#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "trace.h"

#include "hgcd.h"

/* Ways to optimize the code:
 *
 * Compute only the first column of the hgcd and hgcd2 matrices.
 *
 * The second column can be computed later, when needed by
 * schoenhage_step. */

/* This should be a power of 2, to keep the calculation for temporary
   storage simpler. */
#ifndef HGCD_LEHMER_THRESHOLD
# define HGCD_LEHMER_THRESHOLD 32
#endif

#ifndef FGCD_LEHMER_THRESHOLD
# define FGCD_LEHMER_THRESHOLD 64
#endif

#ifndef EUCLID_SUBTRACT_COUNT
# define EUCLID_SUBTRACT_COUNT 2
#endif

/* Must have at least two limbs for hgcd to be defined */
#define HGCD_LIMIT 2

/* For three-limb numbers, one lehmer step may be too large. */
#define LEHMER_STEP_LIMIT 4


#define NORMALIZED(p, l) (!(l) || (p)[ABS((l))-1])

#define MPN_EQUAL_P(ap, asize, bp, bsize)			\
((asize) == (bsize) && !mpn_cmp((ap), (bp), (asize)))

#define MPN_LESS_P(ap, asize, bp, bsize)			\
((asize) < (bsize) || ((asize) == (bsize)			\
                       && mpn_cmp((ap), (bp), (asize)) < 0))

#define MPN_LEQ_P(ap, asize, bp, bsize)				\
((asize) < (bsize) || ((asize) == (bsize)			\
                       && mpn_cmp((ap), (bp), (asize)) <= 0))

#define MPN_MUL_SIZE(asize, bsize) \
((asize) && (bsize) ? ((asize) + (bsize)) : 0)

/* Almost like mpn_mul, but (i) without the requirement that s1size >=
   s2size, and (ii) returns "underflow indication": 1 iff the product
   was one limb shorter then the maximal size s1size + s2size. Input
   must still be non-zero. */
static mp_limb_t
mpn_mul_friendly(mp_ptr rp,
		 mp_srcptr s1p, mp_size_t s1size,
		 mp_srcptr s2p, mp_size_t s2size)
{
  ASSERT(s1size);
  ASSERT(s2size);
  
  if (s1size < s2size)
    MPN_SRCPTR_SWAP(s1p, s1size, s2p, s2size);

  return 0 == mpn_mul(rp, s1p, s1size, s2p, s2size);
}

static int
sizeinbits(mp_limb_t x)
{
  int cnt;

  if (!x)
    return 0;
  
  count_leading_zeros (cnt, x);

  return BITS_PER_MP_LIMB - cnt;
}

#if WANT_CHECKS
void
mpn_check_hgcd(struct hgcd *hgcd, mp_size_t asize)
{
  /* Basic checks, a > 0, at least one non-zero element in each row of
     uv */
  mp_ptr tp;
  
  ASSERT(hgcd->asize);
  ASSERT(!asize || asize >= HGCD_LIMIT);
  ASSERT(NORMALIZED(hgcd->ap, hgcd->asize));
  ASSERT(NORMALIZED(hgcd->bp, hgcd->bsize));
  
  ASSERT(hgcd->uvsize[0] || hgcd->uvsize[1]);
  ASSERT(NORMALIZED(hgcd->uvp[0], hgcd->uvsize[0]));
  ASSERT(NORMALIZED(hgcd->uvp[1], hgcd->uvsize[1]));
  
  ASSERT(hgcd->uvsize[2] || hgcd->uvsize[3]);
  ASSERT(NORMALIZED(hgcd->uvp[2], hgcd->uvsize[2]));
  ASSERT(NORMALIZED(hgcd->uvp[3], hgcd->uvsize[3]));

  /* Check that b <= a */
  ASSERT(MPN_LEQ_P(hgcd->bp, hgcd->bsize, hgcd->ap, hgcd->asize));

  tp = alloca(sizeof(mp_limb_t) * (hgcd->asize + 1));
  
  /* Check on ranges */
  if (asize)
    {
      mp_ptr rp;
      mp_ptr qp;
      mp_size_t rsize;
      
      mp_size_t m = (asize + 1) / 2;
      mp_size_t l = asize / 2 + 5*sizeinbits(asize) + 2;
      unsigned i;

      ASSERT(hgcd->asize > m);
      ASSERT(hgcd->bsize > m);
      
      rp = tp;
      rsize = hgcd->bsize;
      qp = rp + rsize;

      mpn_tdiv_qr(qp, rp, 0,
		  hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
      MPN_NORMALIZE(rp, rsize);
      ASSERT(rsize <= m);
      
      for (i = 0; i<4; i++)
	/* For Euclid and Lehmer, we should have usize,vsize <= m. */
	ASSERT(hgcd->uvsize[i] <= l);
    }
}

static void
hgcd_check_matrix(const struct hgcd *hgcd,
		  mp_srcptr ap, mp_size_t asize,
		  mp_srcptr bp, mp_size_t bsize,
		  mp_srcptr a1p, mp_size_t a1size,
		  mp_srcptr b1p, mp_size_t b1size)
{
  mp_ptr xp[2];
  mp_ptr t1p;
  mp_ptr t2p;
  mp_size_t xsize[2];
  mp_size_t t1size;
  mp_size_t t2size;
  mp_size_t alloc = asize + hgcd->alloc;
  mp_size_t sign = hgcd->sign;
  unsigned i;
      
  xp[0] = alloca(sizeof(mp_limb_t) * 4 * alloc);
  xp[1] = xp[0] + alloc;
  t1p = xp[1] + alloc;
  t2p = t1p + alloc;
  
  for (i = 0; i<2; i++)
    {
      mp_size_t usize = hgcd->uvsize[2*i];
      mp_size_t vsize = hgcd->uvsize[2*i+1];
      mp_limb_t cy;

      ASSERT(usize >= 0);
      ASSERT(vsize >= 0);
      
      t1size = MPN_MUL_SIZE(asize, usize);
      t2size = MPN_MUL_SIZE(bsize, vsize);
      
      if (t1size)
	t1size -= mpn_mul_friendly(t1p, ap, asize,
				   hgcd->uvp[2*i], usize);
      
      if (t2size)
	t2size -= mpn_mul_friendly(t2p, bp, bsize,
				   hgcd->uvp[2*i+1], vsize);

      if (sign < 0)
	MPN_PTR_SWAP(t1p, t1size, t2p, t2size);
      
#if 0
      trace("t1 = %Nd, t2 = %Nd\n", t1p, t1size, t2p, t2size);
#endif
      ASSERT(MPN_LESS_P(t2p, t2size, t1p, t1size));

      cy = mpn_sub(xp[i], t1p, t1size, t2p, t2size);
      ASSERT(!cy);
      xsize[i] = t1size;
      MPN_NORMALIZE(xp[i], xsize[i]);

      /* Switch sign for next row */
      sign = ~sign;
    }
  ASSERT(MPN_EQUAL_P(xp[0], xsize[0], a1p, a1size));
  ASSERT(MPN_EQUAL_P(xp[1], xsize[1], b1p, b1size));
}

int
hgcd_wrap(struct hgcd *hgcd, int (*f)(struct hgcd *))
{
  mp_size_t asize = hgcd->asize;
  mp_size_t bsize = hgcd->bsize;
  mp_ptr ap = alloca(sizeof(mp_limb_t) * (asize + bsize));
  mp_ptr bp = ap + asize;
  int res;
  
  MPN_COPY(ap, hgcd->ap, asize);
  MPN_COPY(bp, hgcd->bp, bsize);
  res = f(hgcd);
  if (res)
    {
      mpn_check_hgcd(hgcd, asize);
      hgcd_check_matrix(hgcd, ap, asize, bp, bsize,
			hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
    }
  
  return res;
}

static void
check_hgcd_2(struct hgcd_2 *h)
{
  ASSERT(h->a[1]);
  ASSERT(h->b[1]);
  ASSERT(h->uv[0] || h->uv[1]);
  ASSERT(h->uv[2] || h->uv[3]);
}

#else
#define mpn_check_hgcd(R, size)
#define hgcd_check_matrix(R, ap, asize, bp, bsize, \
                          a1p, a1size, b1p, b1size)
#define check_hgcd_2(R)
#endif

mp_size_t
mpn_hgcd_itch(mp_size_t asize)
{
  int lg;
  mp_size_t scratch;

  ASSERT(asize);

  lg = sizeinbits(asize);
  ASSERT(asize < (1L << lg));

  /* First, space for the uv matrix.
   * Four numbers of floor(asize/2) + 5 lg + 2 each */
  scratch = 2*asize + 20 * lg + 8;
  
  if (asize <= HGCD_LEHMER_THRESHOLD)
    /* Space for a single euclid or lehmer step at a time. */
    scratch += 2*(asize + 1);

  else
    {
      /* We need enough scratch space for recursive calls.

         Let U(n) = 2 n + 20 #n + 8
	 Let T(n) = T(n/2) + 2 (n+1) + U(n/2)
	          = T(n/2) + 2 (n+1) + 2 n + 20 #n + 8
		  = T(n/2) + 4n + 20 #n + 10
		  = T(n/4) + 2n + 20 (#n - 1) + 10 + 4n + 20 #n + 10
		 <= 8 n + 40 #n + 10 (#n)^2

	 by recursing down to T(N), where N is the lehmer threshold.
      */

      scratch += 8 * asize + 10 * lg * (4 + lg);
    }
  return scratch;
}

void
mpn_hgcd_init(struct hgcd *hgcd,
	      mp_size_t asize,
	      mp_ptr space,
	      mp_size_t alloc)
{
  mp_size_t used = 0;
  unsigned i;
#if 0
  int lg = sizeinbits(asize);
  hgcd->alloc = asize/2 + 5 * lg + 2;
#endif
  hgcd->alloc = asize / 2 + 2;
  
  for (i = 0; i<4; i++)
    {
      hgcd->uvp[i] = space + used;
      used += hgcd->alloc;
    }
  ASSERT(used < alloc);
  hgcd->tp = space + used;
  hgcd->tsize = alloc - used;
  ASSERT(2*(asize + 1) <= hgcd->tsize);
}

#if 0
/* Copied from gmp-4.1.2/mpn/generic/gcdext.c:div2 */
/* Two-limb division optimized for small quotients.  */
static mp_limb_t
div2 (mp_limb_t n1, mp_limb_t n0, mp_limb_t d1, mp_limb_t d0)
{
  if ((mp_limb_signed_t) n1 < 0)
    {
      mp_limb_t q;
      int cnt;
      for (cnt = 1; (mp_limb_signed_t) d1 >= 0; cnt++)
	{
	  d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	  d0 = d0 << 1;
	}

      q = 0;
      while (cnt)
	{
	  q <<= 1;
	  if (n1 > d1 || (n1 == d1 && n0 >= d0))
	    {
	      sub_ddmmss (n1, n0, n1, n0, d1, d0);
	      q |= 1;
	    }
	  d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	  d1 = d1 >> 1;
	  cnt--;
	}

      return q;
    }
  else
    {
      mp_limb_t q;
      int cnt;
      for (cnt = 0; n1 > d1 || (n1 == d1 && n0 >= d0); cnt++)
	{
	  d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	  d0 = d0 << 1;
	}

      q = 0;
      while (cnt)
	{
	  d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	  d1 = d1 >> 1;
	  q <<= 1;
	  if (n1 > d1 || (n1 == d1 && n0 >= d0))
	    {
	      sub_ddmmss (n1, n0, n1, n0, d1, d0);
	      q |= 1;
	    }
	  cnt--;
	}

      return q;
    }
}
#endif

#define HALF_LIMB (((mp_limb_t) 1) << (BITS_PER_MP_LIMB / 2))
#define SHIFT_LEFT(count, x) \
  (((x)[1] << (count)) | ((x)[0] >> (BITS_PER_MP_LIMB - (count))))

/* NOTE: We return with A and B of two limbs, and A mod B of one limb. */
static void
hgcd_2(struct hgcd_2 *h)
{
  ASSERT(h->a[1]);
  ASSERT(h->b[1]);
  ASSERT(h->a[1] >= h->b[1]);
  
  /* Set (u, v; u',v') = (1, 0; 0, 1) */
  memset(h->uv, 0, sizeof(h->uv));
  h->uv[0] = 1;
  h->uv[3] = 1;
  h->sign = 0;
  
  for (;;)
    {
      /* Compute q,r, with r = a - q b < b.
       * Terminate if r fits in a single limb. */ 
      mp_limb_t q = 0;
      mp_limb_t n0 = h->a[0];
      mp_limb_t n1 = h->a[1];
      mp_limb_t d0 = h->b[0];
      mp_limb_t d1 = h->b[1];

      /* Copied from gmp-4.1.2/mpn/generic/gcdext.c:div2 */
      if ((mp_limb_signed_t) n1 < 0)
	{
	  int cnt;
	  for (cnt = 1; (mp_limb_signed_t) d1 >= 0; cnt++)
	    {
	      d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	      d0 = d0 << 1;
	    }

	  q = 0;
	  while (cnt)
	    {
	      q <<= 1;
	      if (n1 > d1 || (n1 == d1 && n0 >= d0))
		{
		  sub_ddmmss (n1, n0, n1, n0, d1, d0);
		  if (n1 == 0)
		    return;
		  q |= 1;
		}
	      d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	      d1 = d1 >> 1;
	      cnt--;
	    }
	}
      else
	{
	  int cnt;
	  for (cnt = 0; n1 > d1 || (n1 == d1 && n0 >= d0); cnt++)
	    {
	      d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	      d0 = d0 << 1;
	    }

	  q = 0;
	  while (cnt)
	    {
	      d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	      d1 = d1 >> 1;
	      q <<= 1;
	      if (n1 > d1 || (n1 == d1 && n0 >= d0))
		{
		  sub_ddmmss (n1, n0, n1, n0, d1, d0);
		  if (n1 == 0)
		    return;
		  q |= 1;
		}
	      cnt--;
	    }

	}
#if 0
      
      mp_limb_t r[2];
      MPN_COPY(r, h->a, 2);
      
      for (;;)
	{
	  mp_limb_t ah, bh;
	  mp_limb_t q1;
      
	  mp_limb_t cy;
	  unsigned shift;
	  count_leading_zeros(shift, r[1]);

	  if (shift)
	    {
	      ah = SHIFT_LEFT(shift, r);
	      bh = SHIFT_LEFT(shift, h->b);
	    }
	  else
	    { /* FIXME: The shift seems not to work for shift == 0.
		 Strange! */
	      ah = r[1];
	      bh = h->b[1];
	    }
	  
	  if (ah < bh)
	    {
	      trace("r = %Nd, b = %Nd, shift = %d, ah = %lu, bh = %lu\n",
		    r, 2, h->b, 2, shift, ah, bh);
	      abort();
	    }
	  ASSERT(ah >= bh);
	  
	  if (bh >= HALF_LIMB)
	    {
	      /* Now bh is large, so we have ah/bh - 1 <= q <= ah/bh */
	      q1 = ah / bh;
	      ASSERT(q1 >= 1);
	      if (q1 == 1)
		{
		  /* This must be the true quotient */
		  cy = mpn_sub_n(r, r, h->b, 2);
		  ASSERT(!cy);
		}
	      else
		{
		  /* q might be one too large */
		  cy = mpn_submul_1(r, h->b, 2, q1 - 1);
		  ASSERT(!cy);
		  if (mpn_cmp(r, h->b, 2) < 0)
		    q1--;		    
		  else
		    {
		      cy = mpn_sub_n(r, r, h->b, 2);
		      ASSERT(!cy);
		      ASSERT(mpn_cmp(r, h->b, 2) < 0);
		    }
		}
	      
	      q += q1;
	      break;
	    }

	  /* Smaller than the true quotient */
	  q1 = ah / (bh + 1);
	  cy = mpn_submul_1(r, h->b, 2, q1);
	  ASSERT(!cy);

	  q += q1;

	  if (mpn_cmp(r, h->b, 2) < 0)
	    break;
	}
      if (!r[1])
	break;

      /* Ok, keep this value. */
      MPN_COPY(h->a, r, 2);
#endif
      
      ASSERT(n1 != 0);
      h->a[0] = n0;
      h->a[1] = n1;
      
      /* The absolute values are computed as
       *
       * (|u_{i+2}|, |v_{i+2}|)
       *   = (|u_i|, |v_i|) + q (|u_{i+1}|, |v_{i+1}|)
       *
       * They will always fit in a single limb each, thanks to the
       * relations a_0 < 2^2k, a_i >= 2^k, v_{i+1} <= a_0 / a_i < 2^k.
       */

#if 0
      /* This doesn't matter, at least not on x86 */
      if (q == 1)
	{
	  h->uv[0] += h->uv[2];
	  h->uv[1] += h->uv[3];
	}
      else
#endif
	{
	  h->uv[0] += q * h->uv[2];
	  h->uv[1] += q * h->uv[3];
	}
      /* FIXME: This swap could be rationalized away by some more
	 book-keping */
      MP_LIMB_T_SWAP(h->uv[0], h->uv[2]);
      MP_LIMB_T_SWAP(h->uv[1], h->uv[3]);

      MP_LIMB_T_SWAP(h->a[0], h->b[0]);
      MP_LIMB_T_SWAP(h->a[1], h->b[1]);
      h->sign = ~h->sign;
    }
}

/* Computes U A - V B. Returns limb count of the result, with sign. */
static mp_size_t
hgcd_2_sub(mp_ptr r, mp_size_t ralloc,
	   mp_limb_t u, mp_ptr ap, mp_size_t asize,
	   mp_limb_t v, mp_ptr bp, mp_size_t bsize)
{
  mp_limb_t cy;
  mp_size_t size;

  /* asize + 1 <= #(U A) <= asize + 2
     bsize + 1 <= #(V B) <= bsize + 2
     #(V B) <= #(U A)

     implies

     bsize <= asize + 1
  */

  size = asize;
  ASSERT(size <= ralloc);
  
  /* R = U * A */
  cy = mpn_mul_1(r, ap, asize, u);
  if (cy)
    {
      ASSERT(size < ralloc);
      r[size++] = cy;
    }
  
  /* R -= V * B */
  ASSERT(bsize <= size);
  
  cy = mpn_submul_1(r, bp, bsize, v);
  if (cy && bsize < size)
    cy = mpn_sub_1(r + bsize, r + bsize, size - bsize, cy);

  ASSERT(!cy);
  
  MPN_NORMALIZE(r, size);
  
  return size;
}

/* Computes U A + V B */
static mp_size_t
hgcd_2_abs(mp_ptr r, mp_size_t ralloc,
	   mp_limb_t u, mp_ptr ap, mp_size_t asize,
	   mp_limb_t v, mp_ptr bp, mp_size_t bsize)
{
  mp_size_t size;
  mp_limb_t cy;

  if (!u)
    asize = 0;
  if (!v)
    bsize = 0;
  
  if (asize < bsize)
    {
      MPN_PTR_SWAP(ap, asize, bp, bsize);
      MP_LIMB_T_SWAP(u, v);
    }
  if (!asize)
    return 0;
  
  ASSERT(asize >= bsize);
  ASSERT(asize);
  ASSERT(u);
  
  ASSERT(ap[asize-1]);
  
  /* R = U * A */
  size = asize;
  ASSERT(size <= ralloc);
  
  cy = mpn_mul_1(r, ap, asize, u);
  if (cy)
    {
      ASSERT(size < ralloc);
      r[size++] = cy;
    } 
  
  /* R += V * B */
  if (bsize)
    {
      ASSERT(v);
      cy = mpn_addmul_1(r, bp, bsize, v);
      if (cy && bsize < size)
	cy = mpn_add_1(r + bsize, r + bsize, size - bsize, cy);
      if (cy)
	{
	  ASSERT(size < ralloc);
	  r[size++] = cy;
	}
    }

  ASSERT(size);
  ASSERT(r[size-1]);

  return size;
}

/* Multiplies the values in R by S, from the left. */
static void
hgcd_2_update(struct hgcd_2 *S, struct hgcd *R)
{
  unsigned i;

  { /* Compute new (a;b) <-- (u,v;u',v')(a;b). */
    mp_ptr ap;
    mp_ptr bp;      
    mp_size_t asize;
    mp_size_t bsize;

    asize = R->asize;
    bsize = R->bsize;

    ASSERT(R->tsize >= asize + bsize + 2);
  
    ap = R->tp;
    bp = R->tp + asize + 1;

    if (S->sign < 0)
      {
	/* -u a + vb */
	asize = hgcd_2_sub(ap, asize + 1,
			   S->uv[1], R->bp, R->bsize,
			   S->uv[0], R->ap, R->asize);
	/* u'a - v'b */
	bsize = hgcd_2_sub(bp, bsize + 1,
			   S->uv[2], R->ap, R->asize,
			   S->uv[3], R->bp, R->bsize);
      }
    else
      {
	/* u a - vb */
	asize = hgcd_2_sub(ap, asize + 1,
			   S->uv[0], R->ap, R->asize,
			   S->uv[1], R->bp, R->bsize);
	/* -u'a + v'b */
	bsize = hgcd_2_sub(bp, bsize + 1,
			   S->uv[3], R->bp, R->bsize,
			   S->uv[2], R->ap, R->asize);
      }

    ASSERT(asize > 0);
    ASSERT(bsize > 0);
      
    if (MPN_LESS_P(ap, asize, bp, bsize))
      {
	/* Swap a and b, as well as the rows of S.
	 *
	 * FIXME: It's probably more correct to subtract a from b
	 *
	 * The size difference should be small here, so that we don't
	 * violate the expectations on the sizes of the resulting a
	 * and b. */
	MPN_PTR_SWAP(ap, asize, bp, bsize);
	MP_LIMB_T_SWAP(S->uv[0], S->uv[2]);
	MP_LIMB_T_SWAP(S->uv[1], S->uv[3]);	
      }

    ASSERT(NORMALIZED(ap, asize));
    ASSERT(NORMALIZED(bp, bsize));
    
    MPN_COPY(R->ap, ap, asize); R->asize = asize;
    MPN_COPY(R->bp, bp, bsize); R->bsize = bsize;    
  }

  /* Handle the R matrix, one column at a time */
  for (i = 0; i<2; i++)
    {
      mp_size_t u1size = R->uvsize[i];
      mp_size_t u2size = R->uvsize[i+2];
      mp_ptr u1p = R->uvp[i];
      mp_ptr u2p = R->uvp[i+2];
      mp_ptr n1p; mp_size_t n1size;
      mp_ptr n2p; mp_size_t n2size;

      ASSERT(u1size >= 0);
      ASSERT(u2size >= 0);
      
      /* Add absolute values. */

      mp_size_t alloc = MAX(u1size, u2size) + 2;
      ASSERT(R->alloc >= alloc);
      ASSERT(R->tsize >= 2*alloc);

      n1p = R->tp;
      n1size = hgcd_2_abs(n1p, alloc,
			  S->uv[0], u1p, u1size,
			  S->uv[1], u2p, u2size);
      ASSERT(NORMALIZED(n1p, n1size));
		 
      n2p = R->tp + n1size;
      n2size = hgcd_2_abs(n2p, alloc,
			  S->uv[2], u1p, u1size,
			  S->uv[3], u2p, u2size);
      ASSERT(NORMALIZED(n2p, n2size));

      MPN_COPY(R->uvp[i], n1p, n1size);
      MPN_COPY(R->uvp[i+2], n2p, n2size);

      R->uvsize[i] = n1size;
      R->uvsize[i+2] = n2size;
    }
  R->sign ^= S->sign;
}

static void
hgcd_update_1_ui(struct hgcd *hgcd, mp_limb_t q)
{
  /* Operate on one column at a time */
  unsigned i;
  for (i = 0; i<2; i++)
    {
      mp_size_t u1size = hgcd->uvsize[i];
      mp_size_t u2size = hgcd->uvsize[i+2];

      mp_ptr u1p = hgcd->uvp[i];
      mp_ptr u2p = hgcd->uvp[i+2];

      mp_limb_t c1, c2;
      mp_size_t size;
      
      ASSERT(hgcd->alloc >= u1size);
      ASSERT(hgcd->alloc >= u2size);

      ASSERT(u1size >= 0);
      ASSERT(u2size >= 0);

      /* Compute u1 += q u2 (these are absolute values */

      if (!u2size)
	/* Do nothing */
	size = u1size;
      
      else if (q == 1)
	{
	  if (u1size >= u2size)
	    {
	      size = u1size;
	      
	      c1 = mpn_add(u1p, u1p, u1size, u2p, u2size);
	      if (c1)
		{
		  ASSERT(size < hgcd->alloc);
		  u1p[size++]= c1;
		}
	    }
	  else
	    {
	      size = u2size;
	      if (u1size)
		{
		  c1 = mpn_add(u1p, u2p, u2size, u1p, u1size);
		  if (c1)
		    {
		      ASSERT(size < hgcd->alloc);
		      u1p[size++]= c1;
		    }
		}
	      else
		MPN_COPY(u1p, u2p, u2size);
	    }
	}
      else if (u1size >= u2size)
	{
	  size = u1size;

	  c1 = mpn_addmul_1(u1p, u2p, u2size, q);
	  if (c1 && u1size > u2size)
	    c1 = mpn_add_1(u1p + u2size, u1p + u2size,
			   u1size - u2size, c1);
	  
	  if (c1)
	    {
	      ASSERT(size < hgcd->alloc);
	      u1p[size++] = c1;
	    }
	}
      else if (u1size)
	{
	  ASSERT(u1size < u2size);
	  size = u2size;
	  c1 = mpn_addmul_1(u1p, u2p, u1size, q);
	  c2 = mpn_mul_1(u1p + u1size, u2p + u1size, u2size - u1size, q);

	  if (c2)
	    {
	      ASSERT(size < hgcd->alloc);
	      u1p[size++] = c1;
	    }
	  if (c1)
	    c1 = mpn_add_1(u1p + u1size,
			   u1p + u1size, size - u1size, c1);
	      
	  if (c1)
	    {
	      ASSERT(size < hgcd->alloc);
	      u1p[size++] = c1;	      
	    }
	}
      else
	{
	  /* u1 == 0, compute u1 = q u2 */
	  size = u2size;
	  c1 = mpn_mul_1(u1p, u2p, u2size, q);
	  if (c1)
	    {
	      ASSERT(size < hgcd->alloc);
	      u1p[size++] = c1;
	    }
	}
      hgcd->uvsize[i] = size;
      MPN_PTR_SWAP(hgcd->uvp[i], hgcd->uvsize[i],
		   hgcd->uvp[i+2], hgcd->uvsize[i+2]);
    }
  hgcd->sign = ~hgcd->sign;
}

static void
hgcd_update_1(struct hgcd *hgcd,
	      /* Scratch space */
	      mp_ptr tp, mp_size_t talloc,
	      mp_srcptr qp, mp_size_t qsize)
{
  /* Operate on one column at a time */
  unsigned i;

  /* FIXME: Move the check for qsize == 1 here? */
  for (i = 0; i<2; i++)
    {
      mp_size_t u1size = hgcd->uvsize[i];
      mp_size_t u2size = hgcd->uvsize[i+2];
      mp_ptr u1p = hgcd->uvp[i];
      mp_ptr u2p = hgcd->uvp[i+2];
      mp_size_t tsize = MPN_MUL_SIZE(qsize, u2size);

      ASSERT(hgcd->alloc >= u1size);
      ASSERT(hgcd->alloc >= u2size);

      ASSERT(u1size >= 0);
      ASSERT(u2size >= 0);
      
      ASSERT(tsize <= talloc);
      ASSERT(tsize < hgcd->alloc);

      /* Compute u1 += q u2 (these are absolute values) */
      if (tsize)
	{
	  if (u1size)
	    {
	      mp_limb_t cy;
	      tsize -= mpn_mul_friendly(tp, u2p, u2size, qp, qsize);
		  
	      if (u1size >= tsize)
		{
		  cy = mpn_add(u1p, u1p, u1size, tp, tsize);
		}
	      else
		{
		  ASSERT(tsize <= hgcd->alloc);
		  cy = mpn_add(u1p, tp, tsize, u1p, u1size);
		  u1size = tsize;
		}
	      if (cy)
		{
		  ASSERT(u1size < hgcd->alloc);
		  u1p[u1size++] = cy;
		}
	    }
	  else
	    {
	      /* u1 = 0, Overwrite u1 directly when forming the
		 product. */
	      tsize -= mpn_mul_friendly(u1p, u2p, u2size, qp, qsize);
	      u1size = hgcd->uvsize[i+2] < 0 ? -tsize : tsize;
	    }
	}
      MPN_PTR_SWAP(hgcd->uvp[i], hgcd->uvsize[i],
		   hgcd->uvp[i+2], hgcd->uvsize[i+2]);
    }
  hgcd->sign = ~hgcd->sign;  
}

/* Like mpn_tdiv_qr, but optimized for small quotients, and returning
 * the size of q. Requires n > d. */
static inline mp_size_t
divn(mp_ptr qp, mp_ptr rp,
     mp_srcptr np, mp_size_t nsize, mp_srcptr dp,
     mp_size_t dsize)
{
  unsigned i;
  mp_limb_t cy;
  mp_size_t rsize;
  mp_size_t qsize;

  /* Necessary to have room in r */
  if (nsize <= dsize + 1)
    {
      rsize = nsize;
      cy = mpn_sub(rp, np, nsize, dp, dsize);
      ASSERT(!cy);
      
      MPN_NORMALIZE(rp, rsize);
      
      for (i = 1;; i++)
	{
	  if (MPN_LEQ_P(dp, dsize, rp, rsize))
	    {
	      if (i < EUCLID_SUBTRACT_COUNT)
		{
		  cy = mpn_sub(rp, rp, rsize, dp, dsize);
		  ASSERT(!cy);
		  MPN_NORMALIZE(rp, rsize);
		}
	      else
		break;
	    }
	  else
	    {
	      *qp = i;
	      return 1;
	    }
	}
    }
  
  /* For simplicity, through away any old r and start over. */
  qsize = nsize - dsize + 1;
  mpn_tdiv_qr(qp, rp, 0, np, nsize, dp, dsize);
  
  MPN_NORMALIZE(qp, qsize);
  
  return qsize;
}

/* If the remainder is of size <= m, return 0, and don't modify
 * a and b */
static int
euclid_step(struct hgcd *hgcd, mp_size_t m)
{
  mp_size_t rsize = hgcd->bsize;
  mp_size_t qsize = hgcd->asize - hgcd->bsize + 1;

  mp_ptr qp = hgcd->tp;
  mp_ptr rp = qp + qsize;
  
  ASSERT(hgcd->tsize >= qsize + rsize + 1);
  /* trace("euclid_step\n"); */

  qsize = divn(qp, rp, hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  MPN_NORMALIZE(rp, rsize);
  ASSERT(qsize);
  
  /* Treat m == 0 specially */
  if (m && rsize <= m)
    return 0;
  
  MPN_COPY(hgcd->ap, rp, rsize);
  MP_PTR_SWAP(hgcd->ap, hgcd->bp);
  hgcd->asize = hgcd->bsize;
  hgcd->bsize = rsize;

  /* Update the (u,v;u',v') matrix, by multiplying from the left with
   *
   * / 0  1 \    u,v   <-  u',v'
   * \ 1 -q /    u',v' <-  (u,v) - q (u',v')
   *
   * Compute it as u -= q u', v -= q v', and then swap.
   */

  /* Q is usually small */
  if (qsize == 1)
    {
      hgcd_update_1_ui(hgcd, qp[0]);
      mpn_check_hgcd(hgcd, 0);
    }
  else
    {
      hgcd_update_1(hgcd, hgcd->tp + qsize, hgcd->tsize - qsize,
		    qp, qsize);
      mpn_check_hgcd(hgcd, 0);
    }

  return 1;
}

/* Does not guarantee progress. Returns zero if the next remainder is
 * smaller than m. */
static int
lehmer_step(struct hgcd *hgcd, mp_size_t m)
{
  struct hgcd_2 h2;
  mp_size_t size;
  
  ASSERT(hgcd->bsize);
  ASSERT(hgcd->asize == hgcd->bsize);

  trace("lehmer_step\n");
  
  size = hgcd->asize;
  ASSERT(size >= LEHMER_STEP_LIMIT);

  ASSERT(hgcd->asize >= HGCD_LIMIT);
  ASSERT(hgcd->bsize >= HGCD_LIMIT);

  MPN_COPY(h2.a, hgcd->ap + size - 2, 2);
  MPN_COPY(h2.b, hgcd->bp + size - 2, 2);

  hgcd_2(&h2);

  check_hgcd_2(&h2);
  mpn_check_hgcd(hgcd, 0);
      
  /* Multiply (a;b) and (u,v;u',v') by the hgcd 2 matrix */
  trace("lehmer_step: before: %Nd, %Nd\n",
	hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  trace("lehmer_step: (u,v;u',v') = %s(%lu,%lu;%lu,%lu)\n",
	h2.sign < 0 ? "-" : "",
	h2.uv[0], h2.uv[1], h2.uv[2], h2.uv[3]);
      
  hgcd_2_update(&h2, hgcd);
      
  trace("lehmer_step: after: %Nd, %Nd\n",
	hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  mpn_check_hgcd(hgcd, 0);
      
  /* We started with a,b < B^size. Now we should have
   *
   *   a > b
   *   a,b >= B^{size - 1} / 2 > B^{size - 2}
   *   a mod b = smaller
   *
   * So if we haven't reduced b by one limb, we have *almost* done
   * so, and it should take at most two euclid steps to fix.
   */

  while (hgcd->bsize == size)
    if (!euclid_step(hgcd, m))
      return 0;

  return 1;
}

/* lehmer_hgcd(HGCD, AP, ASIZE, BP, BSIZE)

   Compute two remainders in the remainder sequence for A/B, such that
   R and R1 are of size larger than M, and size of R mod R1 <= M,
   where M = ceil(ASIZE/2). R and R1 replaces A and B.

   Initial condition: A >= B >= 0, A > 0.

   Returns 0 if B is too small.
*/

static void
lehmer_loop(struct hgcd *hgcd, mp_size_t m)
{
  ASSERT(hgcd->asize >= hgcd->bsize);
  ASSERT(hgcd->asize > m);
  ASSERT(hgcd->bsize > m);

  for (;;)
    {
      /* trace("lehmer_loop: before: m = %d, asize = %d, bsize = %d\n",
	 m, hgcd->asize, hgcd->bsize); */
      mpn_check_hgcd(hgcd, 0);

      /* Lehmer potentially reduce both our values by one limb, so
       * we can't use it when we're getting close. */
      if (hgcd->asize == hgcd->bsize
	  && hgcd->asize > m + 1)

	if (!lehmer_step(hgcd, m))
	  break;
      
      if (!euclid_step(hgcd, m))
	break;
    }
}

/* Computes R = 2^k r + u A' - v B'. When called, we already have r
   stored in the more significant limbs of rp. */

static mp_size_t
hgcd_fix(mp_size_t k, mp_ptr tp, mp_size_t talloc,
	 mp_ptr rp, mp_size_t rsize, mp_size_t ralloc,
	 mp_srcptr up, mp_size_t usize, mp_srcptr ap,
	 mp_srcptr vp, mp_size_t vsize, mp_srcptr bp)
{
  mp_size_t tsize;
  mp_limb_t cy;
  mp_size_t sign = 0;

  /* FIXME: Should normalize ap and bp */

  tsize = MPN_MUL_SIZE(k, usize);
  
  ASSERT(tsize <= talloc);
  ASSERT(tsize <= ralloc);
  ASSERT(usize <= k);

  if (!usize)
    memset(rp, 0, k * sizeof(mp_limb_t));
  else
    {
      tsize -= (0 == mpn_mul(tp,
			     ap, k,
			     up, usize));
      
      trace("hgcd_fix: k = %d, tsize = %d, talloc = %d,\n"
	    "  usize = %d, vsize = %d\n",
	    k, tsize, talloc, usize, vsize);
      
      ASSERT(tsize >= k);
  
      MPN_COPY(rp, tp, k);
      if (tsize <= rsize)
	{
	  cy = mpn_add(rp + k, rp + k, rsize - k, tp + k, tsize - k);
	}
      else
	{
	  cy = mpn_add(rp + k, tp + k, tsize - k, rp + k, rsize - k);
	  rsize = tsize;
	}
  
      if (cy)
	{
	  ASSERT(rsize < ralloc);
	  rp[rsize++] = cy;
	}
    }

  /* A -= v * B' */
  if (vsize)
    {
      tsize = MPN_MUL_SIZE(k, vsize);
      ASSERT(tsize <= talloc);
      ASSERT(tsize <= ralloc);
      ASSERT(vsize <= k);
      tsize -= (0 == mpn_mul(tp,
			     bp, k,
			     vp, vsize));

      if (MPN_LESS_P(tp, tsize, rp, rsize))
	{
	  cy = mpn_sub(rp, rp, rsize,
		       tp, tsize);
	}
      else
	{
	  cy = mpn_sub(rp, tp, tsize,
		       rp, rsize);
	  
	  rsize = tsize;
	  sign = ~sign;
	}

      ASSERT(!cy);
    }

  MPN_NORMALIZE(rp, rsize);
  return sign < 0 ? -rsize : rsize;
}

#if 0
static mp_size_t
get_sign(const struct hgcd *R)
{
  mp_size_t sign = R->uvsize[0];
  if (!sign)
    sign = -R->uvsize[1];

  ASSERT(sign);

  if (sign < 0)
    {
      ASSERT(R->uvsize[0] <= 0);
      ASSERT(R->uvsize[1] >= 0);
      ASSERT(R->uvsize[2] >= 0);
      ASSERT(R->uvsize[3] <= 0);
    }
  else
    {
      ASSERT(R->uvsize[0] >= 0);
      ASSERT(R->uvsize[1] <= 0);
      ASSERT(R->uvsize[2] <= 0);
      ASSERT(R->uvsize[3] >= 0);
    }
    
  return sign;
}
#endif

/* Multiply R by S from the left. */
static void
hgcd_update(const struct hgcd *S, struct hgcd *R,
	    mp_ptr tp, mp_size_t talloc)
{
  unsigned i;

#if 0
  trace("hgcd_update: Before:\n"
	"  R = [ %Nd, %Nd;\n"
	"        %Nd, %Nd ]\n",
	R->uvp[0], R->uvsize[0], R->uvp[1], R->uvsize[1],
	R->uvp[2], R->uvsize[2], R->uvp[3], R->uvsize[3]);

  trace("  S = [ %Nd, %Nd;\n"
	"        %Nd, %Nd ]\n",
	S->uvp[0], S->uvsize[0], S->uvp[1], S->uvsize[1],
	S->uvp[2], S->uvsize[2], S->uvp[3], S->uvsize[3]);
#endif
  
  mp_size_t s0size = S->uvsize[0];
  mp_size_t s1size = S->uvsize[1];
  mp_size_t s2size = S->uvsize[2];
  mp_size_t s3size = S->uvsize[3];
  
  /* One column of R at a time */
  for (i = 0; i<2; i++)
    {
      mp_size_t r1size = R->uvsize[i];
      mp_size_t r2size = R->uvsize[i+2];
      mp_ptr r1p = R->uvp[i];
      mp_ptr r2p = R->uvp[i+2];
      mp_size_t t1size;
      mp_size_t t2size;
      mp_ptr t1p;
      mp_ptr t2p;

      /* The new value for a */
      mp_ptr np;
      mp_size_t nsize;
      
      /* Compute the new value for R[i], and put it in np, nsize */
      /* Allocate space, including one extra limb for the carry from
	 addition. Put the largest block first. */
      t1size = MPN_MUL_SIZE(s0size, r1size);
      t2size = MPN_MUL_SIZE(s1size, r2size);

      ASSERT(t1size < R->alloc);
      ASSERT(t2size < R->alloc);
      ASSERT(t1size + t2size + 1 < talloc);
      
      if (t1size >= t2size)
	{
	  t1p = np = tp;
	  t2p = t1p + t1size + 1;
	}
      else
	{
	  t2p = np = tp;
	  t1p = t2p + t2size + 1;
	}
      
      if (t1size)
	t1size -= mpn_mul_friendly(t1p, S->uvp[0], s0size, r1p, r1size);
      if (t2size)
	t2size -= mpn_mul_friendly(t2p, S->uvp[1], s1size, r2p, r2size);

      if (t2size > t1size)
	MPN_PTR_SWAP(t1p, t1size, t2p, t2size);

      nsize = t1size;
      if (t2size)
	{
	  mp_limb_t cy = mpn_add(np, t1p, t1size, t2p, t2size);
	  if (cy)
	    {
	      ASSERT(nsize < R->alloc);
	      np[nsize++] = cy;
	    }
	}
      else
	{
	  /* np should already have the largest number */
	  ASSERT(np == t1p);
	}

      ASSERT(nsize <= R->alloc);

      /* Compute the new value for the second element. We use
	 temporary space after np. */
      /* NOTE: I'd like to store t2p directly on top of r2p, but
	 mpn_mul doesn't allow source and destination to be the
	 same. */
      t1p = np + nsize;
      t1size = MPN_MUL_SIZE(s2size, r1size);
      t2p = t1p + t1size;
      t2size = MPN_MUL_SIZE(s3size, r2size);
      ASSERT(nsize + t1size + t2size <= talloc);

      ASSERT(t1size <= R->alloc);
      ASSERT(t2size <= R->alloc);
      
      if (t1size)
	t1size -= mpn_mul_friendly(t1p, S->uvp[2], s2size, r1p, r1size);
      if (t2size)
	t2size -= mpn_mul_friendly(t2p, S->uvp[3], s3size, r2p, r2size);

      if (t2size > t1size)
	MPN_PTR_SWAP(t1p, t1size, t2p, t2size);

      if (t2size)
	{
	  mp_limb_t cy = mpn_add(R->uvp[i+2], t1p, t1size, t2p, t2size);
	  if (cy)
	    {
	      ASSERT(t1size < R->alloc);
	      R->uvp[i+2][t1size++] = cy;
	    }
	}
      else
	{
	  MPN_COPY(R->uvp[i+2], t1p, t1size);
	}
      
      /* Now the second element is stored at the right place, and the
	 first is in np, nsize. So copy the latter, and update both
	 signs. */

      MPN_COPY(R->uvp[i], np, nsize);
      R->uvsize[i] = nsize;
      R->uvsize[i+2] = t1size;
    }
  R->sign ^= S->sign;
#if 0
  trace("hgcd_update: After:\n"
	"  R = [ %Nd, %Nd;\n"
	"        %Nd, %Nd ]\n",
	R->uvp[0], R->uvsize[0], R->uvp[1], R->uvsize[1],
	R->uvp[2], R->uvsize[2], R->uvp[3], R->uvsize[3]);
#endif
}

/* Perform a schoenhage step by chopping off the k least significant
   limbs of a */
static void
schoenhage_step(struct hgcd *R, mp_size_t k)
{
  mp_ptr tp;
  struct hgcd S;

  mp_ptr ap;
  mp_ptr bp;

  /* For debugging */
  mp_size_t old_asize;
  mp_size_t old_bsize;
  mp_ptr old_ap;
  mp_ptr old_bp;
  
  ASSERT(k < R->asize);
  
  if (R->bsize <= k)
    return;

  if (R->asize < k + HGCD_LIMIT)
    return;

  old_asize = R->asize;
  old_bsize = R->bsize;
  old_ap = alloca(sizeof(mp_limb_t) * (R->asize + R->bsize));
  old_bp = old_ap + R->asize;

  MPN_COPY(old_ap, R->ap, R->asize);
  MPN_COPY(old_bp, R->bp, R->bsize);
  
  trace("schoenhage_step: before: asize = %d, bsize = %d, k = %d\n"
	"  a = %Nd, b = %Nd\n",
	R->asize, R->bsize, k,
	R->ap, R->asize, R->bp, R->bsize);

  /* We allocate our temporary space as follows: First the space
   * needed for S's uv matrix, and then the rest is made available for
   * the hgcd calculation for S. When that calculation is done, the
   * temporary space for S is reused, we need to store copies of the
   * least significant limbs of a and b (i.e. A' and B'), k limbs
   * each, and hgcd_fix gets to use the rest. */

  if (R->tsize < mpn_hgcd_itch(R->asize - k))
    trace("schoenhage_step: We're getting short on scratch space.\n");
  
  mpn_hgcd_init(&S, R->asize - k, R->tp, R->tsize);
  S.ap = R->ap + k;
  S.bp = R->bp + k;
  S.asize = R->asize - k;
  S.bsize = R->bsize - k;
  
  /* NOTE: Might swap the S.ap and S.bp pointers. */	
  if (!mpn_hgcd(&S))
    {
      /* B already small. */
      return;
    }

  mpn_check_hgcd(&S, R->asize - k);

  trace("schoenhage_step: after mpn_hgcd: asize = %d, bsize = %d\n"
	"  a = %Nd, b = %Nd\n"
	"  uvsize = (%d, %d; %d, %d)\n"
	"  uv = (%Nd, %Nd; %Nd, %Nd)\n",
	S.asize, S.bsize,
	S.ap, S.asize, S.bp, S.bsize,
	S.uvsize[0], S.uvsize[1], S.uvsize[2], S.uvsize[3],
	S.uvp[0], S.uvsize[0], S.uvp[1], S.uvsize[1],
	S.uvp[2], S.uvsize[2], S.uvp[3], S.uvsize[3]);

  /* Copy A' (ap) and B' (bp) into temporary area, as they
     overlap with where we want to store the results */

  ASSERT(2*k < S.tsize);
  MPN_COPY(S.tp, R->ap, k);     ap = S.tp;
  MPN_COPY(S.tp + k, R->bp, k); bp = S.tp + k;
  tp = S.tp + 2*k;

  /* When calling hgcd_fix, we should already have the high order
     limbs, computed by the recursive hgcd call, stored at the right
     place. However, S.ap and S.bp might be swapped, in which case the
     high order limbs of A is stored in R.bp and vice versa. Then we
     swap the ap and bp pointers of R. We use only the high order
     limbs there, as the low order limbs were just copied to temporary
     storage.
  */
  if (S.ap != R->ap + k)
    {
      ASSERT(S.ap == R->bp + k);
      MP_PTR_SWAP(R->ap, R->bp);
    }

  if (S.sign >= 0)
    {
      R->asize = hgcd_fix(k, tp, S.tsize - 2*k,
			  R->ap, k + S.asize, R->asize,
			  S.uvp[0], S.uvsize[0], ap, 
			  S.uvp[1], S.uvsize[1], bp);

      R->bsize = hgcd_fix(k, tp, S.tsize - 2*k,
			  R->bp, k + S.bsize, R->bsize,
			  S.uvp[3], S.uvsize[3], bp,
			  S.uvp[2], S.uvsize[2], ap);
    }
  else
    {
      R->asize = hgcd_fix(k, tp, S.tsize - 2*k,
			  R->ap, k + S.asize, R->asize,
			  S.uvp[1], S.uvsize[1], bp,
			  S.uvp[0], S.uvsize[0], ap);

      R->bsize = hgcd_fix(k, tp, S.tsize - 2*k,
			  R->bp, k + S.bsize, R->bsize,
			  S.uvp[2], S.uvsize[2], ap,
			  S.uvp[3], S.uvsize[3], bp);
    }
  
  if (MPN_LESS_P(R->ap, R->asize, R->bp, R->bsize))
    {
      /* Both a and b must be in acceptable range. Swap them */
      MPN_PTR_SWAP(R->ap, R->asize, R->bp, R->bsize);
      MPN_PTR_SWAP(S.uvp[0], S.uvsize[0],
		   S.uvp[2], S.uvsize[2]);
      MPN_PTR_SWAP(S.uvp[1], S.uvsize[1],
		   S.uvp[3], S.uvsize[3]);
      S.sign = ~S.sign;
    }

#if 0
  /* The a and b parts of S are no longer valid */
  mpn_check_hgcd(&S, 0);
#endif
  hgcd_check_matrix(&S,
		    old_ap, old_asize, old_bp, old_bsize,
		    R->ap, R->asize, R->bp, R->bsize);
		    
  mpn_check_hgcd(R, 0);
  
  /* The available storage is S.tp */  
  hgcd_update(&S, R, S.tp, S.tsize);

  mpn_check_hgcd(R, 0);
}

static unsigned
divide_power_of_two(mp_ptr xp, mp_size_t *xsize)
{
  mp_size_t k;
  int count = 0;
  mp_limb_t cy;
  
  /* Find first non-zero limb. */
  for (k = 0; xp[k] == 0; )
    { k++; ASSERT(k < *xsize); }

  if (k)
    {
      *xsize -= k;
      MPN_COPY_INCR(xp, xp + k, *xsize);
    }

  if ((xp[0] & 1) == 0)
    {
      count_trailing_zeros(count, xp[0]);

      cy = mpn_rshift(xp, xp, *xsize, count);
      ASSERT(!cy);
      /* The macro definition doesn't use as many parentheses as it
	 should. */
      MPN_NORMALIZE(xp, (*xsize));
    }
  return k * BITS_PER_MP_LIMB + count;
}

/* When using a non-lehmer hgcd, we might get some spurious factors */
mp_size_t
mpn_fgcd(mp_ptr rp, mp_ptr ap, mp_size_t asize,
	 mp_ptr bp, mp_size_t bsize)
{
  struct hgcd H;
  mp_size_t scratch;
  
  if (MPN_LESS_P(ap, asize, bp, bsize))
    MPN_PTR_SWAP(ap, asize, bp, bsize);
  
  ASSERT(asize);
  ASSERT(asize >= bsize);

  /* FIXME: Shouldn't use alloca directly */
  scratch = mpn_hgcd_itch(asize);
  mpn_hgcd_init(&H, asize,
		alloca( sizeof(mp_limb_t) * scratch), scratch);
  H.ap = ap;
  H.asize = asize;
  H.bp = bp;
  H.bsize = bsize;
  
  for (;;)
    {
      mp_size_t original = H.asize;
      
      trace("mpn_fgcd: a = %Nd, b = %Nd\n",
	    H.ap, H.asize, H.bp, H.bsize);
  
      if (!H.bsize)
	{
	  MPN_COPY(rp, H.ap, H.asize);
	  return H.asize;
	}
      else if (H.bsize < FGCD_LEHMER_THRESHOLD)
	{
	  /* mpn_gcd actually uses the binary gcd algorithm, not
	     lehmer, and it requires that b is odd. So we must first
	     get rid of any factors of two. */
	  unsigned apower = divide_power_of_two(H.ap, &H.asize);
	  unsigned bpower = divide_power_of_two(H.bp, &H.bsize);
	  unsigned rpower;
	  mp_size_t limbs;
	  unsigned bits;
	  
	  mp_size_t rsize;
	  
	  if (MPN_LESS_P(H.ap, H.asize, H.bp, H.bsize))
	    MPN_PTR_SWAP(H.ap, H.asize, H.bp, H.bsize);

	  rpower = MIN(apower, bpower);
	  limbs = rpower / BITS_PER_MP_LIMB;
	  bits = rpower % BITS_PER_MP_LIMB;
	  
	  rsize = mpn_gcd(rp + limbs,
			  H.ap, H.asize, H.bp, H.bsize);

	  if (bits)
	    {
	      mp_limb_t cy = mpn_lshift(rp + limbs, rp + limbs, rsize, bits);
	      if (cy)
		rp[limbs + rsize++] = cy;
	    }
	  
	  if (limbs)
	    {
	      mp_size_t i;
	      
	      for (i = 0; i < limbs; i++)
		rp[i] = 0;
	      
	      rsize += limbs;
	    }
	  return rsize;
	}
      /* mpn_lehmer_hgcd(&H); */
      if (mpn_hgcd(&H))
	mpn_check_hgcd(&H, original);

      /* Always perform one Euclid step, to guaranteee progress. */
      euclid_step(&H, 0);
    }
}

#undef mpn_hgcd

int
mpn_hgcd(struct hgcd *hgcd)
{
  mp_size_t m = (hgcd->asize + 1) / 2;

  static int entry = 0;
  int count = ++entry;

  /* For debugging */
  mp_size_t old_asize;
  mp_size_t old_bsize;
  mp_ptr old_ap;
  mp_ptr old_bp;
  
  ASSERT(hgcd->asize >= hgcd->bsize);
  ASSERT(hgcd->asize >= 1);
  ASSERT(hgcd->ap[hgcd->asize-1] != 0);
  ASSERT(! hgcd->bsize || hgcd->bp[hgcd->bsize-1] != 0);
  ASSERT(! hgcd->bsize ||
	 ! MPN_OVERLAP_P(hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize));

  /* Space needed for the u's and v's should really be asize - m =
     floor(asize/2), except that isn't tru for asize = 1 */
  ASSERT(hgcd->alloc >= m);

  /* Initialize (u, v; u',v') = identity */
  hgcd->sign = 0;
  hgcd->uvsize[0] = 1; hgcd->uvp[0][0] = 1;
  hgcd->uvsize[1] = 0;
  hgcd->uvsize[2] = 0;
  hgcd->uvsize[3] = 1; hgcd->uvp[3][0] = 1;
  
  if (hgcd->bsize <= m)
    {
      trace("mpn_hgcd(%d): b too small, b = %Nd\n", count,
	    hgcd->bp, hgcd->bsize);
      return 0;
    }

  ASSERT(hgcd->asize >= hgcd->bsize);
  
  if (hgcd->asize == hgcd->bsize)
    {
      mp_limb_t c;
      MPN_CMP(c, hgcd->ap, hgcd->bp, hgcd->asize);
      ASSERT(c >= 0);
      if (!c)
	{
	  trace("mpn_hgcd(%d): a = b = %Nd\n",
		count, hgcd->ap, hgcd->asize);
	  
	  mpn_check_hgcd(hgcd, 0);
	  return 1;
	}
    }

  if (hgcd->asize < HGCD_LEHMER_THRESHOLD)
    {
      trace("mpn_hgcd(%d): calls mpn_lehmer_hgcd b\n", count);
      return mpn_lehmer_hgcd(hgcd);
    }

  old_asize = hgcd->asize;
  old_bsize = hgcd->bsize;
  old_ap = alloca(sizeof(mp_limb_t) * (hgcd->asize + hgcd->bsize));
  old_bp = old_ap + hgcd->asize;

  MPN_COPY(old_ap, hgcd->ap, hgcd->asize);
  MPN_COPY(old_bp, hgcd->bp, hgcd->bsize);
  
  /* Split off half of the limbs */
  trace("mpn_hgcd(%d): calling first schoenhage_step\n", count);
  schoenhage_step(hgcd, hgcd->asize / 2);
  
  ASSERT(hgcd->asize > m);
  ASSERT(hgcd->bsize > m);

  mpn_check_hgcd(hgcd, 0);
  hgcd_check_matrix(hgcd, old_ap, old_asize, old_bp, old_bsize,
		    hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  /* One euclid step, to make sure that both a and b (not just b)
     are of the smaller size */
  trace("mpn_hgcd(%d): calling euclid_step\n", count);
  if (!euclid_step(hgcd, m))
    return 1;
  
  mpn_check_hgcd(hgcd, 0);
  hgcd_check_matrix(hgcd, old_ap, old_asize, old_bp, old_bsize,
		    hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  ASSERT(hgcd->asize > m);
  ASSERT(hgcd->bsize > m);

  /* b should now be of size approximately ceil(3m/2), we want both a
     and b of this size before the second recursive call. */
  /* FIXME: Should we loop over lehmer/euclid to get the numbers in
     range? */
  if (hgcd->bsize > (3*m+1)/2)
    trace("mpn_hgcd(%d): Expected size: %d, have #a = %d, #b = %d\n",
	  count, m+m/2, hgcd->asize, hgcd->bsize);

  if (hgcd->asize > (3*m+1)/2)
    {
      if (hgcd->asize == hgcd->bsize
	  && hgcd->asize > m + 1)
	{
	  if (!lehmer_step(hgcd, m))
	    return 1;
	  
	  mpn_check_hgcd(hgcd, 0);
	  hgcd_check_matrix(hgcd, old_ap, old_asize, old_bp, old_bsize,
			    hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
	}
      if (!euclid_step(hgcd, m))
	return 1;
    }  

  mpn_check_hgcd(hgcd, 0);
  hgcd_check_matrix(hgcd, old_ap, old_asize, old_bp, old_bsize,
		    hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  /* Split off some more limbs, to get the the sizes of a and b close
     to m after the call. */
  trace("mpn_hgcd(%d): calling second schoenhage_step\n", count);
  schoenhage_step(hgcd, 2 * m - hgcd->asize + 1);
  
  ASSERT(hgcd->asize > m);
  ASSERT(hgcd->bsize > m);

  hgcd_check_matrix(hgcd, old_ap, old_asize, old_bp, old_bsize,
		    hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);

  lehmer_loop(hgcd, m);

  hgcd_check_matrix(hgcd, old_ap, old_asize, old_bp, old_bsize,
		    hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  return 1;
}

#undef mpn_lehmer_hgcd

int
mpn_lehmer_hgcd(struct hgcd *hgcd)
{
  mp_size_t m = (hgcd->asize + 1) / 2;

  ASSERT(hgcd->asize >= hgcd->bsize);
  ASSERT(hgcd->asize >= HGCD_LIMIT);
  ASSERT(hgcd->ap[hgcd->asize-1] != 0);
  ASSERT(hgcd->bsize);
  ASSERT(! hgcd->bsize || hgcd->bp[hgcd->bsize-1] != 0);
  ASSERT(! hgcd->bsize
	 || ! MPN_OVERLAP_P(hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize));

  /* Space needed for the u's and v's should really be asize - m =
     floor(asize/2), except that isn't tru for asize = 1 */
  ASSERT(hgcd->alloc >= m);

  /* Initialize (u, v; u',v') = identity */
  hgcd->sign = 0;
  hgcd->uvsize[0] = 1; hgcd->uvp[0][0] = 1;
  hgcd->uvsize[1] = 0;
  hgcd->uvsize[2] = 0;
  hgcd->uvsize[3] = 1; hgcd->uvp[3][0] = 1;

  if (hgcd->bsize <= m)
    return 0;

  lehmer_loop(hgcd, m);

  return 1;
}
