/* Benchmark program */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

/* For getopt */
#include <unistd.h>

#include <gmp.h>
#include <gmp-impl.h>

#include "nhgcd.h"

#include <time.h>

/* Each benchmarked function is called at least BENCH_COUNT times, for
   a total running time at least BENCH_INTERVAL */
#define BENCH_INTERVAL (CLOCKS_PER_SEC/2)
#define BENCH_COUNT 10

static void *
xalloc(size_t size)
{
  void *p = malloc(size);
  if (!p)
    {
      fprintf(stderr, "Virtual memory exhausted.\n");
      abort();
    }
  return p;
}

static double
bench_gcd(mpz_t g, const mpz_t a, const mpz_t b)
{
  clock_t before;
  clock_t after;
  unsigned count = 0;
  
  before = clock();
  
  do {
    mpz_gcd(g, a, b);
    after = clock();
    count++;
  } while (count < BENCH_COUNT || after < before + BENCH_INTERVAL);
  
  return ((double)(after - before)) / CLOCKS_PER_SEC / count;
}

static double
bench_fgcd(const mpz_t g, const mpz_t a, const mpz_t b,
	   mp_size_t (*itch)(mp_size_t asize, mp_size_t bsize),
	   mp_size_t (*f)(mp_ptr gp, mp_srcptr ap, mp_size_t asize,
			  mp_srcptr bp, mp_size_t bsize,
			  mp_ptr tp, mp_size_t talloc))
{
  clock_t before;
  clock_t after;
  unsigned count = 0;

  mp_ptr tp;
  mp_size_t talloc;
  
  mp_size_t asize = mpz_size(a);
  mp_size_t bsize = mpz_size(b);
  mpz_t r;

  mpz_init(r);
  mpz_realloc(r, asize + 1);

  talloc = itch(asize, bsize);
  tp = xalloc(sizeof(mp_limb_t) * talloc);
  
  before = clock();

  do {
    r->_mp_size = f(r->_mp_d, a->_mp_d, asize, b->_mp_d, bsize,
		    tp, talloc);
    after = clock();
    count++;
  } while (count < BENCH_COUNT || after < before + BENCH_INTERVAL);
  
  if (mpz_cmp(g, r) != 0)
    fprintf(stderr, "Bad value from nhgcd_gcd_schoenhage\n");

  return ((double)(after - before)) / CLOCKS_PER_SEC / count;
}

static double
bench_gcdext(mpz_t g, const mpz_t a, const mpz_t b)
{
  mpz_t u;
  clock_t before;
  clock_t after;
  unsigned count = 0;

  mpz_init(u);
  
  before = clock();
  
  do {
    mpz_gcdext(g, u, NULL, a, b);
    after = clock();
    count++;
  } while (count < BENCH_COUNT || after < before + BENCH_INTERVAL);

  mpz_clear(u);
  
  return ((double)(after - before)) / CLOCKS_PER_SEC / count;
}

static int
check_gcdext(const mpz_t a, const mpz_t b,
	     const mpz_t g, const mpz_t u)
{
  mpz_t v;
  mpz_t t;
  
  mpz_init(v);
  mpz_init(t);

  mpz_mul(v, u, a);
  mpz_sub(v, g, v);
  mpz_tdiv_qr(v, t, v, b);

  if (mpz_sgn(t) != 0)
    {
    bad:
      mpz_clear(v);
      mpz_clear(t);
      
      return 0;
    }
  mpz_abs(t, v);
  if (mpz_cmp(v, a) > 0)
    goto bad;
  
  mpz_abs(t, u);
  if (mpz_cmp(u, b) > 0)
    goto bad;

  mpz_clear(v);
  mpz_clear(t);

  return 1;
}

static double
bench_fgcdext(const mpz_t g,
	      const mpz_t a, const mpz_t b)
{
  clock_t before;
  clock_t after;
  unsigned count = 0;

  mp_ptr tp;
  mp_size_t talloc;
  
  mp_size_t asize = mpz_size(a);
  mp_size_t bsize = mpz_size(b);
  mpz_t r;
  mpz_t u;
  
  mpz_init(r);
  mpz_init(u);
  
  mpz_realloc(r, asize + 1);
  mpz_realloc(u, bsize);
  
  talloc = nhgcd_gcdext_schoenhage_itch(asize, bsize);
  tp = xalloc(sizeof(mp_limb_t) * talloc);
  
  before = clock();

  do {
    mp_size_t usize;
    r->_mp_size = nhgcd_gcdext_schoenhage(r->_mp_d, u->_mp_d, &usize,
					  a->_mp_d, asize, b->_mp_d, bsize,
					  tp, talloc);
    u->_mp_size = usize;
    after = clock();
    count++;
  } while (count < BENCH_COUNT || after < before + BENCH_INTERVAL);
  
  if (mpz_cmp(g, r) != 0)
    fprintf(stderr, "Bad g from nhgcd_gcd_schoenhage\n");

  if (!check_gcdext(a, b, g, u))
    fprintf(stderr, "Bad u from nhgcd_gcd_schoenhage\n");
    
  return ((double)(after - before)) / CLOCKS_PER_SEC / count;
}

static void
header(const char *label)
{
  printf("\n%s\nlimbs    gcd     fgcd    gcd/fgcd\n", label);
};

static const char *
choose_unit(double *x)
{
  if (*x < 0.001)
    {
      *x *= 1000000;
      return "us";
    }
  else if (*x < 1.0)
    {
      *x *= 1000;
      return "ms";
    }
  return "s ";
}

static void
display(unsigned limbs, double gcd_time, double fgcd_time)
{
  const char *gcd_unit;
  const char *fgcd_unit;
  
  double q = gcd_time / fgcd_time;
  double l = log(q) / log(2);
  int pos;

  gcd_unit = choose_unit(&gcd_time);
  fgcd_unit = choose_unit(&fgcd_time);

  printf("%5u %6.2f%s %6.2f%s %5.2f",
	 limbs, gcd_time, gcd_unit, fgcd_time, fgcd_unit,
	 q);

  pos = (l + 5 ) * 60 / 20;
  if (pos >= 0 && pos <= 60)
    {
      int i;
      for (i = 0; i<pos; i++)
	printf(" ");
      printf("*");
    }
  printf("\n");
}

#define RANDOM(r, state, size) do {		\
  mpz_urandomb((r), (state), (size));		\
  mpz_setbit(r, (size)-1);			\
} while (0)

int
main(int argc, char **argv)
{
  gmp_randstate_t state;

  mp_size_t (*itch)(mp_size_t asize, mp_size_t bsize);
  mp_size_t (*f)(mp_ptr gp, mp_srcptr ap, mp_size_t asize,
		 mp_srcptr bp, mp_size_t bsize,
		 mp_ptr tp, mp_size_t talloc);
  
  mpz_t a;
  mpz_t b;
  mpz_t c;
  mpz_t g;
  int verbose;
  unsigned limit = 0;
  unsigned i;
  int ch;
  int enable_gcd = 0;
  int enable_extended = 0;
  
  itch = nhgcd_gcd_schoenhage_itch;
  f = nhgcd_gcd_schoenhage;
  verbose = 0;

  while ((ch = getopt (argc, argv, "egvls:")) != -1)
    switch (ch)
      {
      case 's':
	limit = atoi(optarg);
	break;
	
      case 'l':
	itch = nhgcd_gcd_lehmer_itch;
	f = nhgcd_gcd_lehmer;
	break;

      case 'g':
	enable_gcd = 1;
	break;
	
      case 'e':
	enable_extended = 1;
	break;
	
      case 'v':
	verbose = 1;
	break;

      case '?':
	if (isprint (optopt))
	  fprintf (stderr, "Unknown option `-%c'.\n", optopt);
	else
	  fprintf (stderr,
		   "Unknown option character `\\x%x'.\n",
		   optopt);
	return 1;
      default:
	abort ();
      }

  if (!enable_gcd && !enable_extended)
    enable_gcd = enable_extended = 1;
  
  mpz_init(a);
  mpz_init(b);
  mpz_init(c);
  mpz_init(g);

  gmp_randinit_default(state);
  gmp_randseed_ui(state, 4711);

  if (enable_gcd)
    {
      header("Gcd of fibonacci numbers.\n");

      for (i = 10; ; i = i * 4 / 3)
	{
	  double gcd_time;
	  double fgcd_time;
	  unsigned limbs;
      
	  mpz_fib2_ui(a, b, i);

	  limbs = mpz_size(b);

	  if (limit && limbs > limit)
	    break;
      
	  if (verbose && limbs < 1000)
	    gmp_fprintf(stderr,
			"size = %d\n"
			"  a = %Zd\n"
			"  b = %Zd\n",
			limbs, a, b);

	  gcd_time = bench_gcd(g, a, b);
	  fgcd_time = bench_fgcd(g, a, b, itch, f);

	  display(limbs, gcd_time, fgcd_time);
      
	  if (gcd_time > 30.0 || fgcd_time > 30.0)
	    break;
	}
    }

  if (enable_extended)
    {
      header("Gcdext of fibonacci numbers.\n");
      for (i = 10; ; i = i * 4 / 3)
	{
	  double gcdext_time;
	  double fgcdext_time;
	  unsigned limbs;
      
	  mpz_fib2_ui(a, b, i);

	  limbs = mpz_size(b);

	  if (limit && limbs > limit)
	    break;

	  gcdext_time = bench_gcdext(g, a, b);
	  fgcdext_time = bench_fgcdext(g, a, b);

	  display(limbs, gcdext_time, fgcdext_time);
      
	  if (gcdext_time > 30.0 || fgcdext_time > 30.0)
	    break;
	}
    }

  if (enable_gcd)
    {
      header("Gcd of pseudorandom numbers.\n");

      for (i = 10; ; i = i * 4 / 3)
	{
	  double gcd_time;
	  double fgcd_time;
	  unsigned limbs;

	  RANDOM(a, state, i + 47);
	  RANDOM(b, state, i + 17);
	  RANDOM(c, state, i);

	  mpz_mul(a, a, c);
	  mpz_mul(b, b, c);

	  limbs = mpz_size(b);

	  if (limit && limbs > limit)
	    break;

	  if (verbose && limbs < 1000)
	    gmp_fprintf(stderr,
			"size = %d\n"
			"  a = %Zd\n"
			"  b = %Zd\n",
			limbs, a, b);
      
	  gcd_time = bench_gcd(g, a, b);
	  fgcd_time = bench_fgcd(g, a, b, itch, f);

	  display(limbs, gcd_time, fgcd_time);
      
	  if (gcd_time > 30.0 || fgcd_time > 30.0)
	    break;
	}
    }

  if (enable_extended)
    {
      header("Gcdext of pseudorandom numbers.\n");

      for (i = 10; ; i = i * 4 / 3)
	{
	  double gcdext_time;
	  double fgcdext_time;
	  unsigned limbs;

	  RANDOM(a, state, i + 47);
	  RANDOM(b, state, i + 17);
	  RANDOM(c, state, i);

	  mpz_mul(a, a, c);
	  mpz_mul(b, b, c);

	  limbs = mpz_size(b);

	  if (limit && limbs > limit)
	    break;

	  gcdext_time = bench_gcdext(g, a, b);
	  fgcdext_time = bench_fgcdext(g, a, b);

	  display(limbs, gcdext_time, fgcdext_time);
      
	  if (gcdext_time > 30.0 || fgcdext_time > 30.0)
	    break;
	}
    }
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(c);
  mpz_clear(g);

  gmp_randclear(state);
  
  return EXIT_SUCCESS;
}
