#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <gmp.h>
#include <gmp-impl.h>
#include <longlong.h>

#include "trace.h"
#include "nhgcd.h"

#if WANT_INLINE
# define INLINE inline
#else
# define INLINE
#endif

#if GNUC
# define UNUSED __attribute__ ((unused))
#else
# define UNUSED
#endif

#define CHOOSE_M(n) (((n) + 1)/2)
#define CHOOSE_L(n) ((n)/2)

/* This should be tuned while looking at performance for large numbers */
#ifndef HGCD_LEHMER_THRESHOLD
# define HGCD_LEHMER_THRESHOLD 15
#endif

/* While this is important for small and medium size numbers. */
#ifndef FGCD_LEHMER_THRESHOLD
# define FGCD_LEHMER_THRESHOLD 1200
#endif

#ifndef FGCDEXT_LEHMER_THRESHOLD
# define FGCDEXT_LEHMER_THRESHOLD 40
#endif

#ifndef GCD_BINARY_THRESHOLD
# define GCD_BINARY_THRESHOLD 250
#endif

/* Must have at least two limbs for hgcd to be defined */
#define HGCD_LIMIT 2

#define MPN_EQUAL_P(ap, asize, bp, bsize)			\
((asize) == (bsize) && !mpn_cmp((ap), (bp), (asize)))

#define MPN_LESS_P(ap, asize, bp, bsize)			\
((asize) < (bsize) || ((asize) == (bsize)			\
                       && mpn_cmp((ap), (bp), (asize)) < 0))

#define MPN_LEQ_P(ap, asize, bp, bsize)				\
((asize) < (bsize) || ((asize) == (bsize)			\
                       && mpn_cmp((ap), (bp), (asize)) <= 0))

#define MPN_MUL_SIZE(asize, bsize) \
((asize) && (bsize) ? ((asize) + (bsize)) : 0)

#define SHIFT_LEFT(count, x) \
  (((x)[1] << (count)) | ((x)[0] >> (BITS_PER_MP_LIMB - (count))))

/* Copied from mpn/generic/gcdext.c, and modified slightly to return
   the remainder. */
/* Two-limb division optimized for small quotients.  */
static INLINE mp_limb_t
div2 (mp_ptr np, mp_srcptr dp)
{
  mp_limb_t n0 = np[0];
  mp_limb_t n1 = np[1];
  mp_limb_t d0 = dp[0];
  mp_limb_t d1 = dp[1];

  mp_limb_t q = 0;
  
  if ((mp_limb_signed_t) n1 < 0)
    {
      int cnt;
      for (cnt = 1; (mp_limb_signed_t) d1 >= 0; cnt++)
	{
	  d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	  d0 = d0 << 1;
	}

      while (cnt)
	{
	  q <<= 1;
	  if (n1 > d1 || (n1 == d1 && n0 >= d0))
	    {
	      sub_ddmmss (n1, n0, n1, n0, d1, d0);
	      q |= 1;
	    }
	  d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	  d1 = d1 >> 1;
	  cnt--;
	}
    }
  else
    {
      int cnt;
      for (cnt = 0; n1 > d1 || (n1 == d1 && n0 >= d0); cnt++)
	{
	  d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	  d0 = d0 << 1;
	}

      while (cnt)
	{
	  d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	  d1 = d1 >> 1;
	  q <<= 1;
	  if (n1 > d1 || (n1 == d1 && n0 >= d0))
	    {
	      sub_ddmmss (n1, n0, n1, n0, d1, d0);
	      q |= 1;
	    }
	  cnt--;
	}
    }
  
  np[0] = n0;
  np[1] = n1;

  return q;
}

static int
sizeinbits(mp_limb_t x)
{
  int cnt;

  if (!x)
    return 0;
  
  count_leading_zeros (cnt, x);

  return BITS_PER_MP_LIMB - cnt;
}

/* Almost like mpn_add, but without the requirements that s1size >=
   s2size > 0. Returns size of the sum. */
static mp_limb_t
mpn_add_friendly(mp_ptr rp,
		 mp_srcptr s1p, mp_size_t s1size,
		 mp_srcptr s2p, mp_size_t s2size)
{
  mp_limb_t cy;
  
  if (s1size < s2size)
    MPN_SRCPTR_SWAP(s1p, s1size, s2p, s2size);

  if (s1size == 0)
    return 0;

  else if (s2size == 0)
    {
      if (rp != s1p)
	MPN_COPY(rp, s1p, s1size);
    }
  else
    {
      cy = mpn_add(rp, s1p, s1size, s2p, s2size);
      if (cy)
	rp[s1size++] = cy;
    }
  return s1size;
}

/* Almost like mpn_mul, but (i) without the requirement that s1size >=
   s2size, and (ii) returns "underflow indication": 1 iff the product
   was one limb shorter then the maximal size s1size + s2size. Input
   must still be non-zero. */
static mp_limb_t
mpn_mul_friendly(mp_ptr rp,
		 mp_srcptr s1p, mp_size_t s1size,
		 mp_srcptr s2p, mp_size_t s2size)
{
  ASSERT(s1size);
  ASSERT(s2size);
  
  if (s1size < s2size)
    MPN_SRCPTR_SWAP(s1p, s1size, s2p, s2size);

  return 0 == mpn_mul(rp, s1p, s1size, s2p, s2size);
}

/* Compute au + bv */
static mp_size_t
mpn_addmul2_1(mp_ptr rp, mp_size_t ralloc,
	      mp_ptr ap, mp_size_t asize, mp_limb_t u, 
	      mp_ptr bp, mp_size_t bsize, mp_limb_t v)
{
  mp_limb_t cy;
  mp_size_t size;

  if (v != 0 && (u == 0 || asize < bsize))
    {
      MPN_PTR_SWAP(ap, asize, bp, bsize);
      MP_LIMB_T_SWAP(u, v);
    }

  if (asize == 0)
    {
      ASSERT(v == 0 || bsize == 0);
      return 0;
    }
      
  size = asize;
  cy = mpn_mul_1(rp, ap, asize, u);
  if (cy)
    {
      ASSERT(size < ralloc);
      rp[size++] = cy;
    }

  if (bsize != 0 && v != 0)
    {
      ASSERT(bsize <= size);
      cy = mpn_addmul_1(rp, bp, bsize, v);
      if (cy && bsize < size)
	cy = mpn_add_1(rp + bsize,
		       rp + bsize, size - bsize, cy);

      if (cy)
	{
	  ASSERT(size < ralloc);
	  rp[size++] = cy;
	}
    }
  
  return size;
}

/* Returns g and u such that g = u A + v B, for some v. There are
   three different cases for the result:

     g = u A - v B, 0 < u < b, 0 < v < a
     g = a          u = 1, v = 0
     g = b          u = 0, v = 1

   We always return with 0 < u <= b, 0 <= v < a.
*/

mp_limb_t
nhgcd_gcdext_1_odd(mp_limb_t *up, mp_limb_t *vp, mp_limb_t a, mp_limb_t b)
{
  mp_limb_t u0;
  mp_limb_t v0;
  mp_limb_t v1;
  mp_limb_t u1; 

  mp_limb_t B = b;
  mp_limb_t A = a;
  
  /* Through out this function maintain

     a = u0 A - v0 B
     b = u1 A - v1 B

     where A and B are odd. */

  u0 = 1; v0 = 0;
  u1 = b; v1 = a-1;

  if (A == 1)
    {
      *up = u0; *vp = v0;
      return 1;
    }
  
  else if (B == 1)
    {
      *up = u1; *vp = v1;
      return 1;
    }
  
  while (a != b)
    {
      ASSERT(a % 2 == 1);
      ASSERT(b % 2 == 1);

      ASSERT(0 < u0); ASSERT(u0 <= B);
      ASSERT(0 < u1); ASSERT(u1 <= B);

      ASSERT(0 <= v0); ASSERT(v0 < A);
      ASSERT(0 <= v1); ASSERT(v1 < A);
      
      if (a > b)
	{
	  MP_LIMB_T_SWAP(a, b);
	  MP_LIMB_T_SWAP(u0, u1);
	  MP_LIMB_T_SWAP(v0, v1);
	}

      ASSERT(a < b);

      /* Makes b even */
      b -= a;
      
      if (u1 >= u0)
	{
	  ASSERT(v1 >= v0);
	  u1 -= u0;
	  v1 -= v0;
	}
      else
	{
	  ASSERT(v1 < v0); 
	  u1 += (B - u0);
	  v1 += (A - v0);
	}
      
      ASSERT(b % 2 == 0);
      
      do
	{
#if 0
	  trace("a = %lu, b = %lu, u0 = %lu, u1 = %lu\n",
		(unsigned long) a, (unsigned long) b,
		(unsigned long) u0, (unsigned long) u1);
#endif
	  /* As b = u1 A + v1 B is even, while A and B are odd,
	     either both or none of u1, v1 is even */

	  ASSERT(u1 % 2 == v1 % 2);	  
	  if (u1 % 2 == 0)
	    {
	      u1 /= 2;
	      v1 /= 2;
	    }
	  else
	    {
	      /* Overflow-safe way of computing (u1 + B) / 2 */
	      u1 = u1 / 2 + B / 2 + 1;
	      v1 = v1 / 2 + A / 2 + 1;
	    }
	  
	  b /= 2;
	}
      while (b % 2 == 0);
    }

  /* Now g = a = b */
  ASSERT(a == b);
  ASSERT(u1 < B);
  ASSERT(v1 < A);

  ASSERT(A % a == 0);  
  ASSERT(B % a == 0);  
  ASSERT(u0 % (B/a) == u1 % (B/a));
  ASSERT(v0 % (A/a) == v1 % (A/a));

  *up = u0; *vp = v0;
  
  return a;
}

mp_limb_t
nhgcd_gcdext_1(mp_limb_t *up, mp_limb_t *vp, mp_limb_t a, mp_limb_t b)
{
  unsigned shift = 0;
  mp_limb_t g;
  mp_limb_t u;
  mp_limb_t v;
  
  /* We use unsigned values in the range 0, ... B - 1. As the values
     are uniquely determined only modulo B, we can add B at will, to
     get numbers in range or flip the least significant bit. */
  /* Deal with powers of two */
  while (a % 2 == 0 && b % 2 == 0)
    {
      a /= 2; b /= 2; shift++;
    }

  if (b % 2 == 0)
    {
      unsigned k = 0;
      
      do {
	b /= 2; k++;
      } while (b % 2 == 0);

      g = nhgcd_gcdext_1_odd(&u, &v, a, b);

      while (k--)
	{
#if 0
	  trace("a = %lu, b = %lu, u = %lu, v = %lu\n",
		(unsigned long) a, (unsigned long) b,
		(unsigned long) u, (unsigned long) v);
#endif
	  /* We have g = u a + v b, and need to construct
	     g = u'a + v'(2b).
	     
	     If v is even, we can just set u' = u, v' = v/2
	     If v is odd, we can set v' = (v + a)/2, u' = u + b
	  */

	  if (v % 2 == 0)
	    v /= 2;
	  
	  else
	    {
	      u = u + b;
	      v = v/2 + a/2 + 1;
	    }
	  b *= 2;
	}
    }
  else if (a % 2 == 0)
    {
      unsigned k = 0;
      
      do {
	a /= 2; k++;
      } while (a % 2 == 0);

      g = nhgcd_gcdext_1_odd(&u, &v, a, b);

      while (k--)
	{
#if 0
	  trace("a = %lu, b = %lu, u = %lu, v = %lu\n",
		(unsigned long) a, (unsigned long) b,
		(unsigned long) u, (unsigned long) v);
#endif
	  /* We have g = u a + v b, and need to construct
	     g = u'(2a) + v'b.
	     
	     If u is even, we can just set u' = u/2, v' = v.
	     If u is odd, we can set u' = (u + b)/2
	  */

	  if (u % 2 == 0)
	    u /= 2;
	  
	  else
	    {
	      u = u/2 + b/2 + 1;
	      v = v + a;
	    }
	  a *= 2;
	}
    }
  else
    /* Ok, both are odd */
    g = nhgcd_gcdext_1_odd(&u, &v, a, b);

  *up = u;
  *vp = v;
  
  return g << shift;
}

static mp_size_t
nhgcd_tdiv(mp_ptr qp,
	   mp_ptr rp, mp_size_t *rsizep,
	   mp_srcptr ap, mp_size_t asize,
	   mp_srcptr bp, mp_size_t bsize)
{
  mp_size_t qsize = asize - bsize + 1;
  mp_size_t rsize = bsize;

  mpn_tdiv_qr(qp, rp, 0, ap, asize, bp, bsize);
  MPN_NORMALIZE(rp, rsize);
  *rsizep = rsize;

  if (qp[qsize - 1] == 0)
    qsize--;
  
  if (qsize == 1 && qp[0] == 1)
    return 0;

  return qsize;
}


/* Quotient stack */
#if 0
# define trace_stack trace
#else
# define trace_stack 1?0:((void(*)()) NULL)
#endif

mp_size_t
nhgcd_quotient_stack_itch(mp_size_t size)
{
  /* Limit on the recursion depth */
  unsigned k = sizeinbits(1+ size / (HGCD_LEHMER_THRESHOLD  - 5));
  
  ASSERT(2 * k < NHGCD_MAX_QUOTIENTS);
  
  return 2 * size + 12 * (k+1); 
}

#if WANT_CHECKS
static void
nhgcd_quotient_stack_sanity(struct nhgcd_quotient_stack *stack)
{
  mp_size_t next;
  unsigned i;

  for (i = 0, next = 0; i<stack->size_next; i++)
    {
      mp_size_t qsize = stack->size[i];
      ASSERT(next <= stack->limb_alloc);
      
      ASSERT(qsize == 0 || stack->limb[next+qsize - 1] != 0);
      next += qsize;
    }
  
  ASSERT(next == stack->limb_next);
}
#else
# define nhgcd_quotient_stack_sanity(stack)
#endif

void
nhgcd_quotient_stack_reset(struct nhgcd_quotient_stack *stack,
			   mp_size_t asize)
{
  /* Limit on the recursion depth */
  unsigned k = sizeinbits(1 + asize / (HGCD_LEHMER_THRESHOLD  - 5));
  
  stack->size_next = 0;
  stack->limb_next= 0;
  stack->nkeep = 2 * (k + 1);
  ASSERT(stack->nkeep < NHGCD_MAX_QUOTIENTS);
  
  nhgcd_quotient_stack_sanity(stack);
}

void
nhgcd_quotient_stack_init(struct nhgcd_quotient_stack *stack,
			  mp_size_t asize,
			  mp_limb_t *limbs, mp_size_t alloc)
{
  stack->limb = limbs;
  stack->limb_alloc = alloc;
  /* Should depend on input size, we need 2 * recursion depth */
  stack->nkeep = NHGCD_MAX_QUOTIENTS - 1;
  
  nhgcd_quotient_stack_reset(stack, asize);
}


/* Pushing elements */

/* Drop all but the nkeep latest quotients. Drop additional quotients
   if needed to reclain at least SIZE limbs of storage. */
static void
nhgcd_quotient_stack_rotate(struct nhgcd_quotient_stack *stack,
			    mp_size_t size)
{
  unsigned dropped_sizes;
  unsigned kept;
  unsigned i;
  
  mp_size_t dropped_limbs;

  trace_stack("nhgcd_quotient_stack_rotate: size = %d\n", size);
  nhgcd_quotient_stack_sanity(stack);

  if (stack->size_next > stack->nkeep)
    dropped_sizes = stack->size_next - stack->nkeep;
  else
    dropped_sizes = 0;

  for (i = 0, dropped_limbs = 0; i<dropped_sizes; i++)
    dropped_limbs += stack->size[i];

  for (; dropped_limbs < size; dropped_sizes++)
    {
      ASSERT(dropped_sizes < stack->size_next);
      dropped_limbs += stack->size[dropped_sizes];
    }

  ASSERT(dropped_limbs <= stack->limb_next);
  
  kept = stack->size_next - dropped_sizes;

  if (dropped_sizes)
    memmove(stack->size,
	    stack->size + dropped_sizes,
	    kept * sizeof(*stack->size));

  stack->size_next = kept;

  if (dropped_limbs)
    {
      if (dropped_limbs < stack->limb_next)
	{
	  MPN_COPY_INCR(stack->limb, stack->limb + dropped_limbs,
			stack->limb_next - dropped_limbs);
	  ASSERT(dropped_limbs <= stack->limb_next);
	  stack->limb_next -= dropped_limbs;
	}
      else
	stack->limb_next = 0;
    }
  nhgcd_quotient_stack_sanity(stack);
}

/* Prepares stack for pushing SIZE limbs */
static INLINE mp_ptr
nhgcd_quotient_stack_push_start(struct nhgcd_quotient_stack *stack,
				mp_size_t size)
{
  trace_stack("nhgcd_quotient_stack_push_start: size = %d\n", size);

  nhgcd_quotient_stack_sanity(stack);
#if 0
  if (!stack->nkeep)
    {
      ASSERT(size <= stack->limb_alloc);
      return stack->limb;
    }
#endif
  ASSERT(stack->limb_next <= stack->limb_alloc);

  if (size > stack->limb_alloc - stack->limb_next)
    {
      nhgcd_quotient_stack_rotate(stack,
				  size
				  - (stack->limb_alloc - stack->limb_next));
      ASSERT(stack->size_next < NHGCD_MAX_QUOTIENTS);
    }
  else if (stack->size_next >= NHGCD_MAX_QUOTIENTS)
    {
      nhgcd_quotient_stack_rotate(stack, 0); 
    }

  ASSERT(size <= stack->limb_alloc - stack->limb_next);
  ASSERT(stack->size_next < NHGCD_MAX_QUOTIENTS);
  
  return stack->limb + stack->limb_next;
}

static INLINE void
nhgcd_quotient_stack_push_end(struct nhgcd_quotient_stack *stack,
			      mp_size_t size)
{
  ASSERT(stack->size_next < NHGCD_MAX_QUOTIENTS);
  ASSERT(size <= stack->limb_alloc - stack->limb_next);

  trace_stack("nhgcd_quotient_stack_push_end: q = %Nd\n",
	      stack->limb + stack->limb_next, size);

  stack->size[stack->size_next++] = size;
  stack->limb_next += size;
  
  nhgcd_quotient_stack_sanity(stack);
}

static INLINE void
nhgcd_quotient_stack_push_0(struct nhgcd_quotient_stack *stack)
{
  nhgcd_quotient_stack_sanity(stack);

  trace_stack("nhgcd_quotient_stack_push_0\n");
  
  if (stack->size_next >= NHGCD_MAX_QUOTIENTS)
    nhgcd_quotient_stack_rotate(stack, 0);
  
  stack->size[stack->size_next++] = 0;
}

static INLINE void
nhgcd_quotient_stack_push_1(struct nhgcd_quotient_stack *stack, mp_limb_t q)
{
  ASSERT(q >= 2);

  nhgcd_quotient_stack_sanity(stack);

  trace_stack("nhgcd_quotient_stack_push_1: q = %lu\n", (unsigned long) q);
  
  if (stack->limb_next >= stack->limb_alloc)
    {
      nhgcd_quotient_stack_rotate(stack, 1);
    }
  else if (stack->size_next >= NHGCD_MAX_QUOTIENTS)
    {
      nhgcd_quotient_stack_rotate(stack, 0);
    }
  
  stack->size[stack->size_next++] = 1;
  stack->limb[stack->limb_next++] = q;

  nhgcd_quotient_stack_sanity(stack);
}

static INLINE mp_size_t
nhgcd_quotient_stack_push_quotient(struct nhgcd_quotient_stack *stack,
				   mp_ptr *qp,
				   mp_ptr rp, mp_size_t *rsizep,
				   mp_srcptr ap, mp_size_t asize,
				   mp_srcptr bp, mp_size_t bsize)
{
  mp_size_t qsize = asize - bsize + 1;
  mp_size_t rsize = bsize;
  *qp = nhgcd_quotient_stack_push_start(stack, qsize);

  mpn_tdiv_qr(*qp, rp, 0, ap, asize, bp, bsize);
  MPN_NORMALIZE(rp, rsize);
  *rsizep = rsize;

  if ((*qp)[qsize - 1] == 0)
    qsize--;
  
  if (qsize == 1 && (*qp)[0] == 1)
    qsize = 0;

  nhgcd_quotient_stack_push_end(stack, qsize);
  return qsize;
}

static INLINE void
nhgcd_quotient_stack_drop(struct nhgcd_quotient_stack *stack)
{
  ASSERT(stack->size_next);
  trace_stack("nhgcd_quotient_stack_drop\n");
  stack->limb_next -= stack->size[--stack->size_next];
}

/* Get top element, and remove it from the stack. */
static mp_size_t
nhgcd_quotient_stack_pop(struct nhgcd_quotient_stack *stack,
			 mp_srcptr *qp)
{
  mp_size_t qsize;
  ASSERT(stack->size_next);

  qsize = stack->size[--stack->size_next];
  stack->limb_next -= qsize;
  
  *qp = stack->limb + stack->limb_next;

  trace_stack("nhgcd_quotient_stack_pop: q = %Nd\n", *qp, qsize);
  return qsize;
}

/* Get top element */
static mp_size_t
nhgcd_quotient_stack_get_0(struct nhgcd_quotient_stack *stack,
			   mp_srcptr *qp)
{
  mp_size_t qsize;
  ASSERT(stack->size_next);

  qsize = stack->size[stack->size_next - 1]; 
  *qp = stack->limb + stack->limb_next - qsize;

  trace_stack("nhgcd_quotient_stack_get_0: q = %Nd\n", *qp, qsize);
  return qsize;
}


/* Get element just below the top */
static mp_size_t
nhgcd_quotient_stack_get_1(struct nhgcd_quotient_stack *stack,
			   mp_srcptr *qp)
{
  mp_size_t qsize;
  ASSERT(stack->size_next >= 2);

  qsize = stack->size[stack->size_next - 2];
  *qp = stack->limb + stack->limb_next 
    - stack->size[stack->size_next - 1]
    - qsize;
  trace_stack("nhgcd_quotient_stack_get_1: q = %Nd\n", *qp, qsize);
  return qsize;
}

#if WANT_TRACE
void
trace_nhgcd_quotients(const char *label,
		      const struct nhgcd_quotient_stack *stack)
{
  unsigned i;
  mp_size_t limb;
  
  trace("%s: ", label);
  
  for (i = limb = 0; i < stack->size_next; i++);
    {
      mp_size_t qsize = stack->size[i];
      
      ASSERT(limb <= stack->limb_next);
      trace("%Nd, ", stack->limb + limb, qsize);

      limb += qsize;
    }
  trace("\n");
}
#endif

int
nhgcd_quotient_stack_equal(const struct nhgcd_quotient_stack *a,
			   const struct nhgcd_quotient_stack *b)
{
  if (a->size_next > b->size_next)
    {
      const struct nhgcd_quotient_stack *tmp = a;
      a = b;
      b = tmp;
    }
  
  return (a->limb_next <= b->limb_next)
    && !memcmp(a->size, b->size + b->size_next - a->size_next,
	       sizeof(*a->size) * a->size_next)
    && !memcmp(a->limb, b->limb + b->limb_next - a->limb_next,
	       sizeof(*a->limb) * a->limb_next);
}

/* All variants of the nhgcd functions should either return zero, and
   leave the quotient stack unchanged, or push all the quotients from
   the division a/b of the input values, to the division r1/r2 which
   yields the returned r3. */
   
#if WANT_CHECKS
void
nhgcd2_sanity(const struct nhgcd2 *hgcd,
	      mp_srcptr ap, mp_srcptr bp)
{
  int sign = hgcd->sign;
  mp_limb_t limbs[6];;
  mp_ptr t1 = limbs;
  mp_ptr t2 = limbs + 3;
  const struct nhgcd2_row *r;
  mp_limb_t v3 = hgcd->row[3].v;
  unsigned i;
  
  ASSERT(hgcd->row[0].r[1]);
  ASSERT(hgcd->row[1].r[1]);
  ASSERT(hgcd->row[2].r[1]);
  ASSERT(hgcd->row[3].r[1] == 0);

  for (i = 0, r = hgcd->row; i<4; i++, r++, sign = ~sign)
    {
      ASSERT(r->u <= v3);
      ASSERT(r->v <= v3);
      
      t1[2] = mpn_mul_1(t1, ap, 2, r->u);
      t2[2] = mpn_mul_1(t2, bp, 2, r->v);

      if (sign < 0)
	MP_PTR_SWAP(t1, t2);

      ASSERT(MPN_LEQ_P(t2, 3, t1, 3));
      ASSERT_NOCARRY(mpn_sub_n(t1, t1, t2, 3));

      ASSERT(mpn_cmp(t1, r->r, 2) == 0);
    }
}
#endif

void
trace_nhgcd2(const struct nhgcd2 *hgcd, unsigned n)
{
  int sign = hgcd->sign;
  unsigned i;

  for (i = 0; i<n; i++, sign = ~sign)
    {
      trace("r%d = %Nd, u = %s%lu, v = %s%lu\n",
	    i,
	    hgcd->row[i].r, 2,
	    sign < 0 ? "-" : "", (unsigned long) hgcd->row[i].u,
	    sign >= 0 ? "-" : "", (unsigned long) hgcd->row[i].v);
    }
  trace("\n");
};

/* Sets (a, b, c, d)  <--  (b, c, d, a) */
#define NHGCD2_SWAP4_LEFT(row)				\
do {							\
  struct nhgcd2_row __nhgcd2_swap4_left_tmp = row[0];	\
  row[0] = row[1];					\
  row[1] = row[2];					\
  row[2] = row[3];					\
  row[3] = __nhgcd2_swap4_left_tmp;			\
} while (0)

/* Sets (a, b, c, d)  <--  (d, a, b, c) */
#define NHGCD2_SWAP4_RIGHT(row)				\
do {							\
  struct nhgcd2_row __nhgcd2_swap4_right_tmp = row[3];	\
  row[3] = row[2];					\
  row[2] = row[1];					\
  row[1] = row[0];					\
  row[0] = __nhgcd2_swap4_right_tmp;			\
} while (0)

/* Sets (a, b, c, d)  <--  (c, d, a, b) */
#define NHGCD2_SWAP4_2(row)				\
do {							\
  struct nhgcd2_row __nhgcd2_swap4_2_tmp = row[0];	\
  row[0] = row[2];					\
  row[2] = __nhgcd2_swap4_2_tmp;			\
  __nhgcd2_swap4_2_tmp = row[1];			\
  row[1] = row[3];					\
  row[3] = __nhgcd2_swap4_2_tmp;			\
} while (0)


/* Produce r[k] from r[i] and r[j], and push the corresponding
   quotient. */
#define NHGCD2_STEP(i, j, k) do {					\
  sub_ddmmss(hgcd->row[k].r[1], hgcd->row[k].r[0],			\
	     hgcd->row[i].r[1], hgcd->row[i].r[0],			\
	     hgcd->row[j].r[1], hgcd->row[j].r[0]);			\
									\
  /* Could check here for the special case hgcd->row[3].r[1] == 0,	\
     but it's covered by the below condition as well */			\
  if (       hgcd->row[k].r[1] <  hgcd->row[j].r[1]			\
      || (   hgcd->row[k].r[1] == hgcd->row[j].r[1]			\
	  && hgcd->row[k].r[0] <  hgcd->row[j].r[0]))			\
    {									\
	  /* Unit quotient */						\
	  hgcd->row[k].u = hgcd->row[i].u + hgcd->row[j].u;		\
	  hgcd->row[k].v = hgcd->row[i].v + hgcd->row[j].v;		\
									\
	  if (quotients)						\
	    nhgcd_quotient_stack_push_0(quotients);			\
	}								\
      else								\
	{								\
	  mp_limb_t q = 1 + div2(hgcd->row[k].r, hgcd->row[j].r);	\
	  hgcd->row[k].u = hgcd->row[i].u + q * hgcd->row[j].u;		\
	  hgcd->row[k].v = hgcd->row[i].v + q * hgcd->row[j].v;		\
									\
	  if (quotients)						\
	    nhgcd_quotient_stack_push_1(quotients, q);			\
	}								\
} while (0)

/* Returns 0 on failure, otherwise returns 2, 3 or 4 depending on how
   many of the r:s that satisfy Jebelean's criterion. */
/* FIXME: There are two more micro optimizations that could be done to
   this code:

   The div2 function starts with checking the most significant bit of
   the numerator. When we call div2, that bit is know in advance for
   all but the one or two first calls, so we could split div2 in two
   functions, and call the right one.

   We could also have two versions of this code, with and without the
   quotient argument, to avoid checking if it's NULL in the middle of
   the loop. */

int
nhgcd2(struct nhgcd2 *hgcd, mp_srcptr ap, mp_srcptr bp,
       struct nhgcd_quotient_stack *quotients)
{
  /* For all divisions, we special case q = 1, which accounts for
     approximately 41% of the quotients for random numbers (Knuth,
     TAOCP 4.5.3) */
  mp_limb_t s[2];
  mp_limb_t t[2];
  
  ASSERT(ap[1] > bp[1]
	 || (ap[1] == bp[1] && ap[0] >= bp[0]));

  if (!bp[1])
    return 0;
  
  /* Initialize first two rows */
  MPN_COPY(hgcd->row[0].r, ap, 2); hgcd->row[0].u = 1; hgcd->row[0].v = 0;
  MPN_COPY(hgcd->row[1].r, bp, 2); hgcd->row[1].u = 0; hgcd->row[1].v = 1;

  sub_ddmmss(hgcd->row[2].r[1], hgcd->row[2].r[0],
	     hgcd->row[0].r[1], hgcd->row[0].r[0],
	     hgcd->row[1].r[1], hgcd->row[1].r[0]);

  if (hgcd->row[2].r[1] == 0)
    return 0;
  
  if (       hgcd->row[2].r[1] < hgcd->row[1].r[1]
      || (   hgcd->row[2].r[1] == hgcd->row[1].r[1]
	  && hgcd->row[2].r[0] <  hgcd->row[1].r[0]))
    {
      /* Unit quotient */
      hgcd->row[2].v = 1;

      if (quotients)
	nhgcd_quotient_stack_push_0(quotients);
    }
  else
    {
      mp_limb_t q = 1 + div2(hgcd->row[2].r, hgcd->row[1].r);

      if (hgcd->row[2].r[1] == 0)
	return 0;

      hgcd->row[2].v = q;
      
      if (quotients)
	nhgcd_quotient_stack_push_1(quotients, q);
    }

  hgcd->row[2].u = 1;

  /* The simple version of the loop is as follows:
   |
   |   hgcd->sign = 0;
   |   for (;;)
   |     {
   |       (q, r[3]) = divmod(r[1], r[2]);
   |       u[3] = u[1] + q * u[2];
   |       v[3] = v[1] + q * v[2];
   |       nhgcd_quotient_stack_push_1(quotients, q);
   | 	   
   |       if (hgcd->row[3].r[1] == 0)
   |         break;
   |       
   |       NHGCD2_SHIFT4_LEFT(hgcd->row);
   |       hgcd->sign = ~hgcd->sign;
   |     }
   | 
   |   But then we special case for q = 1, and unroll the loop four times
   |   to avoid data movement. */
  
  for (;;)
    {
      NHGCD2_STEP(1, 2, 3);
      if (hgcd->row[3].r[1] == 0)
	{
	  hgcd->sign = 0;
	  break;
	}
      NHGCD2_STEP(2, 3, 0);
      if (hgcd->row[0].r[1] == 0)
	{
	  NHGCD2_SWAP4_LEFT(hgcd->row);
	  hgcd->sign = -1;
	  break;
	}

      NHGCD2_STEP(3, 0, 1);
      if (hgcd->row[1].r[1] == 0)
	{
	  NHGCD2_SWAP4_2(hgcd->row);
	  hgcd->sign = 0;
	  break;
	}

      NHGCD2_STEP(0, 1, 2);
      if (hgcd->row[2].r[1] == 0)
	{
	  NHGCD2_SWAP4_RIGHT(hgcd->row);
	  hgcd->sign = -1;
	  break;
	}
    }

  ASSERT(hgcd->row[0].r[1] != 0);
  ASSERT(hgcd->row[1].r[1] != 0);
  ASSERT(hgcd->row[2].r[1] != 0);
  ASSERT(hgcd->row[3].r[1] == 0);
  
  /* Check r2 */
  /* We always have r2 > u2, v2 */

  if (hgcd->sign >= 0)
    {
      /* Check if r1 - r2 >= u2 - u1 = |u2| + |u1| */
      s[0] = hgcd->row[2].u + hgcd->row[1].u;
      s[1] = (s[0] < hgcd->row[1].u);
    }
  else
    {
      /* Check if r1 - r2 >= v2 - v1 = |v2| + |v1| */
      s[0] = hgcd->row[2].v + hgcd->row[1].v;
      s[1] = (s[0] < hgcd->row[1].v);
    }
  
  sub_ddmmss(t[1], t[0],
	     hgcd->row[1].r[1], hgcd->row[1].r[0],
	     hgcd->row[2].r[1], hgcd->row[2].r[0]);

  if (t[1] < s[1] || (t[1] == s[1] && t[0] < s[0]))
    return 2;
  
  /* Check r3 */

  if (hgcd->sign >= 0)
    {
      /* Check r3 >= max(-u3, -v3) = |u3| */
      if (hgcd->row[3].r[0] < hgcd->row[3].u)
	return 3;

      /* Check r3 - r2 >= v3 - v2 = |v2| + |v1|*/
      s[0] = hgcd->row[3].v + hgcd->row[2].v;
      s[1] = (s[0] < hgcd->row[2].v);
    }
  else
    {
      /* Check r3 >= max(-u3, -v3) = |v3| */
      if (hgcd->row[3].r[0] < hgcd->row[3].v)
	return 3;

      /* Check r3 - r2 >= u3 - u2 = |u2| + |u1| */
      s[0] = hgcd->row[3].u + hgcd->row[2].u;
      s[1] = (s[0] < hgcd->row[2].u);
    }

  sub_ddmmss(t[1], t[0],
	     hgcd->row[2].r[1], hgcd->row[2].r[0],
	     hgcd->row[3].r[1], hgcd->row[3].r[0]);

  if (t[1] < s[1] || (t[1] == s[1] && t[0] < s[0]))
    return 3;

  return 4;
}
#undef NHGCD2_STEP

/* Computes P = R * S. No overlap allowed. */
static void
nhgcd2_mul(struct nhgcd_row *P, mp_size_t alloc,
	   const struct nhgcd2_row *R, const struct nhgcd_row *S)
{
  unsigned i;
  unsigned j;
  
  for (i = 0; i<2; i++)
    for (j = 0; j<2; j++)
      {
	/* Set P[i, j] = R[i, 0] S[0, j] + R[i,1] S[1, j]
	               = u s0 + v s1 */
	P[i].uvsize[j]
	  = mpn_addmul2_1(P[i].uvp[j], alloc,
			  S[0].uvp[j], S[0].uvsize[j], R[i].u, 
			  S[1].uvp[j], S[1].uvsize[j], R[i].v);
      }
}

/* Computes Y = R * X. No overlap allowed.

   FIXME: Ridiculous code duplication. */
static void
nhgcd2_mul_vector(struct nhgcd_row *Y,
		  mp_size_t alloc,
		  const struct nhgcd2_row *R, const struct nhgcd_row *X)
{
  unsigned i;
  
  for (i = 0; i<2; i++)
    {
      /* Set Y[i] = R[i, 0] X[0] + R[i,1] X[1]
	          = u X[0] + v X[0] */
      Y[i].uvsize[0]
	= mpn_addmul2_1(Y[i].uvp[0], alloc,
			X[0].uvp[0], X[0].uvsize[0], R[i].u, 
			X[1].uvp[0], X[1].uvsize[0], R[i].v);
    }
}


static mp_size_t
nhgcd2_fix(mp_ptr rp, mp_size_t ralloc,
	   int sign,
	   mp_limb_t u, mp_srcptr ap, mp_size_t asize,
	   mp_limb_t v, mp_srcptr bp, mp_size_t bsize)
{
  mp_size_t rsize;
  mp_limb_t cy;
  
  if (sign < 0)
    {
      MP_LIMB_T_SWAP(u,v);
      MPN_SRCPTR_SWAP(ap, asize, bp, bsize);
    }

  ASSERT(u > 0);
  
  ASSERT(asize <= ralloc);
  rsize = asize;
  cy = mpn_mul_1(rp, ap, asize, u);
  if (cy)
    {
      ASSERT(rsize < ralloc);
      rp[rsize++] = cy;
    }

  if (v > 0)
    {
      ASSERT(bsize <= rsize);
      cy = mpn_submul_1(rp, bp, bsize, v);
      if (cy)
	{
	  ASSERT(bsize < rsize);
	  ASSERT_NOCARRY(mpn_sub_1(rp + bsize,
				   rp + bsize, rsize - bsize, cy));
	}

      MPN_NORMALIZE(rp, rsize);
    }
  return rsize;
}

mp_size_t
nhgcd_init_itch(mp_size_t asize, mp_size_t bsize)
{
  /* r0 <= a, r1, r2, r3 <= b, but for simplicity, we allocate asize +
     1 for all of them. The size of the uv:s are limited to asize / 2,
     but we allocate one extra limb. */

  return 4 * (asize + 1) + 8 * (CHOOSE_L(asize) + 1);
}

void
nhgcd_init(struct nhgcd *hgcd,
	   mp_size_t asize, mp_size_t bsize UNUSED,
	   mp_limb_t *limbs)
{
  unsigned i;
  unsigned j;
  mp_size_t alloc = CHOOSE_L(asize) + 1;
  
  hgcd->sign = 0;

  for (i = 0; i<4; i++)
    {
      hgcd->row[i].rp = limbs;
      hgcd->row[i].rsize = asize + 1; limbs += asize + 1;
    }

  hgcd->alloc = alloc;
  
  for (i = 0; i<4; i++)
    for (j = 0; j<2; j++)
      {
	hgcd->row[i].uvp[j] = limbs;
	hgcd->row[i].uvsize[j] = alloc; limbs += alloc;
      }
}

#if WANT_CHECKS
void
nhgcd_sanity(const struct nhgcd *hgcd,
	     mp_srcptr ap, mp_size_t asize,
	     mp_srcptr bp, mp_size_t bsize,
	     unsigned start, unsigned end)
{
  int sign;
  unsigned i;
  mp_size_t L = hgcd->row[end-1].uvsize[1];
  mp_ptr t1p;
  mp_ptr t2p;
  const struct nhgcd_row *r;
  
  ASSERT(asize >= bsize);
  ASSERT(L <= asize / 2);

  t1p = alloca(2*(asize + L) * sizeof(mp_limb_t));
  t2p = t1p + (asize + L);

  sign = hgcd->sign;
  if (start % 2)
    sign = ~sign;
  for (i = start, r = &hgcd->row[start]; i<end; i++, sign = ~sign, r++)
    {
      mp_size_t t1size = MPN_MUL_SIZE(asize, r->uvsize[0]);
      mp_size_t t2size = MPN_MUL_SIZE(bsize, r->uvsize[1]);

      ASSERT(r->uvsize[0] <= L);
      ASSERT(r->uvsize[1] <= L);
      
      if (t1size)
	t1size -= mpn_mul_friendly(t1p, ap, asize, r->uvp[0], r->uvsize[0]);

      if (t2size)
	t2size -= mpn_mul_friendly(t2p, bp, bsize, r->uvp[1], r->uvsize[1]);

      if (sign < 0)
	MPN_PTR_SWAP(t1p, t1size, t2p, t2size);

      ASSERT(MPN_LEQ_P(t2p, t2size, t1p, t1size));
      ASSERT_NOCARRY(mpn_sub(t1p, t1p, t1size, t2p, t2size));

      MPN_NORMALIZE(t1p, t1size);
      ASSERT(MPN_EQUAL_P(t1p, t1size, r->rp, r->rsize));
    }
}
#endif

void
trace_nhgcd_row(const struct nhgcd_row *r)
{
  trace("r = %Nd, u = %Nd, v = %Nd\n",
	r->rp, r->rsize,
	r->uvp[0], r->uvsize[0],
	r->uvp[1], r->uvsize[1]);
}

void
trace_nhgcd(struct nhgcd *hgcd)
{
  unsigned i;
  
  trace("sign = %d\n", hgcd->sign);
  for (i = 0; i<4; i++)
    trace_nhgcd_row(hgcd->row+i);
}

int
nhgcd_equal(const struct nhgcd *A, const struct nhgcd *B)
{
  unsigned i;
  
  if (A->sign != B->sign)
    return 0;

  for (i = 0; i<4; i++)
    {
      unsigned j;
      
      if (!MPN_EQUAL_P(A->row[i].rp, A->row[i].rsize,
		       B->row[i].rp, B->row[i].rsize))
	return 0;

      for (j = 0; j < 2; j++)
	if (!MPN_EQUAL_P(A->row[i].uvp[j], A->row[i].uvsize[j],
			 B->row[i].uvp[j], B->row[i].uvsize[j]))
	  return 0;
    }

  return 1;
}

/* Sets (a, b, c, d)  <--  (b, c, d, a) */
#define NHGCD_SWAP4_LEFT(row)				\
do {							\
  struct nhgcd_row __nhgcd_swap4_left_tmp = row[0];	\
  row[0] = row[1];					\
  row[1] = row[2];					\
  row[2] = row[3];					\
  row[3] = __nhgcd_swap4_left_tmp;			\
} while (0)

/* Sets (a, b, c, d)  <--  (d, a, b, c) */
#define NHGCD_SWAP4_RIGHT(row)				\
do {							\
  struct nhgcd_row __nhgcd_swap4_right_tmp = row[3];	\
  row[3] = row[2];					\
  row[2] = row[1];					\
  row[1] = row[0];					\
  row[0] = __nhgcd_swap4_right_tmp;			\
} while (0)

/* Sets (a, b, c, d)  <--  (c, d, a, b) */
#define NHGCD_SWAP4_2(row)				\
do {							\
  struct nhgcd_row __nhgcd_swap4_2_tmp = row[0];	\
  row[0] = row[2];					\
  row[2] = __nhgcd_swap4_2_tmp;				\
  __nhgcd_swap4_2_tmp = row[1];				\
  row[1] = row[3];					\
  row[3] = __nhgcd_swap4_2_tmp;				\
} while (0)

/* Sets (a, b, c)  <--  (b, c, a) */
#define NHGCD_SWAP3_LEFT(row)				\
do {							\
  struct nhgcd_row __nhgcd_swap4_left_tmp = row[0];	\
  row[0] = row[1];					\
  row[1] = row[2];					\
  row[2] = __nhgcd_swap4_left_tmp;			\
} while (0)

/* Computes (u2, v2) = (u0, v0) + q (u1, v1)
 *
 * FIXME: Notation in the function not quite consistent */

static void
nhgcd_update_uv(struct nhgcd_row *r, mp_srcptr qp, mp_size_t qsize,
		/* Limbs allocated for the u and v, for sanity
		   checking */
		mp_size_t alloc)
{
  unsigned i;

  ASSERT(r[0].uvsize[0] < alloc);
  ASSERT(r[0].uvsize[1] < alloc);
  ASSERT(r[1].uvsize[0] < alloc);
  ASSERT(r[1].uvsize[1] < alloc);
  
  /* Compute u3  = u1 + q u2 */
  
  if (qsize == 0)
    /* Represents a unit quotient */
    for (i = 0; i<2; i++)
      {	
	mp_size_t u1size = r[0].uvsize[i];
	mp_size_t u2size = r[1].uvsize[i];
	mp_srcptr u1p = r[0].uvp[i];
	mp_srcptr u2p = r[1].uvp[i];
	mp_ptr u3p = r[2].uvp[i];

	if (u1size == 0)
	  {
	    /* u3 = u2 */
	    MPN_COPY(u3p, u2p, u2size);
	    r[2].uvsize[i] = u2size; 
	  }
	else
	  {	      
	    if (u2size > 0)
	      {
		/* Compute u3 = u1 + u2, then swap */
		mp_limb_t cy;
		mp_size_t u3size;
		
		ASSERT(u2size >= u1size);
		
		cy = mpn_add(u3p, u2p, u2size, u1p, u1size);
		u3size = u2size;
		if (cy)
		  {
		    ASSERT(u3size < alloc);
		    u3p[u3size++] = cy;
		  }
		r[2].uvsize[i] = u3size; 
	      }
	    else
	      {
		/* u2 = 0, u3 = u1 */
		MPN_COPY(u3p, u1p, u1size);
		r[2].uvsize[i] = u1size; 
	      }
	  }
      }
  else if (qsize == 1)
    {
      mp_limb_t q = qp[0];
      for (i = 0; i<2; i++)
	{
	  mp_size_t u1size = r[0].uvsize[i];
	  mp_size_t u2size = r[1].uvsize[i];
	  mp_srcptr u1p = r[0].uvp[i];
	  mp_srcptr u2p = r[1].uvp[i];
	  mp_ptr u3p = r[2].uvp[i];
	  
	  mp_size_t u3size;

	  mp_limb_t cy;
      
	  if (u1size == 0)
	    {
	      /* Compute u1 = q u2 */
	      u3size = u2size;
	      cy = mpn_mul_1(u3p, u2p, u2size, q);
	      if (cy)
		{
		  ASSERT(u3size < alloc);
		  u3p[u3size++] = cy;
		}
	    }
	  else if (u2size > 0)
	    {
	      ASSERT(u2size >= u1size);
	      u3size = u2size;
	      if (u3size > u1size)
		{
		  /* Do high limbs first */
		  cy = mpn_mul_1(u3p + u1size,
				 u2p + u1size, u3size - u1size, q);
		  if (cy)
		    {
		      ASSERT(u3size < alloc);
		      u3p[u3size++] = cy;
		    }
		}
	      /* FIXME: Use mpn_addmul_1? */
	      cy = mpn_mul_1(u3p, u2p, u1size, q);
	      cy += mpn_add_n(u3p, u3p, u1p, u1size);

	      if (cy && u3size > u1size)
		cy = mpn_add_1(u3p + u1size,
			       u3p + u1size, u3size - u1size, cy);
	      if (cy)
		{
		  ASSERT(u3size < alloc);
		  u3p[u3size++] = cy;
		}
	    }
	  else
	    {
	      /* u2 == 0, u3 = u1 */
	      MPN_COPY(u3p, u1p, u1size);
	      u3size = u1size;
	    }

	  r[2].uvsize[i] = u3size;
	}
    }
  
  else for (i = 0; i<2; i++)
    {
      mp_size_t u1size = r[0].uvsize[i];
      mp_size_t u2size = r[1].uvsize[i];
      mp_srcptr u1p = r[0].uvp[i];
      mp_srcptr u2p = r[1].uvp[i];
      mp_ptr u3p = r[2].uvp[i];

      mp_size_t u3size;

      if (u1size == 0)
	{
	  u3size = qsize + u2size;
	  ASSERT(u3size < alloc);

	  u3size -= mpn_mul_friendly(u3p, u2p, u2size, qp, qsize);
	}
      else if (u2size > 0)
	{
	  mp_limb_t cy;
	  
	  u3size = qsize + u2size;
	  ASSERT(u3size < alloc);

	  u3size -= mpn_mul_friendly(u3p, u2p, u2size, qp, qsize);

	  ASSERT(u3size >= u1size);
	  cy = mpn_add(u3p, u3p, u3size, u1p, u1size);
	  if (cy)
	    {
	      ASSERT(u3size < alloc);
	      u3p[u3size++] = cy;
	    }
	}
      else
	{
	  /* u2 == 0, u3 = u1 */
	  MPN_COPY(u3p, u1p, u1size);
	  u3size = u1size;
	}
      
      r[2].uvsize[i] = u3size;
    }
  
  ASSERT(r[1].uvsize[0] <= r[2].uvsize[0]);
  ASSERT(r[1].uvsize[1] <= r[2].uvsize[1]);
}

/* Computes u2 = u0 + q u1
 *
 * FIXME: Notation in the function not quite consistent
 * FIXME: Severe code duplication with nhgcd_update_uv */

static void
nhgcd_update_u(struct nhgcd_row *r, mp_srcptr qp, mp_size_t qsize,
	       /* Limbs allocated for the new u, for sanity
		  checking */
	       mp_size_t alloc)
{
  ASSERT(r[0].uvsize[0] < alloc);
  ASSERT(r[0].uvsize[1] < alloc);
  ASSERT(r[1].uvsize[0] < alloc);
  ASSERT(r[1].uvsize[1] < alloc);
  
  /* Compute u3  = u1 + q u2 */
  
  if (qsize == 0)
    /* Represents a unit quotient */
    {	
      mp_size_t u1size = r[0].uvsize[0];
      mp_size_t u2size = r[1].uvsize[0];
      mp_srcptr u1p = r[0].uvp[0];
      mp_srcptr u2p = r[1].uvp[0];
      mp_ptr u3p = r[2].uvp[0];

      if (u1size == 0)
	{
	  /* u3 = u2 */
	  MPN_COPY(u3p, u2p, u2size);
	  r[2].uvsize[0] = u2size; 
	}
      else
	{	      
	  if (u2size > 0)
	    {
	      /* Compute u3 = u1 + u2, then swap */
	      mp_limb_t cy;
	      mp_size_t u3size;
		
	      ASSERT(u2size >= u1size);
		
	      cy = mpn_add(u3p, u2p, u2size, u1p, u1size);
	      u3size = u2size;
	      if (cy)
		{
		  ASSERT(u3size < alloc);
		  u3p[u3size++] = cy;
		}
	      r[2].uvsize[0] = u3size; 
	    }
	  else
	    {
	      /* u2 = 0, u3 = u1 */
	      MPN_COPY(u3p, u1p, u1size);
	      r[2].uvsize[0] = u1size; 
	    }
	}
    }
  else if (qsize == 1)
    {
      mp_limb_t q = qp[0];

      mp_size_t u1size = r[0].uvsize[0];
      mp_size_t u2size = r[1].uvsize[0];
      mp_srcptr u1p = r[0].uvp[0];
      mp_srcptr u2p = r[1].uvp[0];
      mp_ptr u3p = r[2].uvp[0];
	  
      mp_size_t u3size;

      mp_limb_t cy;
      
      if (u1size == 0)
	{
	  /* Compute u1 = q u2 */
	  u3size = u2size;
	  cy = mpn_mul_1(u3p, u2p, u2size, q);
	  if (cy)
	    {
	      ASSERT(u3size < alloc);
	      u3p[u3size++] = cy;
	    }
	}
      else if (u2size > 0)
	{
	  ASSERT(u2size >= u1size);
	  u3size = u2size;
	  if (u3size > u1size)
	    {
	      /* Do high limbs first */
	      cy = mpn_mul_1(u3p + u1size,
			     u2p + u1size, u3size - u1size, q);
	      if (cy)
		{
		  ASSERT(u3size < alloc);
		  u3p[u3size++] = cy;
		}
	    }
	  /* FIXME: Use mpn_addmul_1? */
	  cy = mpn_mul_1(u3p, u2p, u1size, q);
	  cy += mpn_add_n(u3p, u3p, u1p, u1size);

	  if (cy && u3size > u1size)
	    cy = mpn_add_1(u3p + u1size,
			   u3p + u1size, u3size - u1size, cy);
	  if (cy)
	    {
	      ASSERT(u3size < alloc);
	      u3p[u3size++] = cy;
	    }
	}
      else
	{
	  /* u2 == 0, u3 = u1 */
	  MPN_COPY(u3p, u1p, u1size);
	  u3size = u1size;
	}

      r[2].uvsize[0] = u3size;
    }
  else
    {
      mp_size_t u1size = r[0].uvsize[0];
      mp_size_t u2size = r[1].uvsize[0];
      mp_srcptr u1p = r[0].uvp[0];
      mp_srcptr u2p = r[1].uvp[0];
      mp_ptr u3p = r[2].uvp[0];

      mp_size_t u3size;

      if (u1size == 0)
	{
	  u3size = qsize + u2size;
	  ASSERT(u3size < alloc);

	  u3size -= mpn_mul_friendly(u3p, u2p, u2size, qp, qsize);
	}
      else if (u2size > 0)
	{
	  mp_limb_t cy;
	  
	  u3size = qsize + u2size;
	  ASSERT(u3size < alloc);

	  u3size -= mpn_mul_friendly(u3p, u2p, u2size, qp, qsize);

	  ASSERT(u3size >= u1size);
	  cy = mpn_add(u3p, u3p, u3size, u1p, u1size);
	  if (cy)
	    {
	      ASSERT(u3size < alloc);
	      u3p[u3size++] = cy;
	    }
	}
      else
	{
	  /* u2 == 0, u3 = u1 */
	  MPN_COPY(u3p, u1p, u1size);
	  u3size = u1size;
	}
      
      r[2].uvsize[0] = u3size;
    }
  
  ASSERT(r[1].uvsize[0] <= r[2].uvsize[0]);
}



/* Compute r2 = r0 - q r1 */
static void
nhgcd_update_r(struct nhgcd_row *r, mp_srcptr qp, mp_size_t qsize)
{
  mp_srcptr r0p = r[0].rp;
  mp_srcptr r1p = r[1].rp;
  mp_ptr r2p = r[2].rp;
  mp_size_t r0size = r[0].rsize;
  mp_size_t r1size = r[1].rsize;
  
  mp_size_t r2size = r0size;

  ASSERT(MPN_LESS_P(r1p, r1size, r0p, r0size));
  
  if (qsize == 0)
    {
      ASSERT_NOCARRY(mpn_sub(r2p,
			     r0p, r0size,
			     r1p, r1size));

    }
  else if (qsize == 1)
    {
      mp_limb_t cy;
      MPN_COPY(r2p, r0p, r2size);

      cy = mpn_submul_1(r2p, r1p, r1size, qp[0]);
      if (cy)
	{
	  ASSERT(r1size < r2size);
	  ASSERT_NOCARRY(mpn_sub_1(r2p + r1size,
				   r2p + r1size, r2size - r1size,
				   cy));
	}
    }
  else
    {
      mp_size_t size = MPN_MUL_SIZE(r1size, qsize);
      ASSERT(size);

      size -= mpn_mul_friendly(r2p, r1p, r1size, qp, qsize);
      ASSERT(size <= r0size);
      ASSERT_NOCARRY(mpn_sub(r2p, r0p, r0size, r2p, size));
    }

  MPN_NORMALIZE(r[2].rp, r2size);
  r[2].rsize = r2size;
}

/* Computes P = R * S. No overlap allowed.

   Temporary space is needed for two numbers smaller than the
   resulting matrix elements, i.e. bounded by 2*L <= N. */
static void
nhgcd_mul(struct nhgcd_row *P, mp_size_t alloc,
	  const struct nhgcd_row *R, const struct nhgcd_row *S,
	  mp_ptr tp, mp_size_t talloc)
{
  unsigned i;
  unsigned j;
  
  for (i = 0; i<2; i++)
    for (j = 0; j<2; j++)
      {
	/* Set P[i, j] = R[i, 0] S[0, j] + R[i,1] S[1, j] */
	mp_size_t t1size = MPN_MUL_SIZE(R[i].uvsize[0],
					S[0].uvsize[j]);
	mp_size_t t2size = MPN_MUL_SIZE(R[i].uvsize[1],
					S[1].uvsize[j]);
	ASSERT(t1size <= alloc);
	ASSERT(t2size <= alloc);
	ASSERT(t2size <= talloc);
	
	if (t1size)
	  t1size -= mpn_mul_friendly(P[i].uvp[j],
				     R[i].uvp[0], R[i].uvsize[0],
				     S[0].uvp[j], S[0].uvsize[j]);
	
	if (t2size)
	  t2size -= mpn_mul_friendly(tp,
				     R[i].uvp[1], R[i].uvsize[1],
				     S[1].uvp[j], S[1].uvsize[j]);

	ASSERT(t1size < alloc);
	ASSERT(t2size < alloc);
	
	P[i].uvsize[j] = mpn_add_friendly(P[i].uvp[j],
					  P[i].uvp[j], t1size,
					  tp, t2size);

	ASSERT(P[i].uvsize[j] < alloc);
      }
}

/* Computes Y = R * X. No overlap allowed.

   Temporary space is needed for two numbers smaller than the
   resulting matrix elements, i.e. bounded by 2*L <= N.

   FIXME: Severe code duplication. */

static void
nhgcd_mul_vector(struct nhgcd_row *P,
		 mp_size_t alloc,
		 const struct nhgcd_row *R, const struct nhgcd_row *S,
		 mp_ptr tp, mp_size_t talloc)
{
  unsigned i;
  
  for (i = 0; i<2; i++)
    {
      /* Set P[i, j] = R[i, 0] S[0, j] + R[i,1] S[1, j] */
      mp_size_t t1size = MPN_MUL_SIZE(R[i].uvsize[0],
				      S[0].uvsize[0]);
      mp_size_t t2size = MPN_MUL_SIZE(R[i].uvsize[1],
				      S[1].uvsize[0]);
      ASSERT(t1size <= alloc);
      ASSERT(t2size <= alloc);
      ASSERT(t2size <= talloc);
	
      if (t1size)
	t1size -= mpn_mul_friendly(P[i].uvp[0],
				   R[i].uvp[0], R[i].uvsize[0],
				   S[0].uvp[0], S[0].uvsize[0]);
	
      if (t2size)
	t2size -= mpn_mul_friendly(tp,
				   R[i].uvp[1], R[i].uvsize[1],
				   S[1].uvp[0], S[1].uvsize[0]);

      ASSERT(t1size < alloc);
      ASSERT(t2size < alloc);
	
      P[i].uvsize[0] = mpn_add_friendly(P[i].uvp[0],
					  P[i].uvp[0], t1size,
					  tp, t2size);

      ASSERT(P[i].uvsize[0] < alloc);
    }
}

static int
lehmer_step(struct nhgcd2 *hgcd,
	    mp_srcptr ap, mp_size_t asize,
	    mp_srcptr bp, mp_size_t bsize,
	    struct nhgcd_quotient_stack *quotients)
{
  mp_limb_t ah[2];
  mp_limb_t bh[2];
  
  if (bsize < 2)
    return 0;
  
  /* The case asize == 2 is needed to take care of values that are
     between one and two *full* limbs in size. */
  if (asize == 2 || (ap[asize-1] & GMP_LIMB_HIGHBIT))
    {
      if (bsize < asize)
	return 0;
	  
      ah[0] = ap[asize - 2];
      ah[1] = ap[asize - 1];

      ASSERT(asize == bsize);
      bh[0] = bp[asize - 2];
      bh[1] = bp[asize - 1];
    }
  else
    {
      unsigned shift;
      if (bsize + 1 < asize)
	return 0;
      
      /* We want two *full* limbs */
      ASSERT(asize > 2);

      count_leading_zeros(shift, ap[asize-1]);
      if (bsize == asize)
	bh[1] = SHIFT_LEFT(shift, bp + asize - 2);
      else
	{
	  ASSERT(asize == bsize + 1);
	  bh[1] = bp[asize - 2] >> (BITS_PER_MP_LIMB - shift);
	}
      if (!bh[1])
	return 0;
      
      bh[0] = SHIFT_LEFT(shift, bp + asize - 3);
	  
      ah[0] = SHIFT_LEFT(shift, ap + asize - 3);
      ah[1] = SHIFT_LEFT(shift, ap + asize - 2);
    }

  return nhgcd2(hgcd, ah, bh, quotients);
}

/* Computes R = 2^k H + u A' - v B', which must be non-negative.
   Temporary space needed is k + max(usize, vsize)

   May be called with hp == rp + k. */

static mp_size_t
nhgcd_fix(mp_size_t k,
	  mp_ptr rp, mp_size_t ralloc,
	  mp_ptr hp, mp_size_t hsize,
	  int sign,
	  mp_srcptr up, mp_size_t usize, mp_srcptr ap,
	  mp_srcptr vp, mp_size_t vsize, mp_srcptr bp,
	  mp_ptr tp, mp_size_t talloc)
{
  mp_size_t tsize;
  mp_limb_t cy;
  mp_size_t asize = k;
  mp_size_t bsize = k;
  mp_size_t rsize;

  if (sign < 0)
    {
      MPN_SRCPTR_SWAP(up, usize, vp, vsize);
      MP_SRCPTR_SWAP(ap, bp);
    }
  
  MPN_NORMALIZE(ap, asize);
  MPN_NORMALIZE(bp, bsize);

  tsize = MPN_MUL_SIZE(asize, usize);

  ASSERT(k + hsize <= ralloc);
  ASSERT(tsize <= talloc);
  ASSERT(tsize <= ralloc);

  /* NOTE: We must have usize, vsize <= m, but we don't have m here so
     we can't ASSERT that. */
  
  if (tsize)
    {
      tsize -= mpn_mul_friendly(tp,
				ap, asize,
				up, usize);
#if 0      
      trace("jhgcd_fix: k = %d, tsize = %d, talloc = %d,\n"
	    "  usize = %d, vsize = %d\n",
	    k, tsize, talloc, usize, vsize);
#endif
      if (tsize <= k)
	{
	  MPN_COPY(rp, tp, tsize);
	  MPN_COPY(rp + k, hp, hsize);
	  rsize = k + hsize;

	  if (tsize < k)
	    MPN_ZERO(rp + tsize, k - tsize);
	}
      else
	{
	  MPN_COPY(rp, tp, k);
	  rsize = k + mpn_add_friendly(rp + k,
				       hp, hsize,
				       tp + k, tsize - k);
	}
    }
  else
    {
      MPN_ZERO(rp, k);
      MPN_COPY(rp + k, hp, hsize);
      rsize = k + hsize;
    }

  ASSERT(rsize <= ralloc);
  
  /* A -= v * B' */
  tsize = MPN_MUL_SIZE(bsize, vsize);

  ASSERT(tsize <= talloc);
  ASSERT(tsize <= ralloc);

  if (tsize)
    {
      tsize -= mpn_mul_friendly(tp,
				bp, bsize,
				vp, vsize);

      ASSERT(tsize <= rsize);
      
      cy = mpn_sub(rp, rp, rsize,
		   tp, tsize);
      ASSERT(!cy);

      MPN_NORMALIZE(rp, rsize);
    }
  return rsize;
}

/* Compute r0 = r2 + q r1, and the corresponding uv */
static void
nhgcd_backup(struct nhgcd_row *r, mp_srcptr qp, mp_size_t qsize)
{
  mp_ptr r0p = r[0].rp;
  mp_srcptr r1p = r[1].rp;
  mp_srcptr r2p = r[2].rp;
  mp_size_t r0size;
  mp_size_t r1size = r[1].rsize;
  mp_size_t r2size = r[2].rsize;

  mp_ptr u0p = r[0].uvp[0];
  mp_ptr v0p = r[0].uvp[1];
  mp_srcptr u1p = r[1].uvp[0];
  mp_srcptr v1p = r[1].uvp[1];
  mp_srcptr u2p = r[2].uvp[0];
  mp_srcptr v2p = r[2].uvp[1];
  mp_size_t u1size = r[1].uvsize[0];
  mp_size_t v1size = r[1].uvsize[1];
  mp_size_t u2size = r[2].uvsize[0];
  mp_size_t v2size = r[2].uvsize[1];
  
  mp_size_t u0size = u2size;
  mp_size_t v0size = v2size;

  /* Usually, we should have r1 > r2, but that will not be the case
     when we backup after having used an incorrect quotient. */

  /* u1 may be zero, but v1, u2 and v2 must be non-zero */
  ASSERT(v1size > 0);
  ASSERT(u2size > 0);
  ASSERT(v2size > 0);

  if (u1size == 0)
    MPN_COPY(u0p, u2p, u0size);
  
  if (qsize == 0)
    {
      /* r0 = r2 + r1 */ 
      r0size = mpn_add_friendly(r0p, r1p, r1size, r2p, r2size);

      /* (u0,v0) = (u2,v2) - (u1, v1) */

      if (u1size > 0)
	ASSERT_NOCARRY(mpn_sub(u0p,
			       u2p, u2size,
			       u1p, u1size));

      ASSERT_NOCARRY(mpn_sub(v0p,
			     v2p, v2size,
			     v1p, v1size));
    }
  else if (qsize == 1)
    {
      /* r0 = r2 + q r1 */
      mp_limb_t cy;
      mp_size_t size;
      
      /* FIXME: Use mpn_addmul_1? */      

      r0size = r1size;
      cy = mpn_mul_1(r0p, r1p, r1size, qp[0]);
      if (cy)
	r0p[r0size++] = cy;

      r0size = mpn_add_friendly(r0p, r0p, r0size, r2p, r2size);
      
      /* (u0,v0) = (u2,v2) - q (u1, v1) */

      if (u1size > 0)
	{
	  size = u1size;
	  cy = mpn_mul_1(u0p, u1p, u1size, qp[0]);
	  if (cy)
	    {
	      ASSERT(size < u2size);
	      u0p[size++] = cy;
	    }

	  ASSERT_NOCARRY(mpn_sub(u0p, u2p, u2size, u0p, size));
	}
      size = v1size;

      cy = mpn_mul_1(v0p, v1p, v1size, qp[0]);
      if (cy)
	{
	  ASSERT(size < v2size);
	  v0p[size++] = cy;
	}
      
      ASSERT_NOCARRY(mpn_sub(v0p, v2p, v2size, v0p, size));
    }
  else
    {
      /* r0 = r2 + q r1 */
      mp_size_t size = MPN_MUL_SIZE(r1size, qsize);
      ASSERT(size);

      size -= mpn_mul_friendly(r0p, r1p, r1size, qp, qsize);

      r0size = mpn_add_friendly(r0p, r0p, size, r2p, r2size);

      /* (u0,v0) = (u2,v2) - q (u1, v1) */

      if (u1size > 0)
	{
	  size = MPN_MUL_SIZE(u1size, qsize);
	  ASSERT(size);

	  ASSERT(size <= u2size + 1);
	  size -= mpn_mul_friendly(u0p, u1p, u1size, qp, qsize);
	  ASSERT(size <= u2size);

	  ASSERT_NOCARRY(mpn_sub(u0p, u2p, u2size, u0p, size));
	}

      size = MPN_MUL_SIZE(v1size, qsize);
      ASSERT(size);
      
      ASSERT(size <= v2size + 1);
      size -= mpn_mul_friendly(v0p, v1p, v1size, qp, qsize);
      ASSERT(size <= v2size);
      
      ASSERT_NOCARRY(mpn_sub(v0p, v2p, v2size, v0p, size));
    }

  MPN_NORMALIZE(u0p, u0size);
  MPN_NORMALIZE(v0p, v0size);

  r[0].rsize = r0size;
  r[0].uvsize[0] = u0size;
  r[0].uvsize[1] = v0size;
}

static void
euclid_step(struct nhgcd_row *r,
	    struct nhgcd_quotient_stack *quotients,
	    /* For sanity checking only */
	    mp_size_t alloc)
{
  mp_size_t qsize;
  mp_ptr qp;

  qsize = nhgcd_quotient_stack_push_quotient(quotients, &qp,
					     r[2].rp, &r[2].rsize,
					     r[0].rp, r[0].rsize,
					     r[1].rp, r[1].rsize);
  
  nhgcd_update_uv(r, qp, qsize, alloc);
}

/* Checks if a - b < c. Overwrites c. */
int
nhgcd_diff_smaller_p(mp_srcptr ap, mp_size_t asize,
		     mp_srcptr bp, mp_size_t bsize,
		     mp_ptr cp, mp_size_t csize)
{
  mp_limb_t ch;
  mp_limb_t dh;
  unsigned i;

#if 0
  trace("nhgcd_diff_smaller_p: asize = %d, bsize = %d, csize = %d\n",
	asize, bsize, csize);
  trace("  a = %Nd\n", ap, asize);
  trace("  b = %Nd\n", bp, csize);
  trace("  c = %Nd\n", cp, csize);
#endif

  ASSERT(MPN_LEQ_P(bp, bsize, ap, asize));

  if (csize == 0)
    return 0;

  if (asize < csize)
    return 1;
  
  if (asize > bsize)
    {
      /* #(a - b) >= asize - 1 */
      if (asize > csize + 1)
	return 0;
      else
	{
	  ASSERT(bsize <= csize);
	  ch = mpn_add(cp, cp, csize, bp, bsize);
	  if (ch)
	    {
	      if (asize == csize)
		return 0;
	      ASSERT(asize == csize + 1);
	      if (ap[asize - 1] < ch)
		return 1;
	      else if (ap[asize - 1] > ch)
		return 0;

	      /* The high limb of a is canceled by ch. */
	      asize--;
	    }

	  return MPN_LESS_P(ap, asize, cp, csize);
	}
    }

  /* a and b are of the same size. Equal high limbs cancel out, so
     ignore them */
  while (asize && ap[asize - 1] == bp[asize - 1])
    asize--;

  if (asize < csize)
    return 1;

  /* Now asize = bsize >= csize */
  if (asize == csize)
    {
      ch = mpn_add_n(cp, cp, bp, csize);
      if (ch)
	return 1;
      
      return MPN_LESS_P(ap, csize, cp, csize);
    }

  /* asize == bsize > csize. */
  if (csize == 0)
    return 0;
  
  /* Subtract one from c, transforming the check to a - b <= c */
  ASSERT_NOCARRY(mpn_sub_1(cp, cp, csize, 1));
  MPN_NORMALIZE(cp, csize);

  if (csize == 0)
    return 0;
  
  /* Write a = ah W^k + al, b = bh W^k + bl, where c, al, bl < W^k.
     Then a - b <= c is equivalent to

     (ah - bh) W^k <= c + bl - al

     We always have (ah - bh) >= 0, and

     -W^k < c + bl - ah < 2 W^k
  */

  /* We know that ah - bh >= 1. Do we have ah - bh >= 2? */
  dh = ap[asize - 1] - bp[asize - 1];
  if (dh >= 2)
    return 0;

  /* Compute c += bl - al */
  ch = mpn_add_n(cp, cp, bp, csize);
  if (ch == 0)
    /* c + bl < W^k */
    return 0;
       
  ch -= mpn_sub_n(cp, cp, ap, csize);

  if (ch == 0)
    return 0;

  /* W^k <= c + bl - al < 2 W^k. Then (ah - bh) W^k is smaller iff ah
     = bh + 1, which happens iff ah ends with 0...0, and bh with
     f...f. */

  for (i = csize; i < asize - 1; i++)
    if (ap[i] != 0 || bp[i] != MP_LIMB_T_MAX)
      return 0;

  return 1;
}


#if WANT_CHECKS
# define nhgcd_diff_smaller_p nhgcd_diff_smaller_p_wrapper
static int
nhgcd_diff_smaller_p_wrapper(mp_srcptr ap, mp_size_t asize,
			     mp_srcptr bp, mp_size_t bsize,
			     mp_ptr cp, mp_size_t csize);
#endif


/* Scratch space needed to compute |u2| + |u1| or |v2| + |v1|, where
   all terms are at most floor(asize/2) limbs. */
#define NHGCD_JEBELEAN_ITCH(asize) ((asize)/2 + 1)

static int
nhgcd_jebelean(const struct nhgcd *hgcd, mp_size_t M,
	       mp_ptr tp, mp_size_t talloc)
{
  mp_size_t L;
  mp_size_t tsize;
  
  ASSERT(hgcd->row[0].rsize > M);
  ASSERT(hgcd->row[1].rsize > M);
  ASSERT(hgcd->row[2].rsize > M);
  ASSERT(hgcd->row[3].rsize <= M);

  /* The bound is really floor(N/2), which is <= M = ceil(M/2) */
  L = hgcd->row[3].uvsize[1];
  ASSERT(L <= M);
  
  ASSERT(hgcd->row[0].uvsize[0] <= L);
  ASSERT(hgcd->row[0].uvsize[1] <= L);
  ASSERT(hgcd->row[1].uvsize[0] <= L);
  ASSERT(hgcd->row[1].uvsize[1] <= L);
  ASSERT(hgcd->row[2].uvsize[0] <= L);
  ASSERT(hgcd->row[2].uvsize[1] <= L);
  ASSERT(hgcd->row[3].uvsize[0] <= L);
  
  ASSERT(L < talloc);

  if (hgcd->sign >= 0)
    {
      /* Check if r1 - r2 >= u2 - u1 */
      mp_limb_t cy;
      
      tsize = hgcd->row[2].uvsize[0];
      ASSERT(tsize < talloc);
      cy = mpn_add(tp,
		   hgcd->row[2].uvp[0], hgcd->row[2].uvsize[0],
		   hgcd->row[1].uvp[0], hgcd->row[1].uvsize[0]); 
      if (cy)
	tp[tsize++] = cy;
    }
  else
    {
      /* Check if r1 - r2 >= v2 - v1 */
      mp_limb_t cy;
      
      tsize = hgcd->row[2].uvsize[1];
      ASSERT(tsize < talloc);
      cy = mpn_add(tp,
		   hgcd->row[2].uvp[1], hgcd->row[2].uvsize[1],
		   hgcd->row[1].uvp[1], hgcd->row[1].uvsize[1]); 
      if (cy)
	tp[tsize++] = cy;      
    }

  if (nhgcd_diff_smaller_p(hgcd->row[1].rp, hgcd->row[1].rsize,
			   hgcd->row[2].rp, hgcd->row[2].rsize, tp, tsize))
    return 2;

#if 0
  /* r1 - r2 >= t  <=>  r2 >= t and r2 - t >= r1 */
  if (MPN_LESS_P(hgcd->row[2].rp, hgcd->row[2].rsize, tp, tsize))
    return 2;

  ASSERT_NOCARRY(mpn_sub(tp, hgcd->row[2].rp, hgcd->row[2].rsize, tp, tsize));
  tsize = hgcd->row[2].rsize;
  MPN_NORMALIZE(tp, tsize);

  if (MPN_LESS_P(tp, tsize, hgcd->row[1].rp, hgcd->row[1].rsize))
    return 2;
#endif
  
  /* Ok, r2 is correct */

  if (hgcd->sign >= 0)
    {
      mp_limb_t cy;

      /* Check r3 >= max(-u3, -v3) = u3 */
      if (MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize,
		     hgcd->row[3].uvp[0], hgcd->row[3].uvsize[0]))
	return 3;
      
      /* Check r3 - r2 >= v3 - v2 */
      tsize = hgcd->row[3].uvsize[1];
      ASSERT(tsize < talloc);
      cy = mpn_add(tp,
		   hgcd->row[3].uvp[1], hgcd->row[3].uvsize[1],
		   hgcd->row[2].uvp[1], hgcd->row[2].uvsize[1]); 
      if (cy)
	tp[tsize++] = cy;      
    }
  else
    {
      mp_limb_t cy;

      /* Check r3 >= max(-u3, -v3) = v3 */
      if (MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize,
		     hgcd->row[3].uvp[1], hgcd->row[3].uvsize[1]))
	return 3;

      /* Check r3 - r2 >= u3 - u2 */      
      
      tsize = hgcd->row[3].uvsize[0];
      ASSERT(tsize < talloc);
      cy = mpn_add(tp,
		   hgcd->row[3].uvp[0], hgcd->row[3].uvsize[0],
		   hgcd->row[2].uvp[0], hgcd->row[2].uvsize[0]); 
      if (cy)
	tp[tsize++] = cy;
    }

  if (nhgcd_diff_smaller_p(hgcd->row[2].rp, hgcd->row[2].rsize,
			   hgcd->row[3].rp, hgcd->row[3].rsize, tp, tsize))
    return 3;
#if 0
  /* r3 - r2 >= t  <=> r3 >= t and r3 - t >= r2 */
  if (MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize, tp, tsize))
    return 3;

  ASSERT_NOCARRY(mpn_sub(tp, hgcd->row[3].rp, hgcd->row[3].rsize, tp, tsize));
  tsize = hgcd->row[3].rsize;
  MPN_NORMALIZE(tp, tsize);

  if (MPN_LESS_P(tp, tsize, hgcd->row[2].rp, hgcd->row[2].rsize))
    return 3;
#endif
  
  /* Ok, r3 is correct */
  return 4;
}


/* More helper functions for hgcd */

static void
nhgcd_start(struct nhgcd *hgcd,
	    mp_srcptr ap, mp_size_t asize,
	    mp_srcptr bp, mp_size_t bsize)
{
  MPN_COPY(hgcd->row[0].rp, ap, asize);
  hgcd->row[0].rsize = asize;

  MPN_COPY(hgcd->row[1].rp, bp, bsize);
  hgcd->row[1].rsize = bsize;

  hgcd->sign = 0;
  hgcd->row[0].uvsize[0] = 1; hgcd->row[0].uvp[0][0] = 1;
  hgcd->row[0].uvsize[1] = 0;
  hgcd->row[1].uvsize[0] = 0;
  hgcd->row[1].uvsize[1] = 1; hgcd->row[1].uvp[1][0] = 1;
}

/* Only the first row has v = 0, a = 1 * a + 0 * b */
#define START_ROW_P(r) ((r).uvsize[1] == 0)

/* Called when r2 has been computed, and it is too small. Top element
   on the stack is r0/r1 */
static int
nhgcd_small(struct nhgcd *hgcd, mp_size_t M,
	    struct nhgcd_quotient_stack *quotients,
	    mp_ptr tp, mp_size_t talloc)
{
  mp_srcptr qp;
  mp_size_t qsize;

  if (START_ROW_P(hgcd->row[0]))
    {
      nhgcd_quotient_stack_drop(quotients);
      return 0;
    }

  NHGCD_SWAP4_RIGHT(hgcd->row);
  qsize = nhgcd_quotient_stack_get_1(quotients, &qp);
      
  nhgcd_backup(hgcd->row, qp, qsize);
  hgcd->sign = ~hgcd->sign;

#if WANT_CHECKS
  nhgcd_quotient_stack_rotate(quotients, 0);
#endif

  return nhgcd_jebelean(hgcd, M, tp, talloc);
}

/* Called with r0 and r1 of approximately the right size */
static int
nhgcd_final(struct nhgcd *hgcd, mp_size_t M,
	    struct nhgcd_quotient_stack *quotients,
	    mp_ptr tp, mp_size_t talloc)
{
  ASSERT(hgcd->row[0].rsize > M);
  ASSERT(hgcd->row[1].rsize == M + 1);

  /* Use euclid steps to get to the desired size. */
  /* FIXME: We could perhaps perform one lehmer step */
  euclid_step(hgcd->row, quotients, hgcd->alloc);

  if (hgcd->row[2].rsize <= M)
    return nhgcd_small(hgcd, M, quotients,
		       tp, talloc);
  
  for (;;)
    {
#if 0
      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 3);
#endif 
      euclid_step(hgcd->row + 1,
		  quotients,
		  hgcd->alloc);

      if (hgcd->row[3].rsize <= M)
	{
#if WANT_CHECKS
	  nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
	  return nhgcd_jebelean(hgcd, M, tp, talloc);
	}
      hgcd->sign = ~hgcd->sign;
      
      NHGCD_SWAP4_LEFT(hgcd->row);
    }
}

/* Performs one euclid step on r0, r1. Returns >= if hgcd should be
   terminated, -1 if we should go on */
static int
nhgcd_case0(struct nhgcd *hgcd, mp_size_t M,
	    struct nhgcd_quotient_stack *quotients,
	    mp_ptr tp, mp_size_t talloc)
{
  euclid_step(hgcd->row, quotients, hgcd->alloc);
  
  if (hgcd->row[2].rsize <= M)
    return nhgcd_small(hgcd, M, quotients,
		       tp, talloc);
  else
    {
      /* Keep this remainder */
      hgcd->sign = ~hgcd->sign;
      
      NHGCD_SWAP4_LEFT(hgcd->row);
      return -1;
    }
}

mp_size_t
nhgcd_lehmer_itch(mp_size_t asize, mp_size_t bsize UNUSED)
{
  /* Scratch space is needed for calling nhgcd_lehmer. */
  return NHGCD_JEBELEAN_ITCH(asize);
}

int
nhgcd_lehmer(struct nhgcd *hgcd,
	     mp_srcptr ap, mp_size_t asize,
	     mp_srcptr bp, mp_size_t bsize,
	     struct nhgcd_quotient_stack *quotients,
	     mp_ptr tp, mp_size_t talloc)
{
  /* FIXME: Severe code duplication with nhgcd(). When this code works,
     extract common code to new helper functions. */
  
  mp_size_t N = asize;
  mp_size_t M = CHOOSE_M(N);

  ASSERT(M);

  if (bsize <= M)
    return 0;

  ASSERT(asize >= 2);
  
  /* Initialize, we keep r0 and r1 as the reduced numbers (so far). */
  nhgcd_start(hgcd, ap, asize, bp, bsize);

  while (hgcd->row[1].rsize > M + 1)
    {
      struct nhgcd2 R;

      /* Max size after reduction, plus one */
      mp_size_t ralloc = hgcd->row[1].rsize + 1;
      
      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);
      
      switch(lehmer_step(&R,
			 hgcd->row[0].rp, hgcd->row[0].rsize,
			 hgcd->row[1].rp, hgcd->row[1].rsize,
			 quotients))
	{
	default:
	  abort();
	case 0:
	  {
	    /* The first remainder was small. Then there's a good chance
	       that the remainder A % B is also small. */
	    int res = nhgcd_case0(hgcd, M, quotients, tp, talloc);

	    if (res > 0)
	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	    if (res >= 0)
	      return res;

	    nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);
	    break;
	  }
	case 2:
	  /* r0 and r1 are correct, r2 may be too large. */
	  /* Compute r1 and r2, and check what happened */

	  hgcd->row[2].rsize
	    = nhgcd2_fix(hgcd->row[2].rp, ralloc,
			 ~R.sign,
			 R.row[1].u, hgcd->row[0].rp, hgcd->row[0].rsize,
			 R.row[1].v, hgcd->row[1].rp, hgcd->row[1].rsize);
	    
	  hgcd->row[3].rsize
	    = nhgcd2_fix(hgcd->row[3].rp, ralloc,
			 R.sign,
			 R.row[2].u, hgcd->row[0].rp, hgcd->row[0].rsize,
			 R.row[2].v, hgcd->row[1].rp, hgcd->row[1].rsize);

	  ASSERT(hgcd->row[2].rsize > M);

	  if (MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize,
			 hgcd->row[2].rp, hgcd->row[2].rsize))
	    /* r2 was correct. */
	    goto correct_r2;
	  else
	    {
	      /* Computes the uv matrix for the (incorrect) values r1, r2.
		 The elements must be smaller than the correct ones, since
		 they correspond to a too small q. */
	      nhgcd2_mul(hgcd->row + 2, hgcd->alloc,
			 R.row + 1, hgcd->row);

	      /* Discard r3, and the corresponding quotient */
	      nhgcd_quotient_stack_drop(quotients);

	      /* Get the incorrect r0/r1 quotient used to compute r2 */
	      {
		mp_size_t qsize;
		mp_srcptr qp;

		qsize = nhgcd_quotient_stack_pop(quotients, &qp);
		nhgcd_backup(hgcd->row + 1, qp, qsize);
		hgcd->sign ^= ~R.sign;
	      }
	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 1, 3);
	      
	      /* Compute the correct r2. FIXME: We could do this
		 slightly more efficient by dividing r3/r2, and adding
		 this quotient to the earlier incorrect one. */
	      euclid_step(hgcd->row+1, quotients, hgcd->alloc);

	      /* FIXME: Bound on new r2? */
	      if (hgcd->row[3].rsize <= M)
		{
		  mp_size_t qsize;
		  mp_srcptr qp;
	      
		  if (START_ROW_P(hgcd->row[1]))
		    {
		      nhgcd_quotient_stack_drop(quotients);
		      return 0;
		    }
		  qsize = nhgcd_quotient_stack_get_1(quotients, &qp);
      
		  nhgcd_backup(hgcd->row, qp, qsize);

#if WANT_CHECKS
		  nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
		  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
		  return nhgcd_jebelean(hgcd, M, tp, talloc);
		}
	      NHGCD_SWAP4_2(hgcd->row);

	      continue;
	    }
	case 3:
	  /* Now r0, r1 and r2 are correct, while r3 may be too small
	     or too large. */
	  
	  hgcd->row[2].rsize
	    = nhgcd2_fix(hgcd->row[2].rp, ralloc,
			 ~R.sign,
			 R.row[1].u, hgcd->row[0].rp, hgcd->row[0].rsize,
			 R.row[1].v, hgcd->row[1].rp, hgcd->row[1].rsize);
	    
	  hgcd->row[3].rsize
	    = nhgcd2_fix(hgcd->row[3].rp, ralloc,
			 R.sign,
			 R.row[2].u, hgcd->row[0].rp, hgcd->row[0].rsize,
			 R.row[2].v, hgcd->row[1].rp, hgcd->row[1].rsize);

	correct_r2:
	  /* Discard r3, and the corresponding quotient */
	  nhgcd_quotient_stack_drop(quotients);
	  
	  ASSERT(hgcd->row[2].rsize > M);
	  ASSERT(hgcd->row[3].rsize > M);
	    
	  ASSERT(MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize,
			    hgcd->row[2].rp, hgcd->row[2].rsize));

	  nhgcd2_mul(hgcd->row + 2, hgcd->alloc,
		     R.row + 1, hgcd->row);
	  hgcd->sign ^= ~R.sign;
	  
	  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 2, 4);
	  
	  NHGCD_SWAP4_2(hgcd->row);

	  /* FIXME: Additional euclid step here? */
	  break;
	case 4:
	  /* All of r0, r1, r3 and r3 are correct.
	     Compute r2 and r3 */
	    
	  hgcd->row[2].rsize
	    = nhgcd2_fix(hgcd->row[2].rp, ralloc,
			 R.sign,
			 R.row[2].u, hgcd->row[0].rp, hgcd->row[0].rsize,
			 R.row[2].v, hgcd->row[1].rp, hgcd->row[1].rsize);
	    
	  hgcd->row[3].rsize
	    = nhgcd2_fix(hgcd->row[3].rp, ralloc,
			 ~R.sign,
			 R.row[3].u, hgcd->row[0].rp, hgcd->row[0].rsize,
			 R.row[3].v, hgcd->row[1].rp, hgcd->row[1].rsize);
	    
	  ASSERT(hgcd->row[2].rsize > M);

	  /* FIXME: Bound on r3? */

	  nhgcd2_mul(hgcd->row + 2, hgcd->alloc,
		     R.row + 2, hgcd->row);
	  hgcd->sign ^= R.sign;
	  
	  if (hgcd->row[3].rsize <= M)
	    {
	      /* Backup two steps */
	      mp_size_t qsize;
	      mp_srcptr qp;

	      ASSERT(!START_ROW_P(hgcd->row[2]));
	      
	      qsize = nhgcd_quotient_stack_get_0(quotients, &qp);
	      nhgcd_backup(hgcd->row+1, qp, qsize);

	      ASSERT(!START_ROW_P(hgcd->row[1]));

	      qsize = nhgcd_quotient_stack_get_1(quotients, &qp);
	      nhgcd_backup(hgcd->row, qp, qsize);
		
#if WANT_CHECKS
	      nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	      return nhgcd_jebelean(hgcd, M, tp, talloc);
	    }
	    
	  NHGCD_SWAP4_2(hgcd->row);
	}
    }

#if WANT_CHECKS
  nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);

  return nhgcd_final(hgcd, M, quotients, tp, talloc);
}

mp_size_t
nhgcd_itch(mp_size_t asize, mp_size_t bsize UNUSED)
{
  /* Scratch space is needed for calling nhgcd. We need space for the
     results of all recursive calls. In addition, we need to call
     nhgcd_lehmer, nhgcd_jebelean, nhgcd_fix and nhgcd_mul, for which
     asize limbs should be enough. */

  /* Limit on the recursion depth */
  unsigned k = sizeinbits(1 + asize / (HGCD_LEHMER_THRESHOLD  - 5));
  
  return asize + nhgcd_init_itch(asize + k * (6 + 12), asize);
}

int
nhgcd(struct nhgcd *hgcd,
      mp_srcptr ap, mp_size_t asize,
      mp_srcptr bp, mp_size_t bsize,
      struct nhgcd_quotient_stack *quotients,
      mp_ptr tp, mp_size_t talloc)
{
  mp_size_t N = asize;
  mp_size_t M = CHOOSE_M(N);
  mp_size_t n;
  mp_size_t m;

  struct nhgcd R;
  mp_size_t itch;

  ASSERT(M);

  if (bsize <= M)
    return 0;

  ASSERT(asize >= 2);
  
  if (N < HGCD_LEHMER_THRESHOLD)
    return nhgcd_lehmer(hgcd, ap, asize, bp, bsize,
			quotients, tp, talloc);

  /* Initialize, we keep r0 and r1 as the reduced numbers (so far). */
  nhgcd_start(hgcd, ap, asize, bp, bsize);

  /* Reduce the size to M + m + 1. Usually, only one hgcd call is
     needed, but we may need multiple calls. When finished, the values
     are stored in r0 (potentially large) and r1 (smaller size) */

  n = N - M;
  m = CHOOSE_M(n);

  /* The second recursive call can use numbers of size up to n+3 */
  itch = nhgcd_init_itch(n+3, n+3);
  
  ASSERT(itch <= talloc);
  nhgcd_init(&R, n+3, n+3, tp);
  tp += itch; talloc -= itch;
  
  while (hgcd->row[1].rsize > M + m + 1)
    {
      /* Max size after reduction, plus one */
      mp_size_t ralloc = hgcd->row[1].rsize + 1;
      switch (nhgcd(&R,
		    hgcd->row[0].rp + M, hgcd->row[0].rsize - M,
		    hgcd->row[1].rp + M, hgcd->row[1].rsize - M,
		    quotients, tp, talloc))
	{
	default:
	  abort();
	  
	case 0:
	  {
	    /* The first remainder was small. Then there's a good chance
	       that the remainder A % B is also small. */
	    int res = nhgcd_case0(hgcd, M, quotients, tp, talloc);

	    if (res > 0)
	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	    if (res >= 0)
	      return res;

	    nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);
	    break;
	  }

	case 2:
	  /* Now r0 and r1 are correct, while r2 may be too large.
	   * Compute r1 and r2, and check what happened */

	  nhgcd_sanity(&R,
		       hgcd->row[0].rp + M, hgcd->row[0].rsize - M,
		       hgcd->row[1].rp + M, hgcd->row[1].rsize - M,
		       0, 4);
	
	  /* Store new values in rows 2 and 3, to avoid overlap */
	  hgcd->row[2].rsize
	    = nhgcd_fix(M, hgcd->row[2].rp, ralloc,
			R.row[1].rp, R.row[1].rsize,
			~R.sign,
			R.row[1].uvp[0], R.row[1].uvsize[0], hgcd->row[0].rp,
			R.row[1].uvp[1], R.row[1].uvsize[1], hgcd->row[1].rp,
			tp, talloc);
			    
	  hgcd->row[3].rsize
	    = nhgcd_fix(M, hgcd->row[3].rp, ralloc,
			R.row[2].rp, R.row[2].rsize,
			R.sign,
			R.row[2].uvp[0], R.row[2].uvsize[0], hgcd->row[0].rp,
			R.row[2].uvp[1], R.row[2].uvsize[1], hgcd->row[1].rp,
			tp, talloc);

	  ASSERT(hgcd->row[2].rsize > M);
	  
	  if (MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize,
			 hgcd->row[2].rp, hgcd->row[2].rsize))
	    /* r2 was correct. */
	    goto correct_r2;
	  else
	    {
	      mp_size_t qsize;
	      mp_srcptr qp;
	      
	      /* r2 was too large, i.e. q0 too small. In this case we
		 must have r2 % r1 <= r2 - r1 smaller than M + m + 1. */

	      /* Computes the uv matrix for the (incorrect) values r1, r2.
		 The elements must be smaller than the correct ones, since
		 they correspond to a too small q. */
	      nhgcd_mul(hgcd->row + 2, hgcd->alloc,
			R.row + 1, hgcd->row,
			tp, talloc);
	    
	      /* Discard r3, and the corresponding quotient */
	      nhgcd_quotient_stack_drop(quotients);

	      /* Get the incorrect r0/r1 quotient used to compute r2 */
	      ASSERT(!START_ROW_P(hgcd->row[2]));
	      
	      qsize = nhgcd_quotient_stack_pop(quotients, &qp);
	      nhgcd_backup(hgcd->row + 1, qp, qsize);
	      hgcd->sign ^= ~R.sign;

	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 1, 4);
	    
	      /* Compute the correct r2. FIXME: We could do this
		 slightly more efficient by dividing r3/r2, and adding
		 this quotient to the earlier incorrect one. */
	      euclid_step(hgcd->row+1, quotients, hgcd->alloc);

	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 1, 4);

	      ASSERT(hgcd->row[3].rsize <= M + m + 1);

	      if (hgcd->row[3].rsize <= M)
		{
		  /* Backup one step. */
		  mp_size_t qsize;
		  mp_srcptr qp;

		  if (START_ROW_P(hgcd->row[1]))
		    {
		      nhgcd_quotient_stack_drop(quotients);
		      return 0;
		    }
	      
		  qsize = nhgcd_quotient_stack_get_1(quotients, &qp);
		  nhgcd_backup(hgcd->row, qp, qsize);		
#if WANT_CHECKS
		  nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
		  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
		  return nhgcd_jebelean(hgcd, M, tp, talloc);
		}

	      NHGCD_SWAP4_2(hgcd->row);
	      goto reduction_done;
	    }
	case 3:
	  /* Now r0, r1 and r2 are correct, while r3 may be too small
	     or too large. */
	  /* Compute r1 and r2, and check what happened */
	  nhgcd_sanity(&R,
		       hgcd->row[0].rp + M, hgcd->row[0].rsize - M,
		       hgcd->row[1].rp + M, hgcd->row[1].rsize - M,
		       0, 4);
	  
	  /* Store new values in rows 2 and 3, to avoid overlap */
	  hgcd->row[2].rsize
	    = nhgcd_fix(M, hgcd->row[2].rp, ralloc,
			R.row[1].rp, R.row[1].rsize,
			~R.sign,
			R.row[1].uvp[0], R.row[1].uvsize[0], hgcd->row[0].rp,
			R.row[1].uvp[1], R.row[1].uvsize[1], hgcd->row[1].rp,
			tp, talloc);
			    
	  hgcd->row[3].rsize
	    = nhgcd_fix(M, hgcd->row[3].rp, ralloc,
			R.row[2].rp, R.row[2].rsize,
			R.sign,
			R.row[2].uvp[0], R.row[2].uvsize[0], hgcd->row[0].rp,
			R.row[2].uvp[1], R.row[2].uvsize[1], hgcd->row[1].rp,
			tp, talloc);

	correct_r2:
	  /* Discard r3, and the corresponding quotient */
	  nhgcd_quotient_stack_drop(quotients);

	  ASSERT(hgcd->row[2].rsize > M);
	  ASSERT(hgcd->row[3].rsize > M);

	  ASSERT(MPN_LESS_P(hgcd->row[3].rp, hgcd->row[3].rsize,
			    hgcd->row[2].rp, hgcd->row[2].rsize));

	  nhgcd_mul(hgcd->row+2, hgcd->alloc,
		    R.row+1, hgcd->row,
		    tp, talloc);
	  hgcd->sign ^= ~R.sign;

	  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 2, 4);
	  
	  NHGCD_SWAP4_2(hgcd->row);

	  /* FIXME: Additional euclid step here? */
	  break;
#if 0	  
	  if (hgcd->row[1].rsize <= M + m + 1)
	    goto reduction_done;

	  /* Compute correct r3, there should be a reasonable chance
	     that it is small. */

	  euclid_step(hgcd->row, quotients->top, hgcd->alloc);

	  if (hgcd->row[2].rsize <= M)
	    {
	      if (quotients->depth == start)
		return 0;
	      hgcd->qsize[1] = qsize;

	      /* The same value we pushed earlier */
	      hgcd->qsize[0] = nhgcd_quotient_stack_pop(quotients);
	      ASSERT(hgcd->qsize[0] == R.qsize[0]);

	      NHGCD_SWAP4_RIGHT(hgcd->row);
	      hgcd->sign = ~hgcd->sign;
	    
	      nhgcd_backup(hgcd->row, quotients->top, hgcd->qsize[0]);
	      
#if WANT_CHECKS
	      nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	      return nhgcd_jebelean(hgcd, M, tp, talloc);
	    }

	  else if (hgcd->row[2].rsize <= M + m + 1)
	    {
	      NHGCD_SWAP4_LEFT(hgcd->row);
	      nhgcd_quotient_stack_push(quotients, qsize);

	      goto reduction_done;
	    }
#endif

	case 4:
	  /* All of r0, r1, r3 and r3 are correct.
	     Compute r2 and r3 */

	  nhgcd_sanity(&R,
		       hgcd->row[0].rp + M, hgcd->row[0].rsize - M,
		       hgcd->row[1].rp + M, hgcd->row[1].rsize - M,
		       0, 4);
	
	  /* Store new values in rows 2 and 3, to avoid overlap */
	  hgcd->row[2].rsize
	    = nhgcd_fix(M, hgcd->row[2].rp, ralloc,
			R.row[2].rp, R.row[2].rsize,
			R.sign,
			R.row[2].uvp[0], R.row[2].uvsize[0], hgcd->row[0].rp,
			R.row[2].uvp[1], R.row[2].uvsize[1], hgcd->row[1].rp,
			tp, talloc);
			    
	  hgcd->row[3].rsize
	    = nhgcd_fix(M, hgcd->row[3].rp, ralloc,
			R.row[3].rp, R.row[3].rsize,
			~R.sign,
			R.row[3].uvp[0], R.row[3].uvsize[0], hgcd->row[0].rp,
			R.row[3].uvp[1], R.row[3].uvsize[1], hgcd->row[1].rp,
			tp, talloc);

	  ASSERT(hgcd->row[2].rsize > M);
	  ASSERT(hgcd->row[3].rsize <= M + m + 1);

	  nhgcd_mul(hgcd->row+2, hgcd->alloc,
		    R.row+2, hgcd->row,
		    tp, talloc);
	  hgcd->sign ^= R.sign;
	  
	  if (hgcd->row[3].rsize <= M)
	    {
	      /* Backup two steps */
	      mp_size_t qsize;
	      mp_srcptr qp;

	      ASSERT(!START_ROW_P(hgcd->row[2]));
	      
	      qsize = nhgcd_quotient_stack_get_0(quotients, &qp);
	      nhgcd_backup(hgcd->row+1, qp, qsize);

	      ASSERT(!START_ROW_P(hgcd->row[1]));

	      qsize = nhgcd_quotient_stack_get_1(quotients, &qp);
	      nhgcd_backup(hgcd->row, qp, qsize);
		
#if WANT_CHECKS
	      nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
	      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	      return nhgcd_jebelean(hgcd, M, tp, talloc);
	    }
	  
	  NHGCD_SWAP4_2(hgcd->row);

	  goto reduction_done;
	}
    }

 reduction_done:
  ASSERT(hgcd->row[0].rsize >= hgcd->row[1].rsize);
  ASSERT(hgcd->row[1].rsize > M);
  ASSERT(hgcd->row[1].rsize <= M + m + 1);

  if (hgcd->row[0].rsize > M + m + 1)
    {
      /* One euclid step to reduce size. */
      /* FIXME: Rename nhgcd_case0 */
      int res = nhgcd_case0(hgcd, M, quotients,
			    tp, talloc);

      if (res > 0)
	nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
      if (res >= 0)
	return res;

      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);
    }
  
  ASSERT(hgcd->row[0].rsize >= hgcd->row[1].rsize);
  ASSERT(hgcd->row[0].rsize <= M + m + 1);
  ASSERT(hgcd->row[1].rsize > M);

  /* Second phase, reduce size until we have one number of size > M
     and one of size <= M+1 */
  while (hgcd->row[1].rsize > M + 1)
    {
      mp_size_t k = 2*M - hgcd->row[0].rsize;
      mp_size_t n1 = hgcd->row[0].rsize - k;
      mp_size_t m1 = CHOOSE_M(n1);
      int res;
      
      ASSERT(k + m1 == M);
      ASSERT(n1 >= 2);

      ASSERT(n1 <= 2*(m + 1));
      ASSERT(n1 <= n + 3);

      res = nhgcd(&R,
		  hgcd->row[0].rp + k, hgcd->row[0].rsize - k,
		  hgcd->row[1].rp + k, hgcd->row[1].rsize - k,
		  quotients, tp, talloc);

      if (!res)
	{
	  /* The first remainder was small. Then there's a good chance
	     that the remainder A % B is also small. */

	  res = nhgcd_case0(hgcd, M, quotients,
			    tp, talloc);

	  if (res > 0)
	    nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	  if (res >= 0)
	    return res;

	  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);
	  continue;
	}
      
      /* Now r0 and r1 are always correct. */

      /* Store new values in rows 2 and 3, to avoid overlap */
      hgcd->row[2].rsize
	= nhgcd_fix(k, hgcd->row[2].rp, hgcd->row[0].rsize + 1,
		    R.row[0].rp, R.row[0].rsize,
		    R.sign,
		    R.row[0].uvp[0], R.row[0].uvsize[0], hgcd->row[0].rp,
		    R.row[0].uvp[1], R.row[0].uvsize[1], hgcd->row[1].rp,
		    tp, talloc);
	  
      hgcd->row[3].rsize
	= nhgcd_fix(k, hgcd->row[3].rp, hgcd->row[1].rsize + 1,
		    R.row[1].rp, R.row[1].rsize,
		    ~R.sign,
		    R.row[1].uvp[0], R.row[1].uvsize[0], hgcd->row[0].rp,
		    R.row[1].uvp[1], R.row[1].uvsize[1], hgcd->row[1].rp,
		    tp, talloc);
	  
      ASSERT(hgcd->row[2].rsize > M);
      ASSERT(hgcd->row[3].rsize > k);

      nhgcd_mul(hgcd->row+2, hgcd->alloc,
		R.row, hgcd->row,
		tp, talloc);
      hgcd->sign ^= R.sign;

      nhgcd_sanity(hgcd, ap, asize, bp, bsize, 2, 4);
      
      if (hgcd->row[3].rsize <= M)
	{
	  /* Backup two steps */
	  mp_size_t qsize;
	  mp_srcptr qp;

	  /* We don't use R.row[2] and R.row[3], so drop the
	     corresponding quotients. */
	  nhgcd_quotient_stack_drop(quotients);
	  nhgcd_quotient_stack_drop(quotients);
	  
	  if (START_ROW_P(hgcd->row[2]))
	    return 0;
	  
	  qsize = nhgcd_quotient_stack_get_0(quotients, &qp);
	  nhgcd_backup(hgcd->row+1, qp, qsize);

	  if (START_ROW_P(hgcd->row[1]))
	    return 0;
	  
	  qsize = nhgcd_quotient_stack_get_1(quotients, &qp);
	  nhgcd_backup(hgcd->row, qp, qsize);
		
#if WANT_CHECKS
	  nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
	  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	  return nhgcd_jebelean(hgcd, M, tp, talloc);
	}

      NHGCD_SWAP4_2(hgcd->row);

      /* It's crucial that we drop quotients that we don't use, and
	 it's get a little complicated because we don't use the top
	 quotient first. */

      if (res >= 3)
	{
	  /* We already know the correct q */
	  mp_size_t qsize;
	  mp_srcptr qp;

	  qsize = nhgcd_quotient_stack_get_1(quotients, &qp);

	  nhgcd_update_r(hgcd->row, qp, qsize);
	  nhgcd_update_uv(hgcd->row, qp, qsize, hgcd->alloc);

	  ASSERT(hgcd->row[2].rsize > k);
	  if (hgcd->row[2].rsize <= M)
	    {
	      /* Discard r3 */
	      nhgcd_quotient_stack_drop(quotients);
	      return nhgcd_small(hgcd, M, quotients,
				 tp, talloc);
	    }
	  if (res == 3)
	    /* Drop quotient for r3 */
	    nhgcd_quotient_stack_drop(quotients);
	}
      else
	{
	  /* Discard r2 and r3, compute new r2 */
	  nhgcd_quotient_stack_drop(quotients);
	  nhgcd_quotient_stack_drop(quotients);
	  
	  euclid_step(hgcd->row, quotients, hgcd->alloc); 
	  if (hgcd->row[2].rsize <= M)
	    return nhgcd_small(hgcd, M, quotients,
			       tp, talloc);
	}

      ASSERT(hgcd->row[2].rsize > M);

      if (res == 4)
	{
	  /* We already know the correct q */
	  mp_size_t qsize;
	  mp_srcptr qp;

	  qsize = nhgcd_quotient_stack_get_0(quotients, &qp);

	  nhgcd_update_r(hgcd->row + 1, qp, qsize);
	  nhgcd_update_uv(hgcd->row + 1, qp, qsize, hgcd->alloc);
	  
	  ASSERT(hgcd->row[3].rsize <= M + 1);
	}
      else
	/* At this point, we have already dropped the old
	   quotient.. */
	euclid_step(hgcd->row + 1, quotients, hgcd->alloc);      
      
      if (hgcd->row[3].rsize <= M)
	{
#if WANT_CHECKS
	  nhgcd_quotient_stack_rotate(quotients, 0);
#endif  
	  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 4);
	  return nhgcd_jebelean(hgcd, M, tp, talloc);
	}

      NHGCD_SWAP4_2(hgcd->row);
    }

  nhgcd_sanity(hgcd, ap, asize, bp, bsize, 0, 2);

  return nhgcd_final(hgcd, M, quotients, tp, talloc);
}

mp_size_t
nhgcd_gcd_lehmer_itch(mp_size_t asize, mp_size_t bsize)
{
  return 5 * (asize + 1);
}

#define MAKE_ODD(p, size) do {				\
  int __make_odd_count;					\
  if ((p)[0] % 2 == 0)					\
    {							\
      count_trailing_zeros(__make_odd_count, (p)[0]);	\
      mpn_rshift((p), (p), (size), __make_odd_count);	\
      if ((p)[(size) - 1] == 0)				\
	(size)--;					\
    }							\
} while (0)

/* Called when we know x is even */
#define MAKE_ODD_2(xh, xl) do {					\
  if (xl == 0)							\
    MP_LIMB_T_SWAP(xl, xh);					\
  else								\
    {								\
      int __make_odd_2_count;					\
      count_trailing_zeros(__make_odd_2_count, xl);		\
      xl = (xl >> __make_odd_2_count)				\
         | (xh << (BITS_PER_MP_LIMB - __make_odd_2_count));	\
      xh = xh >> __make_odd_2_count;				\
    }								\
} while (0)

/* Called when we know x is even */
#define MAKE_ODD_1(x) do {			\
  int __make_odd_1_count;			\
  count_trailing_zeros(__make_odd_1_count, x);	\
  x = x >> __make_odd_1_count;			\
} while (0)

/* Unlike the other gcd functions, this one destroys the arguments. a
   and b must include space for an extra high limb. */
mp_size_t
nhgcd_gcd_binary(mp_ptr gp, mp_ptr ap, mp_size_t asize,
		 mp_ptr bp, mp_size_t bsize)
{
  /* General binary algorithm */
  mp_size_t zero_words = 0;
  mp_size_t size;
  unsigned shift = 0;
  
  trace("nhgcd_gcd_binary: a = %Nd, b = %Nd\n", ap, asize, bp, bsize);
  
  /* Numbers must be odd */
  if (ap[0] == 0 && bp[0] == 0)
    {
      do
	gp[zero_words++] = 0;
      while (ap[zero_words] == 0 && bp[zero_words] == 0);

      ap += zero_words; asize -= zero_words;
      bp += zero_words; bsize -= zero_words;
      gp += zero_words;
    }

  if (ap[0] % 2 == 0)
    {
      if (bp[0] % 2 == 0)
	{
	  int acount;
	  int bcount;
	  count_trailing_zeros(acount, ap[0]);
	  count_trailing_zeros(bcount, bp[0]);

	  shift = MIN(acount, bcount);

	  ASSERT_NOCARRY(mpn_rshift(ap, ap, asize, acount));
	  if (ap[asize - 1] == 0)
	    asize--;
	
	  ASSERT_NOCARRY(mpn_rshift(bp, bp, bsize, bcount));
	  if (bp[bsize - 1] == 0)
	    bsize--;
	}
      else
	MAKE_ODD(ap, asize);
    }
  else if (bp[0] % 2 == 0)
    MAKE_ODD(bp, bsize);

  ASSERT(asize);
  ASSERT(bsize);

  if (MPN_LESS_P(ap, asize,
		 bp, bsize))
    MPN_PTR_SWAP(ap, asize,
		 bp, bsize);

  size = mpn_gcd(gp, ap, asize, bp, bsize);
  if (shift)
    {
      mp_limb_t cy = mpn_lshift(gp, gp, size, shift);
      if (cy)
	gp[size++] = cy;
    }
  return zero_words + size;
  
#if 0
  while (asize > 2)
    {
      ASSERT(ap[0] % 2);
      ASSERT(bp[0] % 2);
    
      ASSERT_NOCARRY(mpn_sub(ap, ap, asize, bp, bsize));
      MPN_NORMALIZE(ap, asize);
      if (asize == 0)
	{
	  if (shift)
	    {
	      mp_limb_t cy = mpn_lshift(gp, bp, bsize, shift);
	      if (cy)
		gp[bsize++] = cy;
	    }
	  else
	    MPN_COPY(gp, bp, bsize);

	  return bsize + zero_words;
	}
      MAKE_ODD(ap, asize);

      if (MPN_LESS_P(ap, asize,
		     bp, bsize))
	MPN_PTR_SWAP(ap, asize,
		     bp, bsize);
    }
  {
    /* New locals */
    mp_limb_t ah = ap[1]; mp_limb_t al = ap[0];
    mp_limb_t bh = bp[1]; mp_limb_t bl = bp[0];
      
    while (ah && bh)
      {
	ASSERT(al % 2);
	ASSERT(bl % 2);
	
	{ /* Can we use overlapping source and destination with sub_ddmmss ? */
	  mp_limb_t th;
	  mp_limb_t tl;
	  sub_ddmmss(th, tl, ah, al, bh, bl);

	  ah = th; al = tl;
	}
	if (ah)
	  {
	    MAKE_ODD_2(ah, al);
	    if (ah < bh || (ah == bh && al < bl))
	      {
		MP_LIMB_T_SWAP(ah, bh);
		MP_LIMB_T_SWAP(al, bl);
	      }
	  }
	else
	  {
	    if (al == 0)
	      {
		if (shift)
		  {
		    mp_limb_t cy;

		    gp[0] = bl << shift;
		    gp[1] = (bh << shift) | (bl >> (BITS_PER_MP_LIMB - shift));
		    cy = bh >> (BITS_PER_MP_LIMB - shift);
		    if (cy)
		      {
			gp[2] = cy;
			return zero_words + 3;
		      }
		  }
		else
		  {
		    gp[0] = bl;
		    gp[1] = bh;
		  }
		return zero_words + 2;
	      }
	    MAKE_ODD_1(al);

	    /* Now b is larger, so swap */
	    MP_LIMB_T_SWAP(al, bl);
	    ah = bh;
	    break;
	  }
      }

    /* Now b is one limb, a might be two. Maintain a as the largest */
    if (ah)
      {
	while (ah)
	  {
	    ASSERT(al % 2);
	    ASSERT(bl % 2);
	    
	    if (bl > al)
	      ah--;
	    al -= bl;
	    
	    MAKE_ODD_2(ah, al);
	  }
	if (al < bl)
	  MP_LIMB_T_SWAP(al, bl);
      }
    /* Now we have single limbs left, and a might be zero */
    if (bl)
      while (al != bl)
	{
	  ASSERT(al % 2);
	  ASSERT(bl % 2);

	  if (al > bl)
	    {
	      al -= bl;
	      do
		al /= 2;
	      while (al % 2 == 0);
	    }
	  else /* al < bl */
	    {
	      bl -= al;
	      do
		bl /= 2;
	      while (bl % 2 == 0);
	    }
	}

    /* Now gcd == al 2^shift */
    if (shift)
      {
	mp_limb_t cy;

	gp[0] = al << shift;
	cy = al >> (BITS_PER_MP_LIMB - shift);
	if (cy)
	  {
	    gp[1] = cy;
	    return zero_words + 2;
	  }
      }
    else
      gp[0] = al;
      
    return zero_words + 1;
  }
#endif
}

  
mp_size_t
nhgcd_gcd_lehmer(mp_ptr gp, mp_srcptr ap, mp_size_t asize,
		 mp_srcptr bp, mp_size_t bsize,
		 mp_ptr tp, mp_size_t talloc)
{
  struct nhgcd_row r[4];
  mp_ptr qp;
  mp_size_t qsize;
  mp_size_t ralloc = asize + 1;
  
  ASSERT(asize >= bsize);
  ASSERT(bsize > 0);

  if (asize <= GCD_BINARY_THRESHOLD)
    {
      ASSERT(asize + bsize + 2<= talloc);

      MPN_COPY(tp, ap, asize);
      MPN_COPY(tp + asize + 1, bp, bsize);
      return nhgcd_gcd_binary(gp, tp, asize, tp + asize + 1, bsize);
    }
  
  ASSERT(5 * asize  + 4 <= talloc);
  
  r[0].rp = tp; tp += ralloc; talloc -= ralloc;
  r[1].rp = tp; tp += ralloc; talloc -= ralloc;
  r[2].rp = tp; tp += ralloc; talloc -= ralloc;
  r[3].rp = tp; tp += ralloc; talloc -= ralloc;
  qp = tp; tp += asize; talloc -= asize;
  
  MPN_COPY(r[0].rp, ap, asize); r[0].rsize = asize;
  MPN_COPY(r[1].rp, bp, bsize); r[1].rsize = bsize;

  /* u and v fields aren't used, but zero them out so that we can call
     trace_nhgcd_row */
  r[0].uvsize[0] = r[0].uvsize[1] = 0;
  r[1].uvsize[0] = r[1].uvsize[1] = 0;
  r[2].uvsize[0] = r[2].uvsize[1] = 0;
  r[3].uvsize[0] = r[3].uvsize[1] = 0;
  
  while (r[0].rsize > GCD_BINARY_THRESHOLD && r[1].rsize > 0)
    {
      struct nhgcd2 hgcd;
      int res = lehmer_step(&hgcd,
			    r[0].rp, r[0].rsize,
			    r[1].rp, r[1].rsize,
			    NULL);

      if (!res || (res == 2 && hgcd.row[0].v == 0))
	{
	  qsize = nhgcd_tdiv(qp, r[2].rp, &r[2].rsize,
			     r[0].rp, r[0].rsize,
			     r[1].rp, r[1].rsize);
	  NHGCD_SWAP3_LEFT(r);
	}
      else
	{
	  const struct nhgcd2_row *s = hgcd.row + (res - 2);
	  int sign = hgcd.sign;
	  if (res == 3)
	    sign = ~sign;
	  
	  /* s[0] and s[1] correct. */
	  r[2].rsize
	    = nhgcd2_fix(r[2].rp, ralloc,
			 sign,
			 s[0].u, r[0].rp, r[0].rsize,
			 s[0].v, r[1].rp, r[1].rsize);
	  
	  r[3].rsize
	    = nhgcd2_fix(r[3].rp, ralloc,
			 ~sign,
			 s[1].u, r[0].rp, r[0].rsize,
			 s[1].v, r[1].rp, r[1].rsize);

	  NHGCD_SWAP4_2(r);
	}
    }

  if (r[1].rsize == 0)
    {
      MPN_COPY(gp, r[0].rp, r[0].rsize);
      return r[0].rsize;
    }

  return nhgcd_gcd_binary(gp, r[0].rp, r[0].rsize, r[1].rp, r[1].rsize);
}

mp_size_t
nhgcd_gcd_schoenhage_itch(mp_size_t asize, mp_size_t bsize)
{
  /* Size for nhgcd calls */
  mp_size_t ralloc = bsize + 1;
  mp_size_t nhgcd_size = (bsize + 1) / 2;
  return
    asize                                     /* euclid quotient */
    + 4 * ralloc                              /* Remainder storage */
    + nhgcd_lehmer_itch(bsize, bsize)         /* lehmer call */
    + nhgcd_init_itch(nhgcd_size, nhgcd_size) /* nhgcd storage */
    + nhgcd_quotient_stack_itch(nhgcd_size)
    + nhgcd_itch(nhgcd_size, nhgcd_size)      /* nhgcd call */
    + 1+ 3 * bsize / 4;                       /* nhgcd_fix */
}

mp_size_t
nhgcd_gcd_schoenhage(mp_ptr gp, mp_srcptr ap, mp_size_t asize,
		     mp_srcptr bp, mp_size_t bsize,
		     mp_ptr tp, mp_size_t talloc)
{
  mp_size_t scratch;
  struct nhgcd hgcd;
  struct nhgcd_quotient_stack quotients;
  struct nhgcd_row r[4];
  
  mp_size_t ralloc = bsize + 1;
  
  ASSERT(asize >= bsize);
  ASSERT(bsize > 0);

  /* One in initial division */
  ASSERT(asize + 1 <= talloc);
  r[1].rp = tp; 
  nhgcd_tdiv(tp + bsize, r[1].rp, &r[1].rsize,
	     ap, asize, bp, bsize);

  if (r[1].rsize == 0)
    {
      MPN_COPY(gp, bp, bsize);
      return bsize;
    }

  if (bsize <= FGCD_LEHMER_THRESHOLD)
    return nhgcd_gcd_lehmer(gp, bp, bsize, r[1].rp, r[1].rsize,
			    tp + r[1].rsize, talloc - r[1].rsize);

  ASSERT(4 * ralloc <= talloc);
  tp += ralloc; talloc -= ralloc;
  r[0].rp = tp; tp += ralloc; talloc -= ralloc;
  r[2].rp = tp; tp += ralloc; talloc -= ralloc;
  r[3].rp = tp; tp += ralloc; talloc -= ralloc;
  
  MPN_COPY(r[0].rp, bp, bsize); r[0].rsize = bsize;

  /* We don't use the u and v fields, but zero them out so that we can
     call trace_nhgcd_row while debugging. */
  r[0].uvsize[0] = 0; r[0].uvsize[1] = 0;
  r[1].uvsize[0] = 0; r[1].uvsize[1] = 0;
  r[2].uvsize[0] = 0; r[2].uvsize[1] = 0;
  r[3].uvsize[0] = 0; r[3].uvsize[1] = 0;
  
  scratch = nhgcd_init_itch((bsize + 1)/2, (bsize + 1)/2);
  ASSERT(scratch <= talloc);
  nhgcd_init(&hgcd, (bsize + 1)/2, (bsize + 1)/2, tp);
  tp += scratch; talloc -= scratch;

  {
    mp_size_t nlimbs = nhgcd_quotient_stack_itch((bsize + 1)/2);
    
    ASSERT(nlimbs <= talloc);
    nhgcd_quotient_stack_init(&quotients, (bsize + 1) / 2,
			      tp, nlimbs);
    
    tp += nlimbs;
    talloc -= nlimbs;
  }
  
  while (r[0].rsize > FGCD_LEHMER_THRESHOLD && r[1].rsize > 0)
    {
      mp_size_t k = r[0].rsize / 2;
      int res;

#if 0      
      trace("nhgcd_gcd_schoenhage\n");
      trace_nhgcd_row(r);
      trace_nhgcd_row(r + 1);
#endif      
      if (r[1].rsize <= k)
	goto euclid;

      nhgcd_quotient_stack_reset(&quotients, r[0].rsize - k);
      
      res = nhgcd(&hgcd,
		  r[0].rp + k, r[0].rsize - k,
		  r[1].rp + k, r[1].rsize - k,
		  &quotients,
		  tp, talloc);

      if (!res || (res == 2 && hgcd.row[0].uvsize[1] == 0))
	{
	euclid:
	  ASSERT(r[0].rsize - r[1].rsize + 1 <= talloc);
	  nhgcd_tdiv(tp, r[2].rp, &r[2].rsize,
		     r[0].rp, r[0].rsize,
		     r[1].rp, r[1].rsize);

	  NHGCD_SWAP3_LEFT(r);
	}
      else
	{
	  const struct nhgcd_row *s = hgcd.row + (res - 2);
	  int sign = hgcd.sign;
	  if (res == 3)
	    sign = ~sign;
	  
	  /* s[0] and s[1] are correct */
	  r[2].rsize
	    = nhgcd_fix(k, r[2].rp, ralloc,
			s[0].rp, s[0].rsize,
			sign,
			s[0].uvp[0], s[0].uvsize[0], r[0].rp,
			s[0].uvp[1], s[0].uvsize[1], r[1].rp,
			tp, talloc);
			    
	  r[3].rsize
	    = nhgcd_fix(k, r[3].rp, ralloc,
			s[1].rp, s[1].rsize,
			~sign,
			s[1].uvp[0], s[1].uvsize[0], r[0].rp,
			s[1].uvp[1], s[1].uvsize[1], r[1].rp,
			tp, talloc);

	  NHGCD_SWAP4_2(r);
	}
    }

#if 0
  trace("nhgcd_gcd_schoenhage after loop\n");
  trace_nhgcd_row(r);
  trace_nhgcd_row(r + 1);
#endif
  
  if (r[1].rsize == 0)
    {
      MPN_COPY(gp, r[0].rp, r[0].rsize);
      return r[0].rsize;
    }
  else
    return nhgcd_gcd_lehmer(gp,
			    r[0].rp, r[0].rsize,
			    r[1].rp, r[1].rsize,
			    tp, talloc); 
}

mp_size_t
nhgcd_gcdext_lehmer_itch(mp_size_t asize, mp_size_t bsize)
{
  mp_size_t ralloc = asize + 1;
  mp_size_t ualloc = bsize + 1;

  return 4 * ralloc + 4 * ualloc + asize;
}

mp_size_t
nhgcd_gcdext_lehmer(mp_ptr gp, mp_ptr up, mp_size_t *usize,
		    mp_srcptr ap, mp_size_t asize,
		    mp_srcptr bp, mp_size_t bsize,
		    mp_ptr tp, mp_size_t talloc)
{
  struct nhgcd_row r[4];
  int rsign;
  
  mp_ptr qp;
  mp_size_t qsize;
  mp_size_t ralloc = asize + 1;
  mp_size_t ualloc = bsize + 1;
  ASSERT(asize >= bsize);
  ASSERT(bsize > 0);

  ASSERT(4 * ralloc + 4*ualloc + asize <= talloc);
  
  r[0].rp = tp; tp += ralloc; talloc -= ralloc;
  r[1].rp = tp; tp += ralloc; talloc -= ralloc;
  r[2].rp = tp; tp += ralloc; talloc -= ralloc;
  r[3].rp = tp; tp += ralloc; talloc -= ralloc;
  r[0].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  r[1].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  r[2].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  r[3].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  
  qp = tp; tp += asize; talloc -= asize;
  
  MPN_COPY(r[0].rp, ap, asize); r[0].rsize = asize;
  MPN_COPY(r[1].rp, bp, bsize); r[1].rsize = bsize;
  r[0].uvsize[0] = 1; r[0].uvp[0][0] = 1;
  r[1].uvsize[0] = 0;
  
  /* We don't use the v fields. But we zero them out so we can call
     trace_nhgcd_row while debugging. */
  r[0].uvsize[1] = 0;
  r[1].uvsize[1] = 0;
  r[2].uvsize[1] = 0;
  r[3].uvsize[1] = 0;

  rsign = 0;
  
  while (r[0].rsize >= 2 && r[1].rsize > 0)
    {
      struct nhgcd2 hgcd;
      int res = lehmer_step(&hgcd,
			    r[0].rp, r[0].rsize,
			    r[1].rp, r[1].rsize,
			    NULL);

#if 0
      trace("nhgcd_gcdext_lehmer: res = %d, rsign = %d\n", res, rsign);
      trace_nhgcd_row(r);
      trace_nhgcd_row(r + 1);
#endif
      
      if (!res || (res == 2 && hgcd.row[0].v == 0))
	{
	  qsize = nhgcd_tdiv(qp, r[2].rp, &r[2].rsize,
			     r[0].rp, r[0].rsize,
			     r[1].rp, r[1].rsize);
	  nhgcd_update_u(r, qp, qsize, ualloc);
	  NHGCD_SWAP3_LEFT(r);
	  rsign = ~rsign;
	}
      else
	{
	  const struct nhgcd2_row *s = hgcd.row + (res - 2);
	  int sign = hgcd.sign;
	  if (res == 3)
	    sign = ~sign;
	  
	  /* s[0] and s[1] correct. */
	  r[2].rsize
	    = nhgcd2_fix(r[2].rp, ralloc,
			 sign,
			 s[0].u, r[0].rp, r[0].rsize,
			 s[0].v, r[1].rp, r[1].rsize);
	  
	  r[3].rsize
	    = nhgcd2_fix(r[3].rp, ralloc,
			 ~sign,
			 s[1].u, r[0].rp, r[0].rsize,
			 s[1].v, r[1].rp, r[1].rsize);

	  nhgcd2_mul_vector(r + 2, ralloc, s, r);
	  rsign ^= sign;
	  NHGCD_SWAP4_2(r);
	}
    }

  if (r[1].rsize == 0)
    {
      MPN_COPY(gp, r[0].rp, r[0].rsize);
      MPN_COPY(up, r[0].uvp[0], r[0].uvsize[0]);
      *usize = (rsign >= 0) ? r[0].uvsize[0] : -r[0].uvsize[0];
      
      return r[0].rsize;
    }
  else
    {
      mp_limb_t u;
      mp_limb_t v;

      if (rsign >= 0)
	{
	  gp[0] = nhgcd_gcdext_1(&u, &v, r[0].rp[0], r[1].rp[0]);

	  *usize = mpn_addmul2_1(up, bsize,
				 r[0].uvp[0], r[0].uvsize[0], u,
				 r[1].uvp[0], r[1].uvsize[0], v);
	}
      else
	{
	  gp[0] = nhgcd_gcdext_1(&u, &v, r[1].rp[0], r[0].rp[0]);
	  *usize = mpn_addmul2_1(up, bsize,
				 r[1].uvp[0], r[1].uvsize[0], u,
				 r[0].uvp[0], r[0].uvsize[0], v);
	}
      
      return 1;
    }
}

#define COMPUTE_V_ITCH(asize, bsize, usize) \
((usize) + (asize) + 1 + (bsize))

/* Computes |v| = |(c - u a)| / b, where u may be positive or
   negative. */
static mp_size_t
compute_v(mp_ptr vp, mp_size_t valloc,
	  mp_srcptr ap, mp_size_t asize,
	  mp_srcptr bp, mp_size_t bsize,
	  mp_srcptr cp, mp_size_t csize,
	  mp_srcptr up, mp_size_t usize,
	  mp_ptr tp, mp_size_t talloc)
{
  mp_size_t size;
  mp_size_t vsize;
  mp_ptr rp;
  
  ASSERT(asize);
  ASSERT(bsize);
  ASSERT(csize);
  ASSERT(asize >= bsize);

  trace("compute_v: a = %Nd\n"
	"           b = %Nd\n"
	"           c = %Nd\n"
	"           u = %Nd\n",
	ap, asize, bp, bsize, cp, csize, up, usize);

  ASSERT(usize);
  
  if (usize > 0)
    {
      /* |v| = -v = (u a - c) / b */ 
      size = MPN_MUL_SIZE(asize, usize);
      ASSERT(size <= talloc);

      ASSERT(size);
      size -= mpn_mul_friendly(tp, up, usize, ap, asize);
      
      ASSERT(size >= csize);
      ASSERT_NOCARRY(mpn_sub(tp, tp, size, cp, csize));
      MPN_NORMALIZE(tp, size);
    }
  else if (usize < 0)
    {
      /* |v| = v = (c - u a) / b = (c + |u| a) / b */ 
      usize = -usize;

      size = MPN_MUL_SIZE(asize, usize);
      ASSERT(size <= talloc);

      if (size)
	size -= mpn_mul_friendly(tp, up, usize, ap, asize);

      ASSERT(size < talloc);
      ASSERT(csize < talloc);

      size = mpn_add_friendly(tp, tp, size, cp, csize);
    }
  else
    abort();
  
  if (size == 0)
    return 0;

  /* Now divide t / b. There must be no remainder */
  
  ASSERT(size >= bsize);
  ASSERT(size + bsize <= talloc);
  rp = tp + size;
  
  vsize = size + 1 - bsize;
  ASSERT(vsize <= valloc);

  mpn_tdiv_qr(vp, rp, 0, tp, size, bp, bsize);
  MPN_NORMALIZE(vp, vsize);

  /* Remainder must be zero */
#if WANT_ASSERT
  {
    mp_size_t i;
    for (i = 0; i < bsize; i++)
      {
	ASSERT(rp[i] == 0);
      }
  }
#endif
  return vsize;   
}

mp_size_t
nhgcd_gcdext_schoenhage_itch(mp_size_t asize, mp_size_t bsize)
{
  mp_size_t itch;
  
  mp_size_t ralloc = bsize + 1;
  mp_size_t ualloc = bsize + 1;
  /* Input size for hgcd calls */
  mp_size_t halloc = (bsize + 1) / 2;
  
  /* Storage for the rows, and for initial quotient */
  mp_size_t rstorage = 4 * ralloc + 4 * ualloc + asize;

  /* Storage for hgcd calls */
  mp_size_t tstorage = nhgcd_init_itch(halloc, halloc)
    + nhgcd_quotient_stack_itch(halloc)
    + nhgcd_itch(halloc, halloc);

  /* Storage needed for final nhgcd_gcdext_lehmer */
  mp_size_t lstorage = nhgcd_gcdext_lehmer_itch(FGCDEXT_LEHMER_THRESHOLD,
						FGCDEXT_LEHMER_THRESHOLD);
  
  /* Storage needed after final nhgcd_gcdext_lehmer */
  mp_size_t fstorage
    = COMPUTE_V_ITCH(FGCDEXT_LEHMER_THRESHOLD,
		     FGCDEXT_LEHMER_THRESHOLD,
		     ualloc);

  /* We need rstorage + MAX(tstorage, lstorage, fstorage) */
  
  itch = tstorage;
  if (lstorage > tstorage)
    itch = lstorage;
  if (fstorage > itch)
    itch = fstorage;

  return rstorage + itch;
}

mp_size_t
nhgcd_gcdext_schoenhage(mp_ptr gp, mp_ptr up, mp_size_t *usize,
			mp_srcptr ap, mp_size_t asize,
			mp_srcptr bp, mp_size_t bsize,
			mp_ptr tp, mp_size_t talloc)
{
  int rsign;
  struct nhgcd_row r[4];

  mp_size_t scratch;
  struct nhgcd hgcd;
  struct nhgcd_quotient_stack quotients;

  mp_ptr qp;

  mp_size_t qsize;
  mp_size_t ralloc = bsize + 1;
  mp_size_t ualloc = bsize + 1;
  ASSERT(asize >= bsize);
  ASSERT(bsize > 0);

  ASSERT(4 * ralloc + 4 * ualloc + asize <= talloc);
  
  r[0].rp = tp; tp += ralloc; talloc -= ralloc;
  r[1].rp = tp; tp += ralloc; talloc -= ralloc;
  r[2].rp = tp; tp += ralloc; talloc -= ralloc;
  r[3].rp = tp; tp += ralloc; talloc -= ralloc;
  r[0].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  r[1].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  r[2].uvp[0] = tp; tp += ualloc; talloc -= ualloc;
  r[3].uvp[0] = tp; tp += ualloc; talloc -= ualloc;

  qp = tp; tp += asize; talloc -= asize;
  
  ASSERT(asize >= bsize);
  ASSERT(bsize > 0);

  /* One initial division,

       r0 = b       = 0 a + 1 b;
       r1 = a mod b = 1 a - q b 
   */
  
  nhgcd_tdiv(qp, r[1].rp, &r[1].rsize,
	     ap, asize, bp, bsize);

  if (r[1].rsize == 0)
    {
      /* gcd(a, b) = b */
      *usize = 0;

      MPN_COPY(gp, bp, bsize);
      return bsize;
    }

  MPN_COPY(r[0].rp, bp, bsize); r[0].rsize = bsize;
  r[0].uvsize[0] = 0;
  r[1].uvsize[0] = 1; r[1].uvp[0][0] = 1;

  /* We don't use the v fields. But we zero them out so we can call
     trace_nhgcd_row while debugging. */
  r[0].uvsize[1] = 0;
  r[1].uvsize[1] = 0;
  r[2].uvsize[1] = 0;
  r[3].uvsize[1] = 0;

  rsign = -1;

  scratch = nhgcd_init_itch((bsize + 1)/2, (bsize + 1)/2);
  ASSERT(scratch <= talloc);
  nhgcd_init(&hgcd, (bsize + 1)/2, (bsize + 1)/2, tp);
  tp += scratch; talloc -= scratch;

  {
    mp_size_t nlimbs = nhgcd_quotient_stack_itch((bsize + 1)/2);
    
    ASSERT(nlimbs <= talloc);
    nhgcd_quotient_stack_init(&quotients, (bsize + 1)/2,
			      tp, nlimbs);
    
    tp += nlimbs;
    talloc -= nlimbs;
  }
  
  while (r[0].rsize > FGCDEXT_LEHMER_THRESHOLD && r[1].rsize > 0)
    {
      mp_size_t k = r[0].rsize / 2;
      int res;
      
      if (r[1].rsize <= k)
	goto euclid;

      nhgcd_quotient_stack_reset(&quotients, r[0].rsize - k);
      
      res = nhgcd(&hgcd,
		  r[0].rp + k, r[0].rsize - k,
		  r[1].rp + k, r[1].rsize - k,
		  &quotients,
		  tp, talloc);

      if (!res || (res == 2 && hgcd.row[0].uvsize[1] == 0))
	{
	euclid:
	  qsize = nhgcd_tdiv(qp, r[2].rp, &r[2].rsize,
			     r[0].rp, r[0].rsize,
			     r[1].rp, r[1].rsize);
	  nhgcd_update_u(r, qp, qsize, ualloc);
	  ASSERT(r[2].uvsize[0] < ualloc);
	  
	  NHGCD_SWAP3_LEFT(r);
	  rsign = ~rsign;
	}
      else
	{
	  const struct nhgcd_row *s = hgcd.row + (res - 2);
	  int sign = hgcd.sign;
	  if (res == 3)
	    sign = ~sign;
	  
	  /* s[0] and s[1] are correct */
	  r[2].rsize
	    = nhgcd_fix(k, r[2].rp, ralloc,
			s[0].rp, s[0].rsize,
			sign,
			s[0].uvp[0], s[0].uvsize[0], r[0].rp,
			s[0].uvp[1], s[0].uvsize[1], r[1].rp,
			tp, talloc);
			    
	  r[3].rsize
	    = nhgcd_fix(k, r[3].rp, ralloc,
			s[1].rp, s[1].rsize,
			~sign,
			s[1].uvp[0], s[1].uvsize[0], r[0].rp,
			s[1].uvp[1], s[1].uvsize[1], r[1].rp,
			tp, talloc);

	  nhgcd_mul_vector(r + 2, ualloc, s, r, tp, talloc);
	  ASSERT(r[2].uvsize[0] < ualloc);
	  ASSERT(r[3].uvsize[0] < ualloc);
	  
	  rsign ^= sign;
	  NHGCD_SWAP4_2(r);
	}
    }
  if (r[1].rsize == 0)
    {
      MPN_COPY(gp, r[0].rp, r[0].rsize);
      MPN_COPY(up, r[0].uvp[0], r[0].uvsize[0]);

      *usize = (rsign >= 0) ? r[0].uvsize[0] : - r[0].uvsize[0];
      return r[0].rsize;
    }
  else
    {
      /* We have r0 = u0 a + v0 b,
	         r1 = u1 a + v1 b

	 Compute g = s r0 + t r1 = (s u0 + t u1) a + (...) b
      */
      mp_size_t gsize;
      mp_size_t tsize;
      
      /* We don't need the space for hgcd and the quotient stack any more */
      tp -= scratch; talloc += scratch;

      /* Store s in r[2] and v in r[3] */
      gsize = nhgcd_gcdext_lehmer(gp, r[2].uvp[0], &r[2].uvsize[0],
				  r[0].rp, r[0].rsize,
				  r[1].rp, r[1].rsize,
				  tp, talloc);

      if (r[2].uvsize[0] == 0)
	{
	  /* s == 0  ==>  t = g / b == 1  ==> g = u1 a + (...) b */
	  MPN_COPY(up, r[1].uvp[0], r[1].uvsize[0]);
	  *usize = (rsign >= 0) ? -r[1].uvsize[0] : r[1].uvsize[0];

	  return gsize;
	}
      
      /* Compute t = (g - s r0) / r1, storing it in r[3] */
      r[3].uvsize[0] = compute_v(r[3].uvp[0], ualloc,
				 r[0].rp, r[0].rsize, r[1].rp, r[1].rsize,
				 gp, gsize,
				 r[2].uvp[0], r[2].uvsize[0],
				 tp, talloc);

      if (r[2].uvsize[0] < 0)
	{
	  r[2].uvsize[0] = - r[2].uvsize[0];
	  rsign = ~rsign;
	}

      ASSERT(r[2].uvsize[0] < ualloc);
      ASSERT(r[3].uvsize[0] < ualloc);
      
      tsize = MPN_MUL_SIZE(r[2].uvsize[0], r[0].uvsize[0]);
      ASSERT(tsize <= ualloc);
      if (tsize)
	tsize -= mpn_mul_friendly(up,
				  r[2].uvp[0], r[2].uvsize[0],
				  r[0].uvp[0], r[0].uvsize[0]);

      ASSERT(tsize < ualloc);

      r[2].uvsize[0] = MPN_MUL_SIZE(r[3].uvsize[0], r[1].uvsize[0]);
      ASSERT(r[2].uvsize[0] <= talloc);

      if (r[2].uvsize[0])
	r[2].uvsize[0] -= mpn_mul_friendly(r[2].uvp[0],
					   r[3].uvp[0], r[3].uvsize[0],
					   r[1].uvp[0], r[1].uvsize[0]);

      ASSERT(r[2].uvsize[0] < talloc);
      
      tsize = mpn_add_friendly(up, up, tsize, r[2].uvp[0], r[2].uvsize[0]);
      ASSERT(r[2].uvsize[0] < talloc);

      *usize = (rsign >= 0 ) ? tsize : -tsize;

      return gsize;
    }
}

#if WANT_CHECKS
# undef nhgcd_diff_smaller_p

/* Checks if a - b < c. Overwrites d. */

static int
nhgcd_diff_smaller_p_wrapper(mp_srcptr ap, mp_size_t asize,
			     mp_srcptr bp, mp_size_t bsize,
			     mp_ptr cp, mp_size_t csize)
{
  mp_ptr tp;
  mp_size_t tsize;

  int r1;
  int r2;
  
  tsize = asize;
  tp = alloca(tsize * sizeof(mp_limb_t));

  ASSERT_NOCARRY(mpn_sub(tp, ap, asize, bp, bsize));
  MPN_NORMALIZE(tp, tsize);

  r1 = MPN_LESS_P(tp, tsize, cp, csize);
  r2 = nhgcd_diff_smaller_p(ap, asize, bp, bsize, cp, csize);

  ASSERT(r1 == r2);

  return r1;
}
#endif
