/* For CHAR_BIT */
#include <limits.h>
#include <gmp.h>

#define NHGCD_MAX_QUOTIENTS (sizeof(mp_size_t) * CHAR_BIT + 2)
struct nhgcd_quotient_stack
{
  /* Through out the code we represent q = 1 with qsize = 0. */
  mp_size_t size[NHGCD_MAX_QUOTIENTS];
  mp_ptr limb;
  mp_size_t limb_alloc;
  
  /* Number of quotients to keep when we discard old quotients */
  unsigned nkeep;

  /* Top quotient is of size size[size_next-1], and starts at
     limb+limb_next - size[size_next-1]. We use size_next == 0 for an
     empty stack.*/ 
  unsigned size_next;
  mp_size_t limb_next;
};

mp_size_t
nhgcd_quotient_stack_itch(mp_size_t size);

void
nhgcd_quotient_stack_init(struct nhgcd_quotient_stack *stack,
			  mp_size_t asize,
			  mp_limb_t *limbs, mp_size_t alloc);


void
nhgcd_quotient_stack_reset(struct nhgcd_quotient_stack *stack,
			   mp_size_t asize);

int
nhgcd_quotient_stack_equal(const struct nhgcd_quotient_stack *a,
			   const struct nhgcd_quotient_stack *b);

void
trace_nhgcd_quotients(const char *label,
		      const struct nhgcd_quotient_stack *stack);


struct nhgcd2_row
{
  /* r is a remainder in the remainder sequence */
  mp_limb_t r[2];

  /* r = (-)u a + (-)v b */
  mp_limb_t u;
  mp_limb_t v;
};

struct nhgcd2
{
  /* Sign of the first row, sign >= 0 implies that u >= 0 and v <= 0,
     sign < 0 implies u <= 0, v >= 0 */
  int sign;
  
  struct nhgcd2_row row[4];
};

#if WANT_CHECKS
void
nhgcd2_sanity(const struct nhgcd2 *hgcd,
	      mp_srcptr ap, mp_srcptr bp);
#else
# define nhgcd2_sanity(hgcd, ap, bp)
#endif

void
trace_nhgcd2(const struct nhgcd2 *hgcd, unsigned n);

int
nhgcd2(struct nhgcd2 *hgcd, mp_srcptr ap, mp_srcptr bp,
       struct nhgcd_quotient_stack *quotients);

struct nhgcd_row
{
  mp_ptr rp; mp_size_t rsize;
  mp_ptr uvp[2]; mp_size_t uvsize[2];
};

struct nhgcd
{
  int sign;

  /* Space allocated for the uv entries, for sanity checking */
  mp_size_t alloc;
  
  struct nhgcd_row row[4];
};

mp_size_t
nhgcd_init_itch(mp_size_t asize, mp_size_t bsize);

void
nhgcd_init(struct nhgcd *hgcd,
	   mp_size_t asize, mp_size_t bsize,
	   mp_limb_t *limbs);

#if WANT_CHECKS
void
nhgcd_sanity(const struct nhgcd *hgcd,
	     mp_srcptr ap, mp_size_t asize,
	     mp_srcptr bp, mp_size_t bsize,
	     unsigned start, unsigned end);
#else
# define nhgcd_sanity(hgcd, ap, asize, bp, bsize, start, end)
#endif

int
nhgcd_equal(const struct nhgcd *A, const struct nhgcd *B);

void
trace_nhgcd_row(const struct nhgcd_row *r);

void
trace_nhgcd(struct nhgcd *hgcd);

mp_size_t
nhgcd_lehmer_itch(mp_size_t asize, mp_size_t bsize);

int
nhgcd_lehmer(struct nhgcd *hgcd,
	     mp_srcptr ap, mp_size_t asize,
	     mp_srcptr bp, mp_size_t bsize,
	     struct nhgcd_quotient_stack *quotients,
	     mp_ptr tp, mp_size_t talloc);

mp_size_t
nhgcd_itch(mp_size_t asize, mp_size_t bsize);

int
nhgcd(struct nhgcd *hgcd,
      mp_srcptr ap, mp_size_t asize,
      mp_srcptr bp, mp_size_t bsize,
      struct nhgcd_quotient_stack *quotients,
      mp_ptr tp, mp_size_t talloc);


/* Unlike the other gcd functions, this one destroys the arguments */
mp_size_t
nhgcd_gcd_binary(mp_ptr gp, mp_ptr ap, mp_size_t asize,
		 mp_ptr bp, mp_size_t bsize);

mp_size_t
nhgcd_gcd_lehmer_itch(mp_size_t asize, mp_size_t bsize);

mp_size_t
nhgcd_gcd_lehmer(mp_ptr gp, mp_srcptr ap, mp_size_t asize,
		 mp_srcptr bp, mp_size_t bsize,
		 mp_ptr tp, mp_size_t talloc);

mp_size_t
nhgcd_gcd_schoenhage_itch(mp_size_t asize, mp_size_t bsize);

mp_size_t
nhgcd_gcd_schoenhage(mp_ptr gp, mp_srcptr ap, mp_size_t asize,
		     mp_srcptr bp, mp_size_t bsize,
		     mp_ptr tp, mp_size_t talloc);

mp_size_t
nhgcd_gcdext_lehmer_itch(mp_size_t asize, mp_size_t bsize);

mp_size_t
nhgcd_gcdext_lehmer(mp_ptr gp, mp_ptr up, mp_size_t *usize,
		    mp_srcptr ap, mp_size_t asize,
		    mp_srcptr bp, mp_size_t bsize,
		    mp_ptr tp, mp_size_t talloc);

mp_size_t
nhgcd_gcdext_schoenhage_itch(mp_size_t asize, mp_size_t bsize);

mp_size_t
nhgcd_gcdext_schoenhage(mp_ptr gp, mp_ptr up, mp_size_t *usize,
			mp_srcptr ap, mp_size_t asize,
			mp_srcptr bp, mp_size_t bsize,
			mp_ptr tp, mp_size_t talloc);


/* Hairy helper function */
int
nhgcd_diff_smaller_p(mp_srcptr ap, mp_size_t asize,
		     mp_srcptr bp, mp_size_t bsize,
		     mp_ptr cp, mp_size_t csize);

mp_limb_t
nhgcd_gcdext_1_odd(mp_limb_t *up, mp_limb_t *vp, mp_limb_t a, mp_limb_t b);

mp_limb_t
nhgcd_gcdext_1(mp_limb_t *up, mp_limb_t *vp, mp_limb_t a, mp_limb_t b);
